@@ -39,7 +39,7 @@ WINE_DEFAULT_DEBUG_CHANNEL(dbghelp);
 struct elf_info
 {
     unsigned                    flags;          /* IN  one (or several) of the ELF_INFO constants */
-    DWORD_PTR                   dbg_hdr_addr;   /* OUT address of debug header (if ELF_INFO_DEBUG_HEADER is set) */
+    DWORD64                     dbg_hdr_addr;   /* OUT address of debug header (if ELF_INFO_DEBUG_HEADER is set) */
     struct module*              module;         /* OUT loaded module (if ELF_INFO_MODULE is set) */
     const WCHAR*                module_name;    /* OUT found module name (if ELF_INFO_NAME is set) */
 };
@@ -804,9 +804,11 @@ static const struct elf_sym *elf_lookup_symtab(const struct module* module,
         if (result)
         {
             FIXME("Already found symbol %s (%s) in symtab %s @%08x and %s @%08x\n",
-                  name, compiland_name,
-                  source_get(module, result->compiland->source), (unsigned int)result->sym.st_value,
-                  source_get(module, ste->compiland->source), (unsigned int)ste->sym.st_value);
+                  debugstr_a(name), debugstr_a(compiland_name),
+                  debugstr_a(source_get(module, result->compiland->source)),
+                  (unsigned int)result->sym.st_value,
+                  debugstr_a(source_get(module, ste->compiland->source)),
+                  (unsigned int)ste->sym.st_value);
         }
         else
         {
@@ -817,7 +819,7 @@ static const struct elf_sym *elf_lookup_symtab(const struct module* module,
     if (!result && !(result = weak_result))
     {
         FIXME("Couldn't find symbol %s!%s in symtab\n",
-              debugstr_w(module->modulename), name);
+              debugstr_w(module->modulename), debugstr_a(name));
         return NULL;
     }
     return &result->sym;
@@ -862,19 +864,19 @@ static void elf_finish_stabs_info(struct module* module, const struct hash_table
                 if (((struct symt_function*)sym)->ranges[0].low != elf_info->elf_addr &&
                     ((struct symt_function*)sym)->ranges[0].low != elf_info->elf_addr + symp->st_value)
                     FIXME("Changing address for %p/%s!%s from %I64x to %I64x\n",
-                          sym, debugstr_w(module->modulename), sym->hash_elt.name,
+                          sym, debugstr_w(module->modulename), debugstr_a(sym->hash_elt.name),
                           ((struct symt_function*)sym)->ranges[0].low,
                           elf_info->elf_addr + symp->st_value);
                 if (size && size != symp->st_size)
                     FIXME("Changing size for %p/%s!%s from %I64x to %I64x\n",
-                          sym, debugstr_w(module->modulename), sym->hash_elt.name,
+                          sym, debugstr_w(module->modulename), debugstr_a(sym->hash_elt.name),
                           size, symp->st_size);
 
                 ((struct symt_function*)sym)->ranges[0].low = elf_info->elf_addr + symp->st_value;
                 ((struct symt_function*)sym)->ranges[0].high = elf_info->elf_addr + symp->st_value + symp->st_size;
             } else
                 FIXME("Couldn't find %s!%s\n",
-                      debugstr_w(module->modulename), sym->hash_elt.name);
+                      debugstr_w(module->modulename), debugstr_a(sym->hash_elt.name));
             break;
         case SymTagData:
             switch (((struct symt_data*)sym)->kind)
@@ -891,7 +893,7 @@ static void elf_finish_stabs_info(struct module* module, const struct hash_table
                     if (((struct symt_data*)sym)->u.var.offset != elf_info->elf_addr &&
                         ((struct symt_data*)sym)->u.var.offset != elf_info->elf_addr + symp->st_value)
                         FIXME("Changing address for %p/%s!%s from %I64x to %I64x\n",
-                              sym, debugstr_w(module->modulename), sym->hash_elt.name,
+                              sym, debugstr_w(module->modulename), debugstr_a(sym->hash_elt.name),
                               ((struct symt_function*)sym)->ranges[0].low,
                               elf_info->elf_addr + symp->st_value);
                     ((struct symt_data*)sym)->u.var.offset = elf_info->elf_addr + symp->st_value;
@@ -899,7 +901,7 @@ static void elf_finish_stabs_info(struct module* module, const struct hash_table
                         DataIsFileStatic : DataIsGlobal;
                 } else
                     FIXME("Couldn't find %s!%s\n",
-                          debugstr_w(module->modulename), sym->hash_elt.name);
+                          debugstr_w(module->modulename), debugstr_a(sym->hash_elt.name));
                 break;
             default:;
             }
@@ -1082,11 +1084,8 @@ static BOOL elf_load_debug_info_from_map(struct module* module,
         }
         lret = dwarf2_parse(module, module->reloc_delta, thunks, fmap);
         ret = ret || lret;
-    }
-    if (wcsstr(module->modulename, S_ElfW) || !wcscmp(module->modulename, S_WineLoaderW))
-    {
         /* add the thunks for native libraries */
-        if (!(dbghelp_options & SYMOPT_PUBLICS_ONLY))
+        if (module->is_wine_builtin)
             elf_new_wine_thunks(module, ht_symtab, thunks);
     }
     /* add all the public symbols from symtab */
@@ -1241,7 +1240,8 @@ static BOOL elf_load_file_from_fmap(struct process* pcs, const WCHAR* filename,
         modfmt = HeapAlloc(GetProcessHeap(), 0,
                           sizeof(struct module_format) + sizeof(struct elf_module_info));
         if (!modfmt) return FALSE;
-        elf_info->module = module_new(pcs, filename, DMT_ELF, FALSE, modbase,
+        elf_info->module = module_new(pcs, filename, DMT_ELF,
+                                      module_is_wine_host(filename, L".so"), FALSE, modbase,
                                       fmap->u.elf.elf_size, 0, calc_crc32(fmap->u.elf.handle),
                                       elf_get_machine(fmap->u.elf.elfhdr.e_machine));
         if (!elf_info->module)
@@ -1339,41 +1339,53 @@ static BOOL elf_load_file_cb(void *param, HANDLE handle, const WCHAR *filename)
 /******************************************************************
  *		elf_search_auxv
  *
- * locate some a value from the debuggee auxiliary vector
+ * Locate a value from the debuggee auxiliary vector
  */
 static BOOL elf_search_auxv(const struct process* pcs, unsigned type, ULONG_PTR* val)
 {
     char        buffer[sizeof(SYMBOL_INFO) + MAX_SYM_NAME];
     SYMBOL_INFO*si = (SYMBOL_INFO*)buffer;
-    BYTE*       addr;
-    BYTE*       str;
-    BYTE*       str_max;
+    const unsigned ptr_size = pcs->is_host_64bit ? 8 : 4;
+    UINT64      envp;
+    UINT64      addr;
+    UINT64      str;
+    UINT64      str_max;
 
     si->SizeOfStruct = sizeof(*si);
     si->MaxNameLen = MAX_SYM_NAME;
-    if (!SymFromName(pcs->handle, "libwine.so.1!__wine_main_environ", si) ||
-        !(addr = (void*)(DWORD_PTR)si->Address) ||
-        !ReadProcessMemory(pcs->handle, addr, &addr, sizeof(addr), NULL) ||
-        !addr)
+    if (!SymFromName(pcs->handle, "ntdll.so!main_envp", si) ||
+        !si->Address ||
+        !read_process_integral_value(pcs, si->Address, &envp, ptr_size) ||
+        !envp)
     {
         FIXME("can't find symbol in module\n");
         return FALSE;
     }
     /* walk through envp[] */
     /* envp[] strings are located after the auxiliary vector, so protect the walk */
-    str_max = (void*)(DWORD_PTR)~0L;
-    while (ReadProcessMemory(pcs->handle, addr, &str, sizeof(str), NULL) &&
-           (addr = (void*)((DWORD_PTR)addr + sizeof(str))) != NULL && str != NULL)
-        str_max = min(str_max, str);
+    str_max = ~(UINT64)0u;
+    addr = envp;
+    for (;;)
+    {
+        if (!read_process_integral_value(pcs, addr, &str, ptr_size) || (addr += ptr_size) <= ptr_size)
+            return FALSE;
+        if (!str) break;
+        /* It can be some env vars have been changed, pointing to a different location */
+        if (str >= envp)
+            str_max = min(str_max, str);
+    }
 
     /* Walk through the end of envp[] array.
      * Actually, there can be several NULLs at the end of envp[]. This happens when an env variable is
      * deleted, the last entry is replaced by an extra NULL.
      */
-    while (addr < str_max && ReadProcessMemory(pcs->handle, addr, &str, sizeof(str), NULL) && str == NULL)
-        addr = (void*)((DWORD_PTR)addr + sizeof(str));
+    for (; addr < str_max; addr += ptr_size)
+    {
+        if (!read_process_integral_value(pcs, addr, &str, ptr_size)) return FALSE;
+        if (str) break;
+    }
 
-    if (pcs->is_64bit)
+    if (pcs->is_host_64bit)
     {
         struct
         {
@@ -1381,7 +1393,7 @@ static BOOL elf_search_auxv(const struct process* pcs, unsigned type, ULONG_PTR*
             UINT64 a_val;
         } auxv;
 
-        while (ReadProcessMemory(pcs->handle, addr, &auxv, sizeof(auxv), NULL) && auxv.a_type)
+        while (read_process_memory(pcs, addr, &auxv, sizeof(auxv)) && auxv.a_type)
         {
             if (auxv.a_type == type)
             {
@@ -1399,7 +1411,7 @@ static BOOL elf_search_auxv(const struct process* pcs, unsigned type, ULONG_PTR*
             UINT32 a_val;
         } auxv;
 
-        while (ReadProcessMemory(pcs->handle, addr, &auxv, sizeof(auxv), NULL) && auxv.a_type)
+        while (read_process_memory(pcs, addr, &auxv, sizeof(auxv)) && auxv.a_type)
         {
             if (auxv.a_type == type)
             {
@@ -1445,7 +1457,7 @@ static BOOL elf_search_and_load_file(struct process* pcs, const WCHAR* filename,
         load_elf.elf_info    = elf_info;
 
         ret = search_unix_path(filename, process_getenv(pcs, L"LD_LIBRARY_PATH"), elf_load_file_cb, &load_elf)
-            || search_dll_path(pcs, filename, elf_load_file_cb, &load_elf);
+            || search_dll_path(pcs, filename, IMAGE_FILE_MACHINE_UNKNOWN, elf_load_file_cb, &load_elf);
     }
 
     return ret;
@@ -1519,7 +1531,7 @@ static BOOL elf_enum_modules_internal(const struct process* pcs,
     WCHAR bufstr[MAX_PATH];
     ULONG_PTR lm_addr;
 
-    if (pcs->is_64bit)
+    if (pcs->is_host_64bit)
     {
         struct
         {
@@ -1823,8 +1835,13 @@ BOOL elf_read_wine_loader_dbg_info(struct process* pcs, ULONG_PTR addr)
         HeapFree(GetProcessHeap(), 0, loader);
     }
     if (!ret || !elf_info.dbg_hdr_addr) return FALSE;
-
-    TRACE("Found ELF debug header %#Ix\n", elf_info.dbg_hdr_addr);
+    if (elf_info.dbg_hdr_addr != (ULONG_PTR)elf_info.dbg_hdr_addr)
+    {
+        ERR("Unable to access ELF libraries (outside 32bit limit)\n");
+        module_remove(pcs, elf_info.module);
+        return FALSE;
+    }
+    TRACE("Found ELF debug header %#I64x\n", elf_info.dbg_hdr_addr);
     elf_info.module->format_info[DFI_ELF]->u.elf_info->elf_loader = 1;
     module_set_module(elf_info.module, S_WineLoaderW);
     pcs->dbg_hdr_addr = elf_info.dbg_hdr_addr;
