@@ -26,6 +26,8 @@
 #include <string.h>
 #include <assert.h>
 
+#include "ntstatus.h"
+#define WIN32_NO_STATUS
 #include "dbghelp_private.h"
 #include "image_private.h"
 #include "winternl.h"
@@ -241,11 +243,14 @@ static BOOL pe_is_valid_pointer_table(const IMAGE_NT_HEADERS* nthdr, const void*
  *
  * Maps an PE file into memory (and checks it's a real PE file)
  */
-BOOL pe_map_file(HANDLE file, struct image_file_map* fmap, enum module_type mt)
+BOOL pe_map_file(HANDLE file, struct image_file_map* fmap)
 {
-    void*       mapping;
+    void*                       mapping;
+    IMAGE_NT_HEADERS*           nthdr;
+    IMAGE_SECTION_HEADER*       section;
+    unsigned                    i;
 
-    fmap->modtype = mt;
+    fmap->modtype = DMT_PE;
     fmap->ops = &pe_file_map_ops;
     fmap->alternate = NULL;
     fmap->u.pe.hMap = CreateFileMappingW(file, NULL, PAGE_READONLY, 0, 0, NULL);
@@ -254,78 +259,60 @@ BOOL pe_map_file(HANDLE file, struct image_file_map* fmap, enum module_type mt)
     fmap->u.pe.full_map = NULL;
     if (!(mapping = pe_map_full(fmap, NULL))) goto error;
 
-    switch (mt)
+    if (!(nthdr = RtlImageNtHeader(mapping))) goto error;
+    memcpy(&fmap->u.pe.file_header, &nthdr->FileHeader, sizeof(fmap->u.pe.file_header));
+    switch (nthdr->OptionalHeader.Magic)
     {
-    case DMT_PE:
-        {
-            IMAGE_NT_HEADERS*       nthdr;
-            IMAGE_SECTION_HEADER*   section;
-            unsigned                i;
-
-            if (!(nthdr = RtlImageNtHeader(mapping))) goto error;
-            memcpy(&fmap->u.pe.file_header, &nthdr->FileHeader, sizeof(fmap->u.pe.file_header));
-            switch (nthdr->OptionalHeader.Magic)
-            {
-            case IMAGE_NT_OPTIONAL_HDR32_MAGIC:
-                if (sizeof(void*) == 8 && !(SymGetOptions() & SYMOPT_INCLUDE_32BIT_MODULES))
-                {
-                    TRACE("Won't load 32bit module in 64bit dbghelp when options don't ask for it\n");
-                    goto error;
-                }
-                fmap->addr_size = 32;
-                memcpy(&fmap->u.pe.opt.header32, &nthdr->OptionalHeader, sizeof(fmap->u.pe.opt.header32));
-                break;
-            case IMAGE_NT_OPTIONAL_HDR64_MAGIC:
-                if (sizeof(void*) == 4) return FALSE;
-                fmap->addr_size = 64;
-                memcpy(&fmap->u.pe.opt.header64, &nthdr->OptionalHeader, sizeof(fmap->u.pe.opt.header64));
-                break;
-            default:
-                return FALSE;
-            }
-
-            fmap->u.pe.builtin = !memcmp((const IMAGE_DOS_HEADER*)mapping + 1, builtin_signature, sizeof(builtin_signature));
-            section = (IMAGE_SECTION_HEADER*)
-                ((char*)&nthdr->OptionalHeader + nthdr->FileHeader.SizeOfOptionalHeader);
-            fmap->u.pe.sect = HeapAlloc(GetProcessHeap(), 0,
-                                        nthdr->FileHeader.NumberOfSections * sizeof(fmap->u.pe.sect[0]));
-            if (!fmap->u.pe.sect) goto error;
-            for (i = 0; i < nthdr->FileHeader.NumberOfSections; i++)
-            {
-                memcpy(&fmap->u.pe.sect[i].shdr, section + i, sizeof(IMAGE_SECTION_HEADER));
-                fmap->u.pe.sect[i].mapped = IMAGE_NO_MAP;
-            }
-            if (nthdr->FileHeader.PointerToSymbolTable && nthdr->FileHeader.NumberOfSymbols)
-            {
-                LARGE_INTEGER li;
-
-                if (GetFileSizeEx(file, &li) && pe_is_valid_pointer_table(nthdr, mapping, li.QuadPart))
-                {
-                    /* FIXME ugly: should rather map the relevant content instead of copying it */
-                    const char* src = (const char*)mapping +
-                        nthdr->FileHeader.PointerToSymbolTable +
-                        nthdr->FileHeader.NumberOfSymbols * sizeof(IMAGE_SYMBOL);
-                    char* dst;
-                    DWORD sz = *(DWORD*)src;
-
-                    if ((dst = HeapAlloc(GetProcessHeap(), 0, sz)))
-                        memcpy(dst, src, sz);
-                    fmap->u.pe.strtable = dst;
-                }
-                else
-                {
-                    WARN("Bad coff table... wipping out\n");
-                    /* we have bad information here, wipe it out */
-                    fmap->u.pe.file_header.PointerToSymbolTable = 0;
-                    fmap->u.pe.file_header.NumberOfSymbols = 0;
-                    fmap->u.pe.strtable = NULL;
-                }
-            }
-            else fmap->u.pe.strtable = NULL;
-        }
+    case IMAGE_NT_OPTIONAL_HDR32_MAGIC:
+        fmap->addr_size = 32;
+        memcpy(&fmap->u.pe.opt.header32, &nthdr->OptionalHeader, sizeof(fmap->u.pe.opt.header32));
         break;
-    default: assert(0); goto error;
+    case IMAGE_NT_OPTIONAL_HDR64_MAGIC:
+        fmap->addr_size = 64;
+        memcpy(&fmap->u.pe.opt.header64, &nthdr->OptionalHeader, sizeof(fmap->u.pe.opt.header64));
+        break;
+    default:
+        goto error;
     }
+
+    fmap->u.pe.builtin = !memcmp((const IMAGE_DOS_HEADER*)mapping + 1, builtin_signature, sizeof(builtin_signature));
+    section = IMAGE_FIRST_SECTION( nthdr );
+    fmap->u.pe.sect = HeapAlloc(GetProcessHeap(), 0,
+                                nthdr->FileHeader.NumberOfSections * sizeof(fmap->u.pe.sect[0]));
+    if (!fmap->u.pe.sect) goto error;
+    for (i = 0; i < nthdr->FileHeader.NumberOfSections; i++)
+    {
+        memcpy(&fmap->u.pe.sect[i].shdr, section + i, sizeof(IMAGE_SECTION_HEADER));
+        fmap->u.pe.sect[i].mapped = IMAGE_NO_MAP;
+    }
+    if (nthdr->FileHeader.PointerToSymbolTable && nthdr->FileHeader.NumberOfSymbols)
+    {
+        LARGE_INTEGER li;
+
+        if (GetFileSizeEx(file, &li) && pe_is_valid_pointer_table(nthdr, mapping, li.QuadPart))
+        {
+            /* FIXME ugly: should rather map the relevant content instead of copying it */
+            const char* src = (const char*)mapping +
+                nthdr->FileHeader.PointerToSymbolTable +
+                nthdr->FileHeader.NumberOfSymbols * sizeof(IMAGE_SYMBOL);
+            char* dst;
+            DWORD sz = *(DWORD*)src;
+
+            if ((dst = HeapAlloc(GetProcessHeap(), 0, sz)))
+                memcpy(dst, src, sz);
+            fmap->u.pe.strtable = dst;
+        }
+        else
+        {
+            WARN("Bad coff table... wipping out\n");
+            /* we have bad information here, wipe it out */
+            fmap->u.pe.file_header.PointerToSymbolTable = 0;
+            fmap->u.pe.file_header.NumberOfSymbols = 0;
+            fmap->u.pe.strtable = NULL;
+        }
+    }
+    else fmap->u.pe.strtable = NULL;
+
     pe_unmap_full(fmap);
 
     return TRUE;
@@ -354,6 +341,52 @@ const char* pe_map_directory(struct module* module, int dirno, DWORD* size)
                            nth->OptionalHeader.DataDirectory[dirno].VirtualAddress, NULL);
 }
 
+BOOL pe_unmap_directory(struct module* module, int dirno, const char *dir)
+{
+    if (module->type != DMT_PE || !module->format_info[DFI_PE]) return FALSE;
+    if (dirno >= IMAGE_NUMBEROF_DIRECTORY_ENTRIES) return FALSE;
+    pe_unmap_full(&module->format_info[DFI_PE]->u.pe_info->fmap);
+    return TRUE;
+}
+
+/* Locks a region from a mapped PE file, from its RVA, and for at least 'size' bytes.
+ * Region must fit entirely inside a PE section.
+ * 'length', upon success, gets the size from RVA until end of PE section.
+ */
+const BYTE* pe_lock_region_from_rva(struct module *module, DWORD rva, DWORD size, DWORD *length)
+{
+    IMAGE_NT_HEADERS*     nth;
+    void*                 mapping;
+    IMAGE_SECTION_HEADER *section;
+    const BYTE           *ret;
+
+    if (module->type != DMT_PE || !module->format_info[DFI_PE]) return NULL;
+    if (!(mapping = pe_map_full(&module->format_info[DFI_PE]->u.pe_info->fmap, &nth)))
+        return NULL;
+    section = NULL;
+    ret = RtlImageRvaToVa(nth, mapping, rva, &section);
+    if (ret)
+    {
+        if (rva + size <= section->VirtualAddress + section->SizeOfRawData)
+        {
+            if (length)
+                *length = section->VirtualAddress + section->SizeOfRawData - rva;
+            return ret;
+        }
+        if (rva + size <= section->VirtualAddress + section->Misc.VirtualSize)
+            FIXME("Not able to lock regions not present on file\n");
+    }
+    pe_unmap_full(&module->format_info[DFI_PE]->u.pe_info->fmap);
+    return NULL;
+}
+
+BOOL pe_unlock_region(struct module *module, const BYTE* region)
+{
+    if (module->type != DMT_PE || !module->format_info[DFI_PE] || !region) return FALSE;
+    pe_unmap_full(&module->format_info[DFI_PE]->u.pe_info->fmap);
+    return TRUE;
+}
+
 static void pe_module_remove(struct process* pcs, struct module_format* modfmt)
 {
     image_unmap_file(&modfmt->u.pe_info->fmap);
@@ -408,7 +441,7 @@ static BOOL pe_locate_with_coff_symbol_table(struct module* module)
                     !strcmp(sym->hash_elt.name, name))
                 {
                     TRACE("Changing absolute address for %d.%s: %Ix -> %I64x\n",
-                          isym->SectionNumber, name, sym->u.var.offset,
+                          isym->SectionNumber, debugstr_a(name), sym->u.var.offset,
                           module->module.BaseOfImage +
                           fmap->u.pe.sect[isym->SectionNumber - 1].shdr.VirtualAddress +
                           isym->Value);
@@ -555,15 +588,15 @@ static BOOL pe_load_dwarf(struct module* module)
 static BOOL pe_load_dbg_file(const struct process* pcs, struct module* module,
                              const char* dbg_name, DWORD timestamp)
 {
-    WCHAR tmp[MAX_PATH];
     HANDLE                              hFile = INVALID_HANDLE_VALUE, hMap = 0;
     const BYTE*                         dbg_mapping = NULL;
     BOOL                                ret = FALSE;
+    SYMSRV_INDEX_INFOW                  info;
 
     TRACE("Processing DBG file %s\n", debugstr_a(dbg_name));
 
-    if (path_find_symbol_file(pcs, module, dbg_name, DMT_DBG, NULL, timestamp, 0, tmp, &module->module.DbgUnmatched) &&
-        (hFile = CreateFileW(tmp, GENERIC_READ, FILE_SHARE_READ, NULL,
+    if (path_find_symbol_file(pcs, module, dbg_name, FALSE, NULL, timestamp, 0, &info, &module->module.DbgUnmatched) &&
+        (hFile = CreateFileW(info.dbgfile, GENERIC_READ, FILE_SHARE_READ, NULL,
                              OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL)) != INVALID_HANDLE_VALUE &&
         ((hMap = CreateFileMappingW(hFile, NULL, PAGE_READONLY, 0, 0, NULL)) != 0) &&
         ((dbg_mapping = MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0)) != NULL))
@@ -586,7 +619,7 @@ static BOOL pe_load_dbg_file(const struct process* pcs, struct module* module,
                                       hdr->DebugDirectorySize / sizeof(*dbg));
     }
     else
-        ERR("Couldn't find .DBG file %s (%s)\n", debugstr_a(dbg_name), debugstr_w(tmp));
+        ERR("Couldn't find .DBG file %s (%s)\n", debugstr_a(dbg_name), debugstr_w(info.dbgfile));
 
     if (dbg_mapping) UnmapViewOfFile(dbg_mapping);
     if (hMap) CloseHandle(hMap);
@@ -611,7 +644,7 @@ static BOOL pe_load_msc_debug_info(const struct process* pcs, struct module* mod
     if (!(mapping = pe_map_full(fmap, &nth))) return FALSE;
     /* Read in debug directory */
     dbg = RtlImageDirectoryEntryToData( mapping, FALSE, IMAGE_DIRECTORY_ENTRY_DEBUG, &nDbg );
-    if (!dbg || !(nDbg /= sizeof(IMAGE_DEBUG_DIRECTORY))) goto done;
+    nDbg = dbg ? nDbg / sizeof(IMAGE_DEBUG_DIRECTORY) : 0;
 
     /* Parse debug directory */
     if (nth->FileHeader.Characteristics & IMAGE_FILE_DEBUG_STRIPPED)
@@ -623,8 +656,8 @@ static BOOL pe_load_msc_debug_info(const struct process* pcs, struct module* mod
         if (nDbg != 1 || dbg->Type != IMAGE_DEBUG_TYPE_MISC ||
             misc->DataType != IMAGE_DEBUG_MISC_EXENAME)
         {
-            ERR("-Debug info stripped, but no .DBG file in module %s\n",
-                debugstr_w(module->modulename));
+            WARN("-Debug info stripped, but no .DBG file in module %s\n",
+                 debugstr_w(module->modulename));
         }
         else
         {
@@ -633,12 +666,10 @@ static BOOL pe_load_msc_debug_info(const struct process* pcs, struct module* mod
     }
     else
     {
-        const IMAGE_SECTION_HEADER *sectp = (const IMAGE_SECTION_HEADER*)((const char*)&nth->OptionalHeader + nth->FileHeader.SizeOfOptionalHeader);
         /* Debug info is embedded into PE module */
-        ret = pe_load_debug_directory(pcs, module, mapping, sectp,
+        ret = pe_load_debug_directory(pcs, module, mapping, IMAGE_FIRST_SECTION( nth ),
                                       nth->FileHeader.NumberOfSections, dbg, nDbg);
     }
-done:
     pe_unmap_full(fmap);
     return ret;
 }
@@ -671,10 +702,8 @@ static BOOL pe_load_export_debug_info(const struct process* pcs, struct module*
 #if 0
     /* FIXME: we'd better store addresses linked to sections rather than 
        absolute values */
-    IMAGE_SECTION_HEADER*       section;
+    IMAGE_SECTION_HEADER *section = IMAGE_FIRST_SECTION( nth );
     /* Add start of sections */
-    section = (IMAGE_SECTION_HEADER*)
-        ((char*)&nth->OptionalHeader + nth->FileHeader.SizeOfOptionalHeader);
     for (i = 0; i < nth->FileHeader.NumberOfSections; i++, section++) 
     {
 	symt_new_public(module, NULL, section->Name, FALSE,
@@ -737,9 +766,12 @@ BOOL pe_load_debug_info(const struct process* pcs, struct module* module)
 
     if (!(dbghelp_options & SYMOPT_PUBLICS_ONLY))
     {
-        ret = image_check_alternate(&module->format_info[DFI_PE]->u.pe_info->fmap, module);
-        ret = pe_load_stabs(pcs, module) || ret;
-        ret = pe_load_dwarf(module) || ret;
+        if (!module->dont_load_symbols)
+        {
+            ret = image_check_alternate(&module->format_info[DFI_PE]->u.pe_info->fmap, module);
+            ret = pe_load_stabs(pcs, module) || ret;
+            ret = pe_load_dwarf(module) || ret;
+        }
         ret = pe_load_msc_debug_info(pcs, module) || ret;
         ret = ret || pe_load_coff_symbol_table(module); /* FIXME */
         /* if we still have no debug info (we could only get SymExport at this
@@ -763,13 +795,10 @@ struct builtin_search
 static BOOL search_builtin_pe(void *param, HANDLE handle, const WCHAR *path)
 {
     struct builtin_search *search = param;
-    size_t size;
 
-    if (!pe_map_file(handle, &search->fmap, DMT_PE)) return FALSE;
+    if (!pe_map_file(handle, &search->fmap)) return FALSE;
 
-    size = (lstrlenW(path) + 1) * sizeof(WCHAR);
-    if ((search->path = heap_alloc(size)))
-        memcpy(search->path, path, size);
+    search->path = wcsdup(path);
     return TRUE;
 }
 
@@ -784,6 +813,7 @@ struct module* pe_load_native_module(struct process* pcs, const WCHAR* name,
     BOOL                        opened = FALSE;
     struct module_format*       modfmt;
     WCHAR                       loaded_name[MAX_PATH];
+    WCHAR*                      real_path = NULL;
 
     loaded_name[0] = '\0';
     if (!hFile)
@@ -795,46 +825,74 @@ struct module* pe_load_native_module(struct process* pcs, const WCHAR* name,
             return NULL;
         opened = TRUE;
     }
-    else if (name) lstrcpyW(loaded_name, name);
-    if (!(modfmt = HeapAlloc(GetProcessHeap(), 0, sizeof(struct module_format) + sizeof(struct pe_module_info))))
-        return NULL;
-    modfmt->u.pe_info = (struct pe_module_info*)(modfmt + 1);
-    if (pe_map_file(hFile, &modfmt->u.pe_info->fmap, DMT_PE))
+    else
     {
-        struct builtin_search builtin = { NULL };
-        if (modfmt->u.pe_info->fmap.u.pe.builtin && search_dll_path(pcs, loaded_name, search_builtin_pe, &builtin))
-        {
-            TRACE("reloaded %s from %s\n", debugstr_w(loaded_name), debugstr_w(builtin.path));
-            image_unmap_file(&modfmt->u.pe_info->fmap);
-            modfmt->u.pe_info->fmap = builtin.fmap;
-        }
-        if (!base) base = PE_FROM_OPTHDR(&modfmt->u.pe_info->fmap, ImageBase);
-        if (!size) size = PE_FROM_OPTHDR(&modfmt->u.pe_info->fmap, SizeOfImage);
+        ULONG sz = sizeof(OBJECT_NAME_INFORMATION) + MAX_PATH * sizeof(WCHAR), needed;
+        OBJECT_NAME_INFORMATION *obj_name;
+        NTSTATUS nts;
 
-        module = module_new(pcs, loaded_name, DMT_PE, FALSE, base, size,
-                            modfmt->u.pe_info->fmap.u.pe.file_header.TimeDateStamp,
-                            PE_FROM_OPTHDR(&modfmt->u.pe_info->fmap, CheckSum),
-                            modfmt->u.pe_info->fmap.u.pe.file_header.Machine);
-        if (module)
+        obj_name = RtlAllocateHeap(GetProcessHeap(), 0, sz);
+        if (obj_name)
         {
-            module->real_path = builtin.path;
-            modfmt->module = module;
-            modfmt->remove = pe_module_remove;
-            modfmt->loc_compute = NULL;
-            module->format_info[DFI_PE] = modfmt;
-            module->reloc_delta = base - PE_FROM_OPTHDR(&modfmt->u.pe_info->fmap, ImageBase);
-        }
-        else
-        {
-            ERR("could not load the module '%s'\n", debugstr_w(loaded_name));
-            heap_free(builtin.path);
-            image_unmap_file(&modfmt->u.pe_info->fmap);
+            nts = NtQueryObject(hFile, ObjectNameInformation, obj_name, sz, &needed);
+            if (nts == STATUS_BUFFER_OVERFLOW)
+            {
+                sz = needed;
+                obj_name = RtlReAllocateHeap(GetProcessHeap(), 0, obj_name, sz);
+                nts = NtQueryObject(hFile, ObjectNameInformation, obj_name, sz, &needed);
+            }
+            if (!nts)
+            {
+                obj_name->Name.Buffer[obj_name->Name.Length / sizeof(WCHAR)] = L'\0';
+                real_path = wcsdup(obj_name->Name.Buffer);
+            }
+            RtlFreeHeap(GetProcessHeap(), 0, obj_name);
         }
+        if (name) lstrcpyW(loaded_name, name);
+    }
+
+    if ((modfmt = HeapAlloc(GetProcessHeap(), 0, sizeof(struct module_format) + sizeof(struct pe_module_info))))
+    {
+        modfmt->u.pe_info = (struct pe_module_info*)(modfmt + 1);
+        if (pe_map_file(hFile, &modfmt->u.pe_info->fmap))
+        {
+            struct builtin_search builtin = { NULL };
+            if (opened && modfmt->u.pe_info->fmap.u.pe.builtin &&
+                search_dll_path(pcs, loaded_name, modfmt->u.pe_info->fmap.u.pe.file_header.Machine, search_builtin_pe, &builtin))
+            {
+                TRACE("reloaded %s from %s\n", debugstr_w(loaded_name), debugstr_w(builtin.path));
+                image_unmap_file(&modfmt->u.pe_info->fmap);
+                modfmt->u.pe_info->fmap = builtin.fmap;
+                real_path = builtin.path;
+            }
+            if (!base) base = PE_FROM_OPTHDR(&modfmt->u.pe_info->fmap, ImageBase);
+            if (!size) size = PE_FROM_OPTHDR(&modfmt->u.pe_info->fmap, SizeOfImage);
+
+            module = module_new(pcs, loaded_name, DMT_PE, modfmt->u.pe_info->fmap.u.pe.builtin, FALSE,
+                                base, size,
+                                modfmt->u.pe_info->fmap.u.pe.file_header.TimeDateStamp,
+                                PE_FROM_OPTHDR(&modfmt->u.pe_info->fmap, CheckSum),
+                                modfmt->u.pe_info->fmap.u.pe.file_header.Machine);
+            if (module)
+            {
+                module->real_path = real_path ? pool_wcsdup(&module->pool, real_path) : NULL;
+                modfmt->module = module;
+                modfmt->remove = pe_module_remove;
+                modfmt->loc_compute = NULL;
+                module->format_info[DFI_PE] = modfmt;
+                module->reloc_delta = base - PE_FROM_OPTHDR(&modfmt->u.pe_info->fmap, ImageBase);
+            }
+            else
+            {
+                ERR("could not load the module '%s'\n", debugstr_w(loaded_name));
+                image_unmap_file(&modfmt->u.pe_info->fmap);
+            }
+        }
+        if (!module) HeapFree(GetProcessHeap(), 0, modfmt);
     }
-    if (!module) HeapFree(GetProcessHeap(), 0, modfmt);
 
     if (opened) CloseHandle(hFile);
-
+    free(real_path);
     return module;
 }
 
@@ -842,15 +900,27 @@ struct module* pe_load_native_module(struct process* pcs, const WCHAR* name,
  *		pe_load_nt_header
  *
  */
-BOOL pe_load_nt_header(HANDLE hProc, DWORD64 base, IMAGE_NT_HEADERS* nth)
+BOOL pe_load_nt_header(HANDLE hProc, DWORD64 base, IMAGE_NT_HEADERS* nth, BOOL* is_builtin)
 {
     IMAGE_DOS_HEADER    dos;
 
-    return ReadProcessMemory(hProc, (char*)(DWORD_PTR)base, &dos, sizeof(dos), NULL) &&
-        dos.e_magic == IMAGE_DOS_SIGNATURE &&
-        ReadProcessMemory(hProc, (char*)(DWORD_PTR)(base + dos.e_lfanew),
-                          nth, sizeof(*nth), NULL) &&
-        nth->Signature == IMAGE_NT_SIGNATURE;
+    if (!ReadProcessMemory(hProc, (char*)(DWORD_PTR)base, &dos, sizeof(dos), NULL) ||
+        dos.e_magic != IMAGE_DOS_SIGNATURE ||
+        !ReadProcessMemory(hProc, (char*)(DWORD_PTR)(base + dos.e_lfanew),
+                           nth, sizeof(*nth), NULL) ||
+        nth->Signature != IMAGE_NT_SIGNATURE)
+        return FALSE;
+    if (is_builtin)
+    {
+        if (dos.e_lfanew >= sizeof(dos) + sizeof(builtin_signature))
+        {
+            char sig[sizeof(builtin_signature)];
+            *is_builtin = ReadProcessMemory(hProc, (char*)(DWORD_PTR)base + sizeof(dos), sig, sizeof(sig), NULL) &&
+                !memcmp(sig, builtin_signature, sizeof(builtin_signature));
+        }
+        else *is_builtin = FALSE;
+    }
+    return TRUE;
 }
 
 /******************************************************************
@@ -865,11 +935,12 @@ struct module* pe_load_builtin_module(struct process* pcs, const WCHAR* name,
     if (base && pcs->dbg_hdr_addr)
     {
         IMAGE_NT_HEADERS    nth;
+        BOOL is_builtin;
 
-        if (pe_load_nt_header(pcs->handle, base, &nth))
+        if (pe_load_nt_header(pcs->handle, base, &nth, &is_builtin))
         {
             if (!size) size = nth.OptionalHeader.SizeOfImage;
-            module = module_new(pcs, name, DMT_PE, FALSE, base, size,
+            module = module_new(pcs, name, DMT_PE, is_builtin, FALSE, base, size,
                                 nth.FileHeader.TimeDateStamp,
                                 nth.OptionalHeader.CheckSum,
                                 nth.FileHeader.Machine);
@@ -938,3 +1009,54 @@ PVOID WINAPI ImageDirectoryEntryToData( PVOID base, BOOLEAN image, USHORT dir, P
 {
     return ImageDirectoryEntryToDataEx( base, image, dir, size, NULL );
 }
+
+DWORD pe_get_file_indexinfo(void* image, DWORD size, SYMSRV_INDEX_INFOW* info)
+{
+    const IMAGE_NT_HEADERS* nthdr;
+    const IMAGE_DEBUG_DIRECTORY* dbg;
+    ULONG dirsize;
+
+    if (!(nthdr = RtlImageNtHeader(image))) return ERROR_BAD_FORMAT;
+
+    dbg = RtlImageDirectoryEntryToData(image, FALSE, IMAGE_DIRECTORY_ENTRY_DEBUG, &dirsize);
+    if (!dbg) dirsize = 0;
+
+    /* fill in information from NT header */
+    info->timestamp = nthdr->FileHeader.TimeDateStamp;
+    info->stripped = (nthdr->FileHeader.Characteristics & IMAGE_FILE_DEBUG_STRIPPED) != 0;
+    if (nthdr->OptionalHeader.Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC)
+    {
+        const IMAGE_NT_HEADERS64* nthdr64 = (const IMAGE_NT_HEADERS64*)nthdr;
+        info->size = nthdr64->OptionalHeader.SizeOfImage;
+    }
+    else if (nthdr->OptionalHeader.Magic == IMAGE_NT_OPTIONAL_HDR32_MAGIC)
+    {
+        const IMAGE_NT_HEADERS32* nthdr32 = (const IMAGE_NT_HEADERS32*)nthdr;
+        info->size = nthdr32->OptionalHeader.SizeOfImage;
+    }
+    return msc_get_file_indexinfo(image, dbg, dirsize / sizeof(*dbg), info);
+}
+
+/* check if image contains a debug entry that contains a gcc/mingw - clang build-id information */
+BOOL pe_has_buildid_debug(struct image_file_map *fmap, GUID *guid)
+{
+    BOOL ret = FALSE;
+
+    if (fmap->modtype == DMT_PE)
+    {
+        SYMSRV_INDEX_INFOW info = {.sizeofstruct = sizeof(info)};
+        const void *image = pe_map_full(fmap, NULL);
+
+        if (image)
+        {
+            DWORD retval = pe_get_file_indexinfo((void*)image, GetFileSize(fmap->u.pe.hMap, NULL), &info);
+            if ((retval == ERROR_SUCCESS || retval == ERROR_BAD_EXE_FORMAT) && info.age && !info.pdbfile[0])
+            {
+                *guid = info.guid;
+                ret = TRUE;
+            }
+            pe_unmap_full(fmap);
+        }
+    }
+    return ret;
+}
