@@ -49,8 +49,8 @@
 #include "dsound.h"
 #include "dsound_private.h"
 
-/* default velocity of sound in the air */
-#define DEFAULT_VELOCITY 340
+/* according to the tests, native uses 360 as the speed of sound */
+#define DEFAULT_VELOCITY 360
 
 WINE_DEFAULT_DEBUG_CHANNEL(dsound3d);
 
@@ -65,7 +65,7 @@ static inline D3DVALUE ScalarProduct (const D3DVECTOR *a, const D3DVECTOR *b)
 	c = (a->x*b->x) + (a->y*b->y) + (a->z*b->z);
 	TRACE("(%f,%f,%f) * (%f,%f,%f) = %f)\n", a->x, a->y, a->z, b->x, b->y,
 	      b->z, c);
-	return c;
+	return !isnan(c) ? c : 0;
 }
 
 /* vector product (I believe it's called cross product in English */
@@ -77,6 +77,12 @@ static inline D3DVECTOR VectorProduct (const D3DVECTOR *a, const D3DVECTOR *b)
 	c.z = (a->x*b->y) - (a->y*b->x);
 	TRACE("(%f,%f,%f) x (%f,%f,%f) = (%f,%f,%f)\n", a->x, a->y, a->z, b->x, b->y,
 	      b->z, c.x, c.y, c.z);
+	if (isnan(c.x) || isnan(c.y) || isnan(c.z))
+	{
+		c.x = 0;
+		c.y = 0;
+		c.z = 0;
+	}
 	return c;
 }
 
@@ -136,6 +142,12 @@ static inline D3DVECTOR VectorBetweenTwoPoints (const D3DVECTOR *a, const D3DVEC
 	c.z = b->z - a->z;
 	TRACE("A (%f,%f,%f), B (%f,%f,%f), AB = (%f,%f,%f)\n", a->x, a->y, a->z, b->x, b->y,
 	      b->z, c.x, c.y, c.z);
+	if (isnan(c.x) || isnan(c.y) || isnan(c.z))
+	{
+		c.x = 0;
+		c.y = 0;
+		c.z = 0;
+	}
 	return c;
 }
 
@@ -144,7 +156,7 @@ static inline D3DVALUE ProjectVector (const D3DVECTOR *a, const D3DVECTOR *p)
 {
 	D3DVALUE prod, result;
 	prod = ScalarProduct(a, p);
-	result = prod/VectorMagnitude(p);
+	result = (VectorMagnitude(p) != 0) ? prod / VectorMagnitude(p) : 0;
 	TRACE("length projection of (%f,%f,%f) on (%f,%f,%f) = %f\n", a->x, a->y, a->z, p->x,
               p->y, p->z, result);
 	return result;
@@ -167,6 +179,8 @@ void DSOUND_Calc3DBuffer(IDirectSoundBufferImpl *dsb)
 	int i, num_main_speakers;
 	float a, ingain;
 	/* doppler shift related stuff */
+	D3DVECTOR vRelativeVel;
+	D3DVALUE flFreq, flRelativeVel, flLimitedVel, flVelocityFactor;
 
 	TRACE("(%p)\n",dsb);
 
@@ -180,18 +194,22 @@ void DSOUND_Calc3DBuffer(IDirectSoundBufferImpl *dsb)
 			/* we need to calculate distance between buffer and listener*/
 			vDistance = VectorBetweenTwoPoints(&dsb->device->ds3dl.vPosition, &dsb->ds3db_ds3db.vPosition);
 			flDistance = VectorMagnitude (&vDistance);
+			vRelativeVel = VectorBetweenTwoPoints(&dsb->device->ds3dl.vVelocity, &dsb->ds3db_ds3db.vVelocity);
 			break;
 		case DS3DMODE_HEADRELATIVE:
 			TRACE("Head-relative 3D processing mode\n");
 			/* distance between buffer and listener is same as buffer's position */
 			vDistance = dsb->ds3db_ds3db.vPosition;
 			flDistance = VectorMagnitude (&vDistance);
+			vRelativeVel = dsb->ds3db_ds3db.vVelocity;
 			break;
 		default:
 			TRACE("3D processing disabled\n");
 			/* this one is here only to eliminate annoying warning message */
 			dsb->volpan.lVolume = dsb->ds3db_lVolume;
 			DSOUND_RecalcVolPan (&dsb->volpan);
+			dsb->freq = dsb->ds3db_freq;
+			DSOUND_RecalcFormat(dsb);
 			return;
 	}
 	
@@ -288,37 +306,30 @@ void DSOUND_Calc3DBuffer(IDirectSoundBufferImpl *dsb)
 	}
 	TRACE("panning: Angle = %f rad, lPan = %ld\n", flAngle, dsb->volpan.lPan);
 
-	/* FIXME: Doppler Effect disabled since i have no idea which frequency to change and how to do it */
-if(0)
-{
-	D3DVALUE flFreq, flBufferVel, flListenerVel;
+	dsb->freq = dsb->ds3db_freq;
+
 	/* doppler shift*/
 	if (!VectorMagnitude(&dsb->ds3db_ds3db.vVelocity) && !VectorMagnitude(&dsb->device->ds3dl.vVelocity))
 	{
 		TRACE("doppler: Buffer and Listener don't have velocities\n");
 	}
-	else if (!(dsb->ds3db_ds3db.vVelocity.x == dsb->device->ds3dl.vVelocity.x &&
-	           dsb->ds3db_ds3db.vVelocity.y == dsb->device->ds3dl.vVelocity.y &&
-	           dsb->ds3db_ds3db.vVelocity.z == dsb->device->ds3dl.vVelocity.z))
+	else if (!(vRelativeVel.x == 0.0f && vRelativeVel.y == 0.0f && vRelativeVel.z == 0.0f) &&
+	         !(vDistance.x == 0.0f && vDistance.y == 0.0f && vDistance.z == 0.0f))
 	{
-		/* calculate length of ds3db_ds3db.vVelocity component which causes Doppler Effect
+		/* calculate length of vRelativeVel component which causes Doppler Effect
 		   NOTE: if buffer moves TOWARDS the listener, its velocity component is NEGATIVE
 		         if buffer moves AWAY from listener, its velocity component is POSITIVE */
-		flBufferVel = ProjectVector(&dsb->ds3db_ds3db.vVelocity, &vDistance);
-		/* calculate length of ds3dl.vVelocity component which causes Doppler Effect
-		   NOTE: if listener moves TOWARDS the buffer, its velocity component is POSITIVE
-		         if listener moves AWAY from buffer, its velocity component is NEGATIVE */
-		flListenerVel = ProjectVector(&dsb->device->ds3dl.vVelocity, &vDistance);
+		flRelativeVel = ProjectVector(&vRelativeVel, &vDistance);
+		flVelocityFactor = dsb->device->ds3dl.flDistanceFactor * dsb->device->ds3dl.flDopplerFactor;
+		flLimitedVel = max(-DEFAULT_VELOCITY/2, min(DEFAULT_VELOCITY/2, flRelativeVel * flVelocityFactor));
 		/* formula taken from Gianicoli D.: Physics, 4th edition: */
-		/* FIXME: replace dsb->freq with appropriate frequency ! */
-		flFreq = dsb->freq * ((DEFAULT_VELOCITY + flListenerVel)/(DEFAULT_VELOCITY + flBufferVel));
-		TRACE("doppler: Buffer velocity (component) = %f, Listener velocity (component) = %f => Doppler shift: %ld Hz -> %f Hz\n",
-		      flBufferVel, flListenerVel, dsb->freq, flFreq);
-		/* FIXME: replace following line with correct frequency setting ! */
-		dsb->freq = flFreq;
-		DSOUND_RecalcFormat(dsb);
+		flFreq = dsb->ds3db_freq * (DEFAULT_VELOCITY/(DEFAULT_VELOCITY + flLimitedVel));
+		TRACE("doppler: Relative velocity (component) = %f => Doppler shift: %ld Hz -> %f Hz\n",
+		      flRelativeVel, dsb->ds3db_freq, flFreq);
+		dsb->freq = max(DSBFREQUENCY_MIN, min(DSBFREQUENCY_MAX, (DWORD)flFreq));
 	}
-}
+
+	DSOUND_RecalcFormat(dsb);
 
 	for (i = 0; i < dsb->device->pwfx->nChannels; i++)
 		dsb->volpan.dwTotalAmpFactor[i] = 0;
@@ -363,11 +374,13 @@ static void DSOUND_ChangeListener(IDirectSoundBufferImpl *ds3dl)
 	TRACE("(%p)\n",ds3dl);
 	for (i = 0; i < ds3dl->device->nrofbuffers; i++)
 	{
+		AcquireSRWLockExclusive(&ds3dl->device->buffers[i]->lock);
 		/* check if this buffer is waiting for recalculation */
 		if (ds3dl->device->buffers[i]->ds3db_need_recalc)
 		{
 			DSOUND_Mix3DBuffer(ds3dl->device->buffers[i]);
 		}
+		ReleaseSRWLockExclusive(&ds3dl->device->buffers[i]->lock);
 	}
 }
 
@@ -545,6 +558,9 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetAllParameters(IDirectSound3DBu
 	}
 
 	TRACE("setting: all parameters; dwApply = %ld\n", dwApply);
+
+	AcquireSRWLockExclusive(&This->lock);
+
 	This->ds3db_ds3db = *lpcDs3dBuffer;
 
 	if (dwApply == DS3D_IMMEDIATE)
@@ -554,6 +570,8 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetAllParameters(IDirectSound3DBu
 	This->ds3db_need_recalc = TRUE;
 	status = DS_OK;
 
+	ReleaseSRWLockExclusive(&This->lock);
+
 	return status;
 }
 
@@ -564,11 +582,17 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetConeAngles(IDirectSound3DBuffe
 
     TRACE("setting: Inside Cone Angle = %ld; Outside Cone Angle = %ld; dwApply = %ld\n",
             dwInsideConeAngle, dwOutsideConeAngle, dwApply);
+
+    AcquireSRWLockExclusive(&This->lock);
+
     This->ds3db_ds3db.dwInsideConeAngle = dwInsideConeAngle;
     This->ds3db_ds3db.dwOutsideConeAngle = dwOutsideConeAngle;
     if (dwApply == DS3D_IMMEDIATE)
         DSOUND_Mix3DBuffer(This);
     This->ds3db_need_recalc = TRUE;
+
+    ReleaseSRWLockExclusive(&This->lock);
+
     return DS_OK;
 }
 
@@ -578,6 +602,9 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetConeOrientation(IDirectSound3D
     IDirectSoundBufferImpl *This = impl_from_IDirectSound3DBuffer(iface);
 
     TRACE("setting: Cone Orientation vector = (%f,%f,%f); dwApply = %ld\n", x, y, z, dwApply);
+
+    AcquireSRWLockExclusive(&This->lock);
+
     This->ds3db_ds3db.vConeOrientation.x = x;
     This->ds3db_ds3db.vConeOrientation.y = y;
     This->ds3db_ds3db.vConeOrientation.z = z;
@@ -587,6 +614,9 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetConeOrientation(IDirectSound3D
         DSOUND_Mix3DBuffer(This);
     }
     This->ds3db_need_recalc = TRUE;
+
+    ReleaseSRWLockExclusive(&This->lock);
+
     return DS_OK;
 }
 
@@ -596,6 +626,9 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetConeOutsideVolume(IDirectSound
     IDirectSoundBufferImpl *This = impl_from_IDirectSound3DBuffer(iface);
 
     TRACE("setting: ConeOutsideVolume = %ld; dwApply = %ld\n", lConeOutsideVolume, dwApply);
+
+    AcquireSRWLockExclusive(&This->lock);
+
     This->ds3db_ds3db.lConeOutsideVolume = lConeOutsideVolume;
     if (dwApply == DS3D_IMMEDIATE)
     {
@@ -603,6 +636,9 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetConeOutsideVolume(IDirectSound
         DSOUND_Mix3DBuffer(This);
     }
     This->ds3db_need_recalc = TRUE;
+
+    ReleaseSRWLockExclusive(&This->lock);
+
     return DS_OK;
 }
 
@@ -612,6 +648,9 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetMaxDistance(IDirectSound3DBuff
     IDirectSoundBufferImpl *This = impl_from_IDirectSound3DBuffer(iface);
 
     TRACE("setting: MaxDistance = %f; dwApply = %ld\n", fMaxDistance, dwApply);
+
+    AcquireSRWLockExclusive(&This->lock);
+
     This->ds3db_ds3db.flMaxDistance = fMaxDistance;
     if (dwApply == DS3D_IMMEDIATE)
     {
@@ -619,6 +658,9 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetMaxDistance(IDirectSound3DBuff
         DSOUND_Mix3DBuffer(This);
     }
     This->ds3db_need_recalc = TRUE;
+
+    ReleaseSRWLockExclusive(&This->lock);
+
     return DS_OK;
 }
 
@@ -628,6 +670,9 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetMinDistance(IDirectSound3DBuff
     IDirectSoundBufferImpl *This = impl_from_IDirectSound3DBuffer(iface);
 
     TRACE("setting: MinDistance = %f; dwApply = %ld\n", fMinDistance, dwApply);
+
+    AcquireSRWLockExclusive(&This->lock);
+
     This->ds3db_ds3db.flMinDistance = fMinDistance;
     if (dwApply == DS3D_IMMEDIATE)
     {
@@ -635,6 +680,9 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetMinDistance(IDirectSound3DBuff
         DSOUND_Mix3DBuffer(This);
     }
     This->ds3db_need_recalc = TRUE;
+
+    ReleaseSRWLockExclusive(&This->lock);
+
     return DS_OK;
 }
 
@@ -644,6 +692,9 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetMode(IDirectSound3DBuffer *ifa
     IDirectSoundBufferImpl *This = impl_from_IDirectSound3DBuffer(iface);
 
     TRACE("setting: Mode = %ld; dwApply = %ld\n", dwMode, dwApply);
+
+    AcquireSRWLockExclusive(&This->lock);
+
     This->ds3db_ds3db.dwMode = dwMode;
     if (dwApply == DS3D_IMMEDIATE)
     {
@@ -651,6 +702,9 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetMode(IDirectSound3DBuffer *ifa
         DSOUND_Mix3DBuffer(This);
     }
     This->ds3db_need_recalc = TRUE;
+
+    ReleaseSRWLockExclusive(&This->lock);
+
     return DS_OK;
 }
 
@@ -660,6 +714,9 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetPosition(IDirectSound3DBuffer
     IDirectSoundBufferImpl *This = impl_from_IDirectSound3DBuffer(iface);
 
     TRACE("setting: Position vector = (%f,%f,%f); dwApply = %ld\n", x, y, z, dwApply);
+
+    AcquireSRWLockExclusive(&This->lock);
+
     This->ds3db_ds3db.vPosition.x = x;
     This->ds3db_ds3db.vPosition.y = y;
     This->ds3db_ds3db.vPosition.z = z;
@@ -669,6 +726,9 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetPosition(IDirectSound3DBuffer
         DSOUND_Mix3DBuffer(This);
     }
     This->ds3db_need_recalc = TRUE;
+
+    ReleaseSRWLockExclusive(&This->lock);
+
     return DS_OK;
 }
 
@@ -678,6 +738,9 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetVelocity(IDirectSound3DBuffer
     IDirectSoundBufferImpl *This = impl_from_IDirectSound3DBuffer(iface);
 
     TRACE("setting: Velocity vector = (%f,%f,%f); dwApply = %ld\n", x, y, z, dwApply);
+
+    AcquireSRWLockExclusive(&This->lock);
+
     This->ds3db_ds3db.vVelocity.x = x;
     This->ds3db_ds3db.vVelocity.y = y;
     This->ds3db_ds3db.vVelocity.z = z;
@@ -687,6 +750,9 @@ static HRESULT WINAPI IDirectSound3DBufferImpl_SetVelocity(IDirectSound3DBuffer
         DSOUND_Mix3DBuffer(This);
     }
     This->ds3db_need_recalc = TRUE;
+
+    ReleaseSRWLockExclusive(&This->lock);
+
     return DS_OK;
 }
 
@@ -903,16 +969,22 @@ static HRESULT WINAPI IDirectSound3DListenerImpl_SetOrientation(IDirectSound3DLi
         D3DVALUE xFront, D3DVALUE yFront, D3DVALUE zFront, D3DVALUE xTop, D3DVALUE yTop,
         D3DVALUE zTop, DWORD dwApply)
 {
-        IDirectSoundBufferImpl *This = impl_from_IDirectSound3DListener(iface);
+	IDirectSoundBufferImpl *This = impl_from_IDirectSound3DListener(iface);
+	D3DVECTOR front = {.x = xFront, .y = yFront, .z = zFront};
+	D3DVECTOR top = {.x = xTop, .y = yTop, .z = zTop};
+	D3DVALUE angle = 0;
 
 	TRACE("setting: Front vector = (%f,%f,%f); Top vector = (%f,%f,%f); dwApply = %ld\n",
 	xFront, yFront, zFront, xTop, yTop, zTop, dwApply);
-	This->device->ds3dl.vOrientFront.x = xFront;
-	This->device->ds3dl.vOrientFront.y = yFront;
-	This->device->ds3dl.vOrientFront.z = zFront;
-	This->device->ds3dl.vOrientTop.x = xTop;
-	This->device->ds3dl.vOrientTop.y = yTop;
-	This->device->ds3dl.vOrientTop.z = zTop;
+
+	if ((angle = AngleBetweenVectorsDeg(&front, &top)) == 0.0f)
+	{
+		WARN("Angle %f is 0 degrees\n", angle);
+		return DSERR_INVALIDPARAM;
+	}
+
+	This->device->ds3dl.vOrientFront = front;
+	This->device->ds3dl.vOrientTop = top;
 	if (dwApply == DS3D_IMMEDIATE)
 	{
 		This->device->ds3dl_need_recalc = FALSE;
