@@ -39,7 +39,6 @@ enum { MAX_QUOTA = 5000000 };
 typedef struct {
     DispatchEx dispex;
     IHTMLStorage IHTMLStorage_iface;
-    LONG ref;
     unsigned num_props;
     BSTR *props;
     HTMLInnerWindow *window;
@@ -200,24 +199,24 @@ static inline HTMLStorage *impl_from_IHTMLStorage(IHTMLStorage *iface)
 static HRESULT build_session_origin(IUri*,BSTR,BSTR*);
 
 struct storage_event_task {
-    task_t header;
-    HTMLInnerWindow *window;
+    event_task_t header;
     DOMEvent *event;
 };
 
-static void storage_event_proc(task_t *_task)
+static void storage_event_proc(event_task_t *_task)
 {
     struct storage_event_task *task = (struct storage_event_task*)_task;
-    HTMLInnerWindow *window = task->window;
+    HTMLInnerWindow *window = task->header.window;
     DOMEvent *event = task->event;
+    compat_mode_t compat_mode;
     VARIANT_BOOL cancelled;
     HRESULT hres;
     VARIANT var;
 
-    if(event->event_id == EVENTID_STORAGE && dispex_compat_mode(&window->event_target.dispex) >= COMPAT_MODE_IE9) {
+    if(event->event_id == EVENTID_STORAGE && (compat_mode = dispex_compat_mode(&window->event_target.dispex)) >= COMPAT_MODE_IE9) {
         dispatch_event(&window->event_target, event);
         if(window->doc) {
-            hres = create_event_obj(window->doc, event, (IHTMLEventObj**)&V_DISPATCH(&var));
+            hres = create_event_obj(event, window->doc, (IHTMLEventObj**)&V_DISPATCH(&var));
             if(SUCCEEDED(hres)) {
                 V_VT(&var) = VT_DISPATCH;
                 fire_event(&window->doc->node, L"onstorage", &var, &cancelled);
@@ -229,11 +228,10 @@ static void storage_event_proc(task_t *_task)
     }
 }
 
-static void storage_event_destr(task_t *_task)
+static void storage_event_destr(event_task_t *_task)
 {
     struct storage_event_task *task = (struct storage_event_task*)_task;
     IDOMEvent_Release(&task->event->IDOMEvent_iface);
-    IHTMLWindow2_Release(&task->window->base.IHTMLWindow2_iface);
 }
 
 struct send_storage_event_ctx {
@@ -262,9 +260,7 @@ static HRESULT push_storage_event_task(struct send_storage_event_ctx *ctx, HTMLI
     }
 
     task->event = event;
-    task->window = window;
-    IHTMLWindow2_AddRef(&task->window->base.IHTMLWindow2_iface);
-    return push_task(&task->header, storage_event_proc, storage_event_destr, window->task_magic);
+    return push_event_task(&task->header, window, storage_event_proc, storage_event_destr, window->task_magic);
 }
 
 static HRESULT send_storage_event_impl(struct send_storage_event_ctx *ctx, HTMLInnerWindow *window)
@@ -294,6 +290,8 @@ static HRESULT send_storage_event_impl(struct send_storage_event_ctx *ctx, HTMLI
         origin = storage->session_storage->origin;
         origin_len = ctx->skip_window ? wcslen(origin) : storage->session_storage->origin_len;
         bstr = NULL;
+    }else if(!window->base.outer_window->uri) {
+        return S_OK;
     }else {
         hres = IUri_GetHost(window->base.outer_window->uri, &bstr);
         if(hres != S_OK) {
@@ -333,8 +331,10 @@ static HRESULT send_storage_event(HTMLStorage *storage, BSTR key, BSTR old_value
     HRESULT hres = S_OK;
 
     ctx.url = NULL;
-    if(!window)
+
+    if(!window->base.outer_window)
         goto done;
+
     if(window->base.outer_window->uri_nofrag) {
         hres = IUri_GetDisplayUri(window->base.outer_window->uri_nofrag, &ctx.url);
         if(hres != S_OK)
@@ -370,88 +370,7 @@ done:
     return FAILED(hres) ? hres : S_OK;
 }
 
-static HRESULT WINAPI HTMLStorage_QueryInterface(IHTMLStorage *iface, REFIID riid, void **ppv)
-{
-    HTMLStorage *This = impl_from_IHTMLStorage(iface);
-
-    TRACE("(%p)->(%s %p)\n", This, debugstr_mshtml_guid(riid), ppv);
-
-    if(IsEqualGUID(&IID_IUnknown, riid)) {
-        *ppv = &This->IHTMLStorage_iface;
-    }else if(IsEqualGUID(&IID_IHTMLStorage, riid)) {
-        *ppv = &This->IHTMLStorage_iface;
-    }else if(dispex_query_interface(&This->dispex, riid, ppv)) {
-        return *ppv ? S_OK : E_NOINTERFACE;
-    }else {
-        *ppv = NULL;
-        WARN("(%p)->(%s %p)\n", This, debugstr_mshtml_guid(riid), ppv);
-        return E_NOINTERFACE;
-    }
-
-    IUnknown_AddRef((IUnknown*)*ppv);
-    return S_OK;
-}
-
-static ULONG WINAPI HTMLStorage_AddRef(IHTMLStorage *iface)
-{
-    HTMLStorage *This = impl_from_IHTMLStorage(iface);
-    LONG ref = InterlockedIncrement(&This->ref);
-
-    TRACE("(%p) ref=%ld\n", This, ref);
-
-    return ref;
-}
-
-static ULONG WINAPI HTMLStorage_Release(IHTMLStorage *iface)
-{
-    HTMLStorage *This = impl_from_IHTMLStorage(iface);
-    LONG ref = InterlockedDecrement(&This->ref);
-
-    TRACE("(%p) ref=%ld\n", This, ref);
-
-    if(!ref) {
-        release_session_map_entry(This->session_storage);
-        release_dispex(&This->dispex);
-        free(This->filename);
-        CloseHandle(This->mutex);
-        release_props(This);
-        free(This);
-    }
-
-    return ref;
-}
-
-static HRESULT WINAPI HTMLStorage_GetTypeInfoCount(IHTMLStorage *iface, UINT *pctinfo)
-{
-    HTMLStorage *This = impl_from_IHTMLStorage(iface);
-    return IDispatchEx_GetTypeInfoCount(&This->dispex.IDispatchEx_iface, pctinfo);
-}
-
-static HRESULT WINAPI HTMLStorage_GetTypeInfo(IHTMLStorage *iface, UINT iTInfo,
-        LCID lcid, ITypeInfo **ppTInfo)
-{
-    HTMLStorage *This = impl_from_IHTMLStorage(iface);
-
-    return IDispatchEx_GetTypeInfo(&This->dispex.IDispatchEx_iface, iTInfo, lcid, ppTInfo);
-}
-
-static HRESULT WINAPI HTMLStorage_GetIDsOfNames(IHTMLStorage *iface, REFIID riid, LPOLESTR *rgszNames, UINT cNames,
-        LCID lcid, DISPID *rgDispId)
-{
-    HTMLStorage *This = impl_from_IHTMLStorage(iface);
-
-    return IDispatchEx_GetIDsOfNames(&This->dispex.IDispatchEx_iface, riid, rgszNames, cNames,
-            lcid, rgDispId);
-}
-
-static HRESULT WINAPI HTMLStorage_Invoke(IHTMLStorage *iface, DISPID dispIdMember, REFIID riid, LCID lcid,
-        WORD wFlags, DISPPARAMS *pDispParams, VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT *puArgErr)
-{
-    HTMLStorage *This = impl_from_IHTMLStorage(iface);
-
-    return IDispatchEx_Invoke(&This->dispex.IDispatchEx_iface, dispIdMember, riid, lcid, wFlags,
-            pDispParams, pVarResult, pExcepInfo, puArgErr);
-}
+DISPEX_IDISPATCH_IMPL(HTMLStorage, IHTMLStorage, impl_from_IHTMLStorage(iface)->dispex)
 
 static BOOL create_path(const WCHAR *path)
 {
@@ -1075,6 +994,45 @@ static inline HTMLStorage *impl_from_DispatchEx(DispatchEx *iface)
     return CONTAINING_RECORD(iface, HTMLStorage, dispex);
 }
 
+static void *HTMLStorage_query_interface(DispatchEx *dispex, REFIID riid)
+{
+    HTMLStorage *This = impl_from_DispatchEx(dispex);
+
+    if(IsEqualGUID(&IID_IHTMLStorage, riid))
+        return &This->IHTMLStorage_iface;
+
+    return NULL;
+}
+
+static void HTMLStorage_traverse(DispatchEx *dispex, nsCycleCollectionTraversalCallback *cb)
+{
+    HTMLStorage *This = impl_from_DispatchEx(dispex);
+
+    if(This->window)
+        note_cc_edge((nsISupports*)&This->window->base.IHTMLWindow2_iface, "window", cb);
+}
+
+static void HTMLStorage_unlink(DispatchEx *dispex)
+{
+    HTMLStorage *This = impl_from_DispatchEx(dispex);
+
+    if(This->window) {
+        HTMLInnerWindow *window = This->window;
+        This->window = NULL;
+        IHTMLWindow2_Release(&window->base.IHTMLWindow2_iface);
+    }
+}
+
+static void HTMLStorage_destructor(DispatchEx *dispex)
+{
+    HTMLStorage *This = impl_from_DispatchEx(dispex);
+    release_session_map_entry(This->session_storage);
+    free(This->filename);
+    CloseHandle(This->mutex);
+    release_props(This);
+    free(This);
+}
+
 static HRESULT check_item(HTMLStorage *This, const WCHAR *key)
 {
     struct session_entry *session_entry;
@@ -1141,7 +1099,7 @@ static HRESULT get_prop(HTMLStorage *This, const WCHAR *name, DISPID *dispid)
     return S_OK;
 }
 
-static HRESULT HTMLStorage_get_dispid(DispatchEx *dispex, BSTR name, DWORD flags, DISPID *dispid)
+static HRESULT HTMLStorage_get_dispid(DispatchEx *dispex, const WCHAR *name, DWORD flags, DISPID *dispid)
 {
     HTMLStorage *This = impl_from_DispatchEx(dispex);
     HRESULT hres;
@@ -1158,19 +1116,8 @@ static HRESULT HTMLStorage_get_dispid(DispatchEx *dispex, BSTR name, DWORD flags
     return get_prop(This, name, dispid);
 }
 
-static HRESULT HTMLStorage_get_name(DispatchEx *dispex, DISPID id, BSTR *name)
-{
-    HTMLStorage *This = impl_from_DispatchEx(dispex);
-    DWORD idx = id - MSHTML_DISPID_CUSTOM_MIN;
-
-    if(idx >= This->num_props)
-        return DISP_E_MEMBERNOTFOUND;
-
-    return (*name = SysAllocString(This->props[idx])) ? S_OK : E_OUTOFMEMORY;
-}
-
-static HRESULT HTMLStorage_invoke(DispatchEx *dispex, IDispatch *this_obj, DISPID id, LCID lcid, WORD flags,
-        DISPPARAMS *params, VARIANT *res, EXCEPINFO *ei, IServiceProvider *caller)
+static HRESULT HTMLStorage_invoke(DispatchEx *dispex, DISPID id, LCID lcid, WORD flags, DISPPARAMS *params,
+        VARIANT *res, EXCEPINFO *ei, IServiceProvider *caller)
 {
     HTMLStorage *This = impl_from_DispatchEx(dispex);
     DWORD idx = id - MSHTML_DISPID_CUSTOM_MIN;
@@ -1232,7 +1179,7 @@ static HRESULT HTMLStorage_delete(DispatchEx *dispex, DISPID id)
     return HTMLStorage_removeItem(&This->IHTMLStorage_iface, This->props[idx]);
 }
 
-static HRESULT HTMLStorage_next_dispid(DispatchEx *dispex, DISPID id, DISPID *pid)
+static HRESULT HTMLStorage_next_dispid(DispatchEx *dispex, DISPID id, BOOL enum_all_own_props, DISPID *pid)
 {
     DWORD idx = (id == DISPID_STARTENUM) ? 0 : id - MSHTML_DISPID_CUSTOM_MIN + 1;
     HTMLStorage *This = impl_from_DispatchEx(dispex);
@@ -1319,7 +1266,18 @@ static HRESULT HTMLStorage_next_dispid(DispatchEx *dispex, DISPID id, DISPID *pi
     return S_OK;
 }
 
-static HRESULT HTMLStorage_override(DispatchEx *dispex, const WCHAR *name, VARIANT *value)
+static HRESULT HTMLStorage_get_prop_desc(DispatchEx *dispex, DISPID id, struct property_info *desc)
+{
+    HTMLStorage *This = impl_from_DispatchEx(dispex);
+
+    desc->name = This->props[id - MSHTML_DISPID_CUSTOM_MIN];
+    desc->id = id;
+    desc->flags = PROPF_WRITABLE | PROPF_CONFIGURABLE | PROPF_ENUMERABLE;
+    desc->prototype_id = 0;
+    return S_OK;
+}
+
+static HRESULT HTMLStorage_override(DispatchEx *dispex, const WCHAR *name, VARIANT *get_value)
 {
     HTMLStorage *This = impl_from_DispatchEx(dispex);
     BSTR bstrName;
@@ -1328,10 +1286,10 @@ static HRESULT HTMLStorage_override(DispatchEx *dispex, const WCHAR *name, VARIA
     if(!(bstrName = SysAllocString(name)))
         return E_OUTOFMEMORY;
 
-    if(value) {
-        hres = HTMLStorage_getItem(&This->IHTMLStorage_iface, bstrName, value);
+    if(get_value) {
+        hres = HTMLStorage_getItem(&This->IHTMLStorage_iface, bstrName, get_value);
         if(SUCCEEDED(hres))
-            hres = (V_VT(value) != VT_NULL) ? S_OK : S_FALSE;
+            hres = (V_VT(get_value) != VT_NULL) ? S_OK : S_FALSE;
     }else {
         hres = HTMLStorage_removeItem(&This->IHTMLStorage_iface, bstrName);
         if(SUCCEEDED(hres))
@@ -1342,27 +1300,35 @@ static HRESULT HTMLStorage_override(DispatchEx *dispex, const WCHAR *name, VARIA
     return hres;
 }
 
-static const dispex_static_data_vtbl_t HTMLStorage_dispex_vtbl = {
-    NULL,
-    HTMLStorage_get_dispid,
-    HTMLStorage_get_name,
-    HTMLStorage_invoke,
-    HTMLStorage_delete,
-    HTMLStorage_next_dispid,
-    HTMLStorage_override,
-    NULL
+static HRESULT HTMLStorage_fill_props(DispatchEx *dispex)
+{
+    return S_FALSE;
+}
+
+static const dispex_static_data_vtbl_t Storage_dispex_vtbl = {
+    .query_interface  = HTMLStorage_query_interface,
+    .destructor       = HTMLStorage_destructor,
+    .traverse         = HTMLStorage_traverse,
+    .unlink           = HTMLStorage_unlink,
+    .get_dispid       = HTMLStorage_get_dispid,
+    .invoke           = HTMLStorage_invoke,
+    .delete           = HTMLStorage_delete,
+    .next_dispid      = HTMLStorage_next_dispid,
+    .get_prop_desc    = HTMLStorage_get_prop_desc,
+    .override         = HTMLStorage_override,
+    .fill_props       = HTMLStorage_fill_props,
 };
 
 static const tid_t HTMLStorage_iface_tids[] = {
     IHTMLStorage_tid,
     0
 };
-dispex_static_data_t HTMLStorage_dispex = {
-    L"Storage",
-    &HTMLStorage_dispex_vtbl,
-    PROTO_ID_HTMLStorage,
-    IHTMLStorage_tid,
-    HTMLStorage_iface_tids
+dispex_static_data_t Storage_dispex = {
+    .id         = PROT_Storage,
+    .vtbl       = &Storage_dispex_vtbl,
+    .disp_tid   = IHTMLStorage_tid,
+    .iface_tids = HTMLStorage_iface_tids,
+    .js_flags   = HOSTOBJ_VOLATILE_PROPS
 };
 
 static HRESULT build_session_origin(IUri *uri, BSTR hostname, BSTR *ret)
@@ -1512,18 +1478,12 @@ HRESULT create_html_storage(HTMLInnerWindow *window, BOOL local, IHTMLStorage **
     }
 
     storage->IHTMLStorage_iface.lpVtbl = &HTMLStorageVtbl;
-    storage->ref = 1;
     storage->window = window;
+    IHTMLWindow2_AddRef(&window->base.IHTMLWindow2_iface);
 
-    init_dispatch(&storage->dispex, (IUnknown*)&storage->IHTMLStorage_iface, &HTMLStorage_dispex,
-                  window, dispex_compat_mode(&window->event_target.dispex));
+    init_dispatch(&storage->dispex, &Storage_dispex, window,
+                  dispex_compat_mode(&window->event_target.dispex));
 
     *p = &storage->IHTMLStorage_iface;
     return S_OK;
 }
-
-void detach_html_storage(IHTMLStorage *iface)
-{
-    HTMLStorage *storage = impl_from_IHTMLStorage(iface);
-    storage->window = NULL;
-}
