@@ -23,7 +23,6 @@
 #include <limits.h>
 #include <sys/types.h>
 
-#define NONAMELESSUNION
 #include "ntstatus.h"
 #define WIN32_NO_STATUS
 #include "winternl.h"
@@ -94,24 +93,30 @@ static int append_output( struct debug_info *info, const char *str, size_t len )
 unsigned char __cdecl __wine_dbg_get_channel_flags( struct __wine_debug_channel *channel )
 {
     int min, max, pos, res;
-    unsigned char default_flags;
+    unsigned char flags;
+
+    if (!(channel->flags & (1 << __WINE_DBCL_INIT))) return channel->flags;
 
     if (!debug_options) init_options();
 
+    flags = debug_options[nb_debug_options].flags;
     min = 0;
     max = nb_debug_options - 1;
     while (min <= max)
     {
         pos = (min + max) / 2;
         res = strcmp( channel->name, debug_options[pos].name );
-        if (!res) return debug_options[pos].flags;
+        if (!res)
+        {
+            flags = debug_options[pos].flags;
+            break;
+        }
         if (res < 0) max = pos - 1;
         else min = pos + 1;
     }
-    /* no option for this channel */
-    default_flags = debug_options[nb_debug_options].flags;
-    if (channel->flags & (1 << __WINE_DBCL_INIT)) channel->flags = default_flags;
-    return default_flags;
+
+    if (!(flags & (1 << __WINE_DBCL_INIT))) channel->flags = flags; /* not dynamically changeable */
+    return flags;
 }
 
 /***********************************************************************
@@ -165,6 +170,16 @@ int __cdecl __wine_dbg_header( enum __wine_debug_class cls, struct __wine_debug_
     return info->out_pos;
 }
 
+/***********************************************************************
+ *		__wine_dbg_write  (NTDLL.@)
+ */
+int WINAPI __wine_dbg_write( const char *str, unsigned int len )
+{
+    struct wine_dbg_write_params params = { str, len };
+
+    return WINE_UNIX_CALL( unix_wine_dbg_write, &params );
+}
+
 /***********************************************************************
  *		__wine_dbg_output  (NTDLL.@)
  */
@@ -245,98 +260,6 @@ void WINAPI RtlExitUserThread( ULONG status )
 }
 
 
-/***********************************************************************
- *           RtlUserThreadStart (NTDLL.@)
- */
-#ifdef __i386__
-__ASM_STDCALL_FUNC( RtlUserThreadStart, 8,
-                   "movl %ebx,8(%esp)\n\t"  /* arg */
-                   "movl %eax,4(%esp)\n\t"  /* entry */
-                   "jmp " __ASM_NAME("call_thread_func") )
-
-/* wrapper to call BaseThreadInitThunk */
-extern void DECLSPEC_NORETURN call_thread_func_wrapper( void *thunk, PRTL_THREAD_START_ROUTINE entry, void *arg );
-__ASM_GLOBAL_FUNC( call_thread_func_wrapper,
-                  "pushl %ebp\n\t"
-                  __ASM_CFI(".cfi_adjust_cfa_offset 4\n\t")
-                  __ASM_CFI(".cfi_rel_offset %ebp,0\n\t")
-                  "movl %esp,%ebp\n\t"
-                  __ASM_CFI(".cfi_def_cfa_register %ebp\n\t")
-                   "subl $4,%esp\n\t"
-                   "andl $~0xf,%esp\n\t"
-                   "xorl %ecx,%ecx\n\t"
-                   "movl 12(%ebp),%edx\n\t"
-                   "movl 16(%ebp),%eax\n\t"
-                   "movl %eax,(%esp)\n\t"
-                   "call *8(%ebp)" )
-
-void DECLSPEC_HIDDEN call_thread_func( PRTL_THREAD_START_ROUTINE entry, void *arg )
-{
-    __TRY
-    {
-        TRACE_(relay)( "\1Starting thread proc %p (arg=%p)\n", entry, arg );
-        call_thread_func_wrapper( pBaseThreadInitThunk, entry, arg );
-    }
-    __EXCEPT(call_unhandled_exception_filter)
-    {
-        NtTerminateProcess( GetCurrentProcess(), GetExceptionCode() );
-    }
-    __ENDTRY
-}
-
-#elif /* __i386__ */ defined(__x86_64__) && defined(__ASM_SEH_SUPPORTED)
-EXCEPTION_DISPOSITION WINAPI call_thread_func_handler( EXCEPTION_RECORD *rec, ULONG64 frame,
-                                                       CONTEXT *context, DISPATCHER_CONTEXT *dispatch )
-{
-    EXCEPTION_POINTERS ep = { rec, context };
-
-    WARN( "Unhandled exception, calling filter.\n" );
-
-    switch (call_unhandled_exception_filter( &ep ))
-    {
-        case EXCEPTION_CONTINUE_SEARCH:
-            return ExceptionContinueSearch;
-        case EXCEPTION_CONTINUE_EXECUTION:
-            return ExceptionContinueExecution;
-        case EXCEPTION_EXECUTE_HANDLER:
-            break;
-    }
-    NtTerminateProcess( GetCurrentProcess(), rec->ExceptionCode );
-    return ExceptionContinueExecution;
-}
-
-extern void WINAPI RtlUserThreadStart( PRTL_THREAD_START_ROUTINE entry, void *arg );
-__ASM_GLOBAL_FUNC( RtlUserThreadStart,
-                  "subq $0x28, %rsp\n\t"
-                  __ASM_SEH(".seh_stackalloc 0x28\n\t")
-                  __ASM_SEH(".seh_endprologue\n\t")
-                  "movq %rdx,%r8\n\t"
-                  "movq %rcx,%rdx\n\t"
-                  "xorq %rcx,%rcx\n\t"
-                  "movq pBaseThreadInitThunk(%rip),%r9\n\t"
-                  "call *%r9\n\t"
-                  "int3\n\t"
-                   __ASM_SEH(".seh_handler call_thread_func_handler, @except\n\t") )
-
-#else /* defined(__x86_64__) && defined(__ASM_SEH_SUPPORTED) */
-
-void WINAPI RtlUserThreadStart( PRTL_THREAD_START_ROUTINE entry, void *arg )
-{
-    __TRY
-    {
-        TRACE_(relay)( "\1Starting thread proc %p (arg=%p)\n", entry, arg );
-        pBaseThreadInitThunk( 0, (LPTHREAD_START_ROUTINE)entry, arg );
-    }
-    __EXCEPT(call_unhandled_exception_filter)
-    {
-        NtTerminateProcess( GetCurrentProcess(), GetExceptionCode() );
-    }
-    __ENDTRY
-}
-
-#endif  /* __i386__ */
-
-
 /***********************************************************************
  *              RtlCreateUserThread   (NTDLL.@)
  */
@@ -425,10 +348,9 @@ NTSTATUS WINAPI RtlCreateUserStack( SIZE_T commit, SIZE_T reserve, ULONG zero_bi
                                       &alloc, sizeof(alloc) );
     if (!status)
     {
-        void *addr = alloc.StackBase;
+        void *addr;
         SIZE_T size = page_size;
 
-        NtAllocateVirtualMemory( GetCurrentProcess(), &addr, 0, &size, MEM_COMMIT, PAGE_NOACCESS );
         addr = (char *)alloc.StackBase + page_size;
         NtAllocateVirtualMemory( GetCurrentProcess(), &addr, 0, &size, MEM_COMMIT, PAGE_READWRITE | PAGE_GUARD );
         addr = (char *)alloc.StackBase + 2 * page_size;
@@ -496,6 +418,49 @@ TEB_ACTIVE_FRAME * WINAPI RtlGetFrame(void)
 }
 
 
+/******************************************************************************
+ *              RtlIsCurrentThread  (NTDLL.@)
+ */
+BOOLEAN WINAPI RtlIsCurrentThread( HANDLE handle )
+{
+    return handle == NtCurrentThread() || !NtCompareObjects( handle, NtCurrentThread() );
+}
+
+
+/***********************************************************************
+ *              RtlSetThreadErrorMode  (NTDLL.@)
+ */
+NTSTATUS WINAPI RtlSetThreadErrorMode( DWORD mode, LPDWORD oldmode )
+{
+    if (mode & ~0x70)
+        return STATUS_INVALID_PARAMETER_1;
+
+    if (oldmode)
+        *oldmode = NtCurrentTeb()->HardErrorMode;
+
+    NtCurrentTeb()->HardErrorMode = mode;
+    return STATUS_SUCCESS;
+}
+
+
+/***********************************************************************
+ *              RtlGetThreadErrorMode  (NTDLL.@)
+ */
+DWORD WINAPI RtlGetThreadErrorMode( void )
+{
+    return NtCurrentTeb()->HardErrorMode;
+}
+
+
+/***********************************************************************
+ *           _errno  (NTDLL.@)
+ */
+int * CDECL _errno(void)
+{
+    return (int *)&NtCurrentTeb()->TlsSlots[NTDLL_TLS_ERRNO];
+}
+
+
 /***********************************************************************
  * Fibers
  ***********************************************************************/
@@ -567,10 +532,8 @@ NTSTATUS WINAPI DECLSPEC_HOTPATCH RtlFlsAlloc( PFLS_CALLBACK_FUNCTION callback,
 {
     unsigned int chunk_index, index, i;
     FLS_INFO_CHUNK *chunk;
-    TEB_FLS_DATA *fls;
 
-    if (!(fls = NtCurrentTeb()->FlsSlots)
-            && !(NtCurrentTeb()->FlsSlots = fls = fls_alloc_data()))
+    if (!NtCurrentTeb()->FlsSlots && !(NtCurrentTeb()->FlsSlots = fls_alloc_data()))
         return STATUS_NO_MEMORY;
 
     lock_fls_data();
