@@ -43,6 +43,15 @@
 
 WINE_DEFAULT_DEBUG_CHANNEL(opengl);
 
+static const BOOL is_win64 = (sizeof(void *) > sizeof(int));
+
+static BOOL is_wow64(void)
+{
+    return !!NtCurrentTeb()->WowTebOffset;
+}
+
+static UINT64 call_gl_debug_message_callback;
+
 static pthread_mutex_t wgl_lock = PTHREAD_MUTEX_INITIALIZER;
 
 /* handle management */
@@ -59,17 +68,18 @@ enum wgl_handle_type
 struct opengl_context
 {
     DWORD tid;                   /* thread that the context is current in */
-    void (CALLBACK *debug_callback)( GLenum, GLenum, GLuint, GLenum, GLsizei, const GLchar *, const void * ); /* debug callback */
-    const void *debug_user;      /* debug user parameter */
+    UINT64 debug_callback;       /* client pointer */
+    UINT64 debug_user;           /* client pointer */
     GLubyte *extensions;         /* extension string */
     GLuint *disabled_exts;       /* indices of disabled extensions */
     struct wgl_context *drv_ctx; /* driver context */
+    GLubyte *version_string;
 };
 
 struct wgl_handle
 {
     UINT handle;
-    struct opengl_funcs *funcs;
+    const struct opengl_funcs *funcs;
     union
     {
         struct opengl_context *context; /* for HANDLE_CONTEXT */
@@ -109,7 +119,7 @@ static struct wgl_handle *get_handle_ptr( HANDLE handle, enum wgl_handle_type ty
     return NULL;
 }
 
-static HANDLE alloc_handle( enum wgl_handle_type type, struct opengl_funcs *funcs, void *user_ptr )
+static HANDLE alloc_handle( enum wgl_handle_type type, const struct opengl_funcs *funcs, void *user_ptr )
 {
     HANDLE handle = 0;
     struct wgl_handle *ptr = NULL;
@@ -168,7 +178,13 @@ static GLubyte *filter_extensions_list( const char *extensions, const char *disa
         memcpy( p, extensions, end - extensions );
         p[end - extensions] = 0;
 
-        if (!has_extension( disabled, p, strlen( p ) ))
+        /* We do not support GL_MAP_PERSISTENT_BIT, and hence
+         * ARB_buffer_storage, on wow64. */
+        if (is_win64 && is_wow64() && (!strcmp( p, "GL_ARB_buffer_storage" ) || !strcmp( p, "GL_EXT_buffer_storage" )))
+        {
+            TRACE( "-- %s (disabled due to wow64)\n", p );
+        }
+        else if (!has_extension( disabled, p, strlen( p ) ))
         {
             TRACE( "++ %s\n", p );
             p += end - extensions;
@@ -184,13 +200,32 @@ static GLubyte *filter_extensions_list( const char *extensions, const char *disa
     return (GLubyte *)str;
 }
 
+static const char *parse_gl_version( const char *gl_version, int *major, int *minor )
+{
+    const char *ptr = gl_version;
+
+    *major = atoi( ptr );
+    if (*major <= 0)
+        ERR( "Invalid OpenGL major version %d.\n", *major );
+
+    while (isdigit( *ptr )) ++ptr;
+    if (*ptr++ != '.')
+        ERR( "Invalid OpenGL version string %s.\n", debugstr_a(gl_version) );
+
+    *minor = atoi( ptr );
+
+    while (isdigit( *ptr )) ++ptr;
+    return ptr;
+}
+
 static GLuint *filter_extensions_index( TEB *teb, const char *disabled )
 {
     const struct opengl_funcs *funcs = teb->glTable;
+    const char *ext, *version;
     GLuint *disabled_index;
     GLint extensions_count;
     unsigned int i = 0, j;
-    const char *ext;
+    int major, minor;
 
     if (!funcs->ext.p_glGetStringi)
     {
@@ -199,6 +234,11 @@ static GLuint *filter_extensions_index( TEB *teb, const char *disabled )
         if (!funcs->ext.p_glGetStringi) return NULL;
     }
 
+    version = (const char *)funcs->gl.p_glGetString( GL_VERSION );
+    parse_gl_version( version, &major, &minor );
+    if (major < 3)
+        return NULL;
+
     funcs->gl.p_glGetIntegerv( GL_NUM_EXTENSIONS, &extensions_count );
     disabled_index = malloc( extensions_count * sizeof(*disabled_index) );
     if (!disabled_index) return NULL;
@@ -208,7 +248,15 @@ static GLuint *filter_extensions_index( TEB *teb, const char *disabled )
     for (j = 0; j < extensions_count; ++j)
     {
         ext = (const char *)funcs->ext.p_glGetStringi( GL_EXTENSIONS, j );
-        if (!has_extension( disabled, ext, strlen( ext ) ))
+
+        /* We do not support GL_MAP_PERSISTENT_BIT, and hence
+         * ARB_buffer_storage, on wow64. */
+        if (is_win64 && is_wow64() && (!strcmp( ext, "GL_ARB_buffer_storage" ) || !strcmp( ext, "GL_EXT_buffer_storage" )))
+        {
+            TRACE( "-- %s (disabled due to wow64)\n", ext );
+            disabled_index[i++] = j;
+        }
+        else if (!has_extension( disabled, ext, strlen( ext ) ))
         {
             TRACE( "++ %s\n", ext );
         }
@@ -272,13 +320,13 @@ static HKEY open_hkcu_key( const char *name )
             return 0;
 
         sid = ((TOKEN_USER *)sid_data)->User.Sid;
-        len = sprintf( buffer, "\\Registry\\User\\S-%u-%u", sid->Revision,
-                       (int)MAKELONG( MAKEWORD( sid->IdentifierAuthority.Value[5],
-                                                sid->IdentifierAuthority.Value[4] ),
-                                      MAKEWORD( sid->IdentifierAuthority.Value[3],
-                                                sid->IdentifierAuthority.Value[2] )));
+        len = snprintf( buffer, sizeof(buffer), "\\Registry\\User\\S-%u-%u", sid->Revision,
+                        (int)MAKELONG( MAKEWORD( sid->IdentifierAuthority.Value[5],
+                                                 sid->IdentifierAuthority.Value[4] ),
+                                       MAKEWORD( sid->IdentifierAuthority.Value[3],
+                                                 sid->IdentifierAuthority.Value[2] )));
         for (i = 0; i < sid->SubAuthorityCount; i++)
-            len += sprintf( buffer + len, "-%u", (int)sid->SubAuthority[i] );
+            len += snprintf( buffer + len, sizeof(buffer) - len, "-%u", (int)sid->SubAuthority[i] );
 
         ascii_to_unicode( bufferW, buffer, len );
         hkcu = reg_open_key( NULL, bufferW, len * sizeof(WCHAR) );
@@ -329,7 +377,6 @@ static BOOL filter_extensions( TEB * teb, const char *extensions, GLubyte **exts
         else disabled = "";
     }
 
-    if (!disabled[0]) return FALSE;
     if (extensions && !*exts_list) *exts_list = filter_extensions_list( extensions, disabled );
     if (!*disabled_exts) *disabled_exts = filter_extensions_index( teb, disabled );
     return (exts_list && *exts_list) || *disabled_exts;
@@ -401,6 +448,21 @@ static void wrap_glGetIntegerv( TEB *teb, GLenum pname, GLint *data )
 
     if (pname == GL_NUM_EXTENSIONS && (disabled = disabled_extensions_index( teb )))
         while (*disabled++ != ~0u) (*data)--;
+
+    if (is_win64 && is_wow64())
+    {
+        /* 4.4 depends on ARB_buffer_storage, which we don't support on wow64. */
+        if (pname == GL_MAJOR_VERSION && *data > 4)
+            *data = 4;
+        else if (pname == GL_MINOR_VERSION)
+        {
+            GLint major;
+
+            funcs->gl.p_glGetIntegerv( GL_MAJOR_VERSION, &major );
+            if (major == 4 && *data > 3)
+                *data = 3;
+        }
+    }
 }
 
 static const GLubyte *wrap_glGetString( TEB *teb, GLenum name )
@@ -408,12 +470,37 @@ static const GLubyte *wrap_glGetString( TEB *teb, GLenum name )
     const struct opengl_funcs *funcs = teb->glTable;
     const GLubyte *ret;
 
-    if ((ret = funcs->gl.p_glGetString( name )) && name == GL_EXTENSIONS)
+    if ((ret = funcs->gl.p_glGetString( name )))
     {
-        struct wgl_handle *ptr = get_current_context_ptr( teb );
-        GLubyte **extensions = &ptr->u.context->extensions;
-        GLuint **disabled = &ptr->u.context->disabled_exts;
-        if (*extensions || filter_extensions( teb, (const char *)ret, extensions, disabled )) return *extensions;
+        if (name == GL_EXTENSIONS)
+        {
+            struct wgl_handle *ptr = get_current_context_ptr( teb );
+            GLubyte **extensions = &ptr->u.context->extensions;
+            GLuint **disabled = &ptr->u.context->disabled_exts;
+            if (*extensions || filter_extensions( teb, (const char *)ret, extensions, disabled )) return *extensions;
+        }
+        else if (name == GL_VERSION && is_win64 && is_wow64())
+        {
+            struct wgl_handle *ptr = get_current_context_ptr( teb );
+            int major, minor;
+            const char *rest;
+
+            if (ptr->u.context->version_string)
+                return ptr->u.context->version_string;
+
+            rest = parse_gl_version( (const char *)ret, &major, &minor );
+
+            /* 4.4 depends on ARB_buffer_storage, which we don't support on wow64. */
+            if (major > 4 || (major == 4 && minor >= 4))
+            {
+                char *str = NULL;
+
+                asprintf( &str, "4.3%s", rest );
+                if (InterlockedCompareExchangePointer( (void **)&ptr->u.context->version_string, str, NULL ))
+                    free( str );
+                return ptr->u.context->version_string;
+            }
+        }
     }
 
     return ret;
@@ -451,7 +538,7 @@ static char *build_extension_list( TEB *teb )
         capacity = max( capacity, len + strlen( extension ) + 2 );
         if (!(tmp = realloc( available_extensions, capacity ))) break;
         available_extensions = tmp;
-        len += sprintf( available_extensions + len, "%s ", extension );
+        len += snprintf( available_extensions + len, capacity - len, "%s ", extension );
     }
     if (len) available_extensions[len - 1] = 0;
 
@@ -487,6 +574,23 @@ static int registry_entry_cmp( const void *a, const void *b )
     return strcmp( entry_a->name, entry_b->name );
 }
 
+static BOOL ignore_extenstions_for_get_proc_address(void)
+{
+    static int cached = -1;
+
+    if (cached == -1)
+    {
+        const char *sgi;
+
+        cached = (sgi = getenv( "SteamGameId" )) && (
+                 !strcmp( sgi, "2293310" )
+                 || !strcmp( sgi, "2914160" )
+                 );
+    }
+
+    return cached;
+}
+
 static PROC wrap_wglGetProcAddress( TEB *teb, LPCSTR name )
 {
     const struct registry_entry entry = {.name = name}, *found;
@@ -513,7 +617,7 @@ static PROC wrap_wglGetProcAddress( TEB *teb, LPCSTR name )
     {
         void *driver_func = funcs->wgl.p_wglGetProcAddress( name );
 
-        if (!is_extension_supported( teb, found->extension ))
+        if (!ignore_extenstions_for_get_proc_address() && !is_extension_supported( teb, found->extension ))
         {
             unsigned int i;
             static const struct { const char *name, *alt; } alternatives[] =
@@ -568,7 +672,7 @@ static HGLRC wrap_wglCreateContext( HDC hdc )
     HGLRC ret = 0;
     struct wgl_context *drv_ctx;
     struct opengl_context *context;
-    struct opengl_funcs *funcs = get_dc_funcs( hdc );
+    const struct opengl_funcs *funcs = get_dc_funcs( hdc );
 
     if (!funcs) return 0;
     if (!(drv_ctx = funcs->wgl.p_wglCreateContext( hdc ))) return 0;
@@ -600,7 +704,7 @@ static BOOL wrap_wglMakeCurrent( TEB *teb, HDC hdc, HGLRC hglrc )
                 teb->glReserved1[0] = hdc;
                 teb->glReserved1[1] = hdc;
                 teb->glCurrentRC = hglrc;
-                teb->glTable = ptr->funcs;
+                teb->glTable = (void *)ptr->funcs;
             }
         }
         else
@@ -637,6 +741,7 @@ static BOOL wrap_wglDeleteContext( TEB *teb, HGLRC hglrc )
     }
     if (hglrc == teb->glCurrentRC) wrap_wglMakeCurrent( teb, 0, 0 );
     ptr->funcs->wgl.p_wglDeleteContext( ptr->u.context->drv_ctx );
+    free( ptr->u.context->version_string );
     free( ptr->u.context->disabled_exts );
     free( ptr->u.context->extensions );
     free( ptr->u.context );
@@ -671,7 +776,7 @@ static HGLRC wrap_wglCreateContextAttribsARB( HDC hdc, HGLRC share, const int *a
     struct wgl_context *drv_ctx;
     struct wgl_handle *share_ptr = NULL;
     struct opengl_context *context;
-    struct opengl_funcs *funcs = get_dc_funcs( hdc );
+    const struct opengl_funcs *funcs = get_dc_funcs( hdc );
 
     if (!funcs)
     {
@@ -716,7 +821,7 @@ static HPBUFFERARB wrap_wglCreatePbufferARB( HDC hdc, int format, int width, int
 {
     HPBUFFERARB ret;
     struct wgl_pbuffer *pbuffer;
-    struct opengl_funcs *funcs = get_dc_funcs( hdc );
+    const struct opengl_funcs *funcs = get_dc_funcs( hdc );
 
     if (!funcs || !funcs->ext.p_wglCreatePbufferARB) return 0;
     if (!(pbuffer = funcs->ext.p_wglCreatePbufferARB( hdc, format, width, height, attribs ))) return 0;
@@ -762,7 +867,7 @@ static BOOL wrap_wglMakeContextCurrentARB( TEB *teb, HDC draw_hdc, HDC read_hdc,
                 teb->glReserved1[0] = draw_hdc;
                 teb->glReserved1[1] = read_hdc;
                 teb->glCurrentRC = hglrc;
-                teb->glTable = ptr->funcs;
+                teb->glTable = (void *)ptr->funcs;
             }
         }
         else
@@ -810,21 +915,15 @@ static BOOL wrap_wglSetPbufferAttribARB( HPBUFFERARB handle, const int *attribs
 }
 
 static void gl_debug_message_callback( GLenum source, GLenum type, GLuint id, GLenum severity,
-                                       GLsizei length, const GLchar *message, const void *userParam )
+                                       GLsizei length, const GLchar *message, const void *user )
 {
-    struct wine_gl_debug_message_params params =
-    {
-        .source = source,
-        .type = type,
-        .id = id,
-        .severity = severity,
-        .length = length,
-        .message = message,
-    };
+    struct gl_debug_message_callback_params *params;
     void *ret_ptr;
     ULONG ret_len;
-    struct wgl_handle *ptr = (struct wgl_handle *)userParam;
+    struct wgl_handle *ptr = (struct wgl_handle *)user;
+    UINT len = strlen( message ) + 1, size;
 
+    if (!ptr->u.context->debug_callback) return;
     if (!NtCurrentTeb())
     {
         fprintf( stderr, "msg:gl_debug_message_callback called from native thread, severity %#x, message \"%.*s\".\n",
@@ -832,46 +931,55 @@ static void gl_debug_message_callback( GLenum source, GLenum type, GLuint id, GL
         return;
     }
 
-    if (!(params.user_callback = ptr->u.context->debug_callback)) return;
-    params.user_data = ptr->u.context->debug_user;
+    size = offsetof(struct gl_debug_message_callback_params, message[len] );
+    if (!(params = malloc( size ))) return;
+    params->dispatch.callback = call_gl_debug_message_callback;
+    params->debug_callback = ptr->u.context->debug_callback;
+    params->debug_user = ptr->u.context->debug_user;
+    params->source = source;
+    params->type = type;
+    params->id = id;
+    params->severity = severity;
+    params->length = length;
+    memcpy( params->message, message, len );
 
-    KeUserModeCallback( NtUserCallOpenGLDebugMessageCallback, &params, sizeof(params),
-                        &ret_ptr, &ret_len );
+    KeUserDispatchCallback( &params->dispatch, size, &ret_ptr, &ret_len );
+    free( params );
 }
 
-static void wrap_glDebugMessageCallback( TEB *teb, GLDEBUGPROC callback, const void *userParam )
+static void wrap_glDebugMessageCallback( TEB *teb, GLDEBUGPROC callback, const void *user )
 {
     struct wgl_handle *ptr = get_current_context_ptr( teb );
     const struct opengl_funcs *funcs = teb->glTable;
 
     if (!funcs->ext.p_glDebugMessageCallback) return;
 
-    ptr->u.context->debug_callback = callback;
-    ptr->u.context->debug_user     = userParam;
+    ptr->u.context->debug_callback = (UINT_PTR)callback;
+    ptr->u.context->debug_user     = (UINT_PTR)user;
     funcs->ext.p_glDebugMessageCallback( gl_debug_message_callback, ptr );
 }
 
-static void wrap_glDebugMessageCallbackAMD( TEB *teb, GLDEBUGPROCAMD callback, void *userParam )
+static void wrap_glDebugMessageCallbackAMD( TEB *teb, GLDEBUGPROCAMD callback, void *user )
 {
     struct wgl_handle *ptr = get_current_context_ptr( teb );
     const struct opengl_funcs *funcs = teb->glTable;
 
     if (!funcs->ext.p_glDebugMessageCallbackAMD) return;
 
-    ptr->u.context->debug_callback = callback;
-    ptr->u.context->debug_user     = userParam;
+    ptr->u.context->debug_callback = (UINT_PTR)callback;
+    ptr->u.context->debug_user     = (UINT_PTR)user;
     funcs->ext.p_glDebugMessageCallbackAMD( gl_debug_message_callback, ptr );
 }
 
-static void wrap_glDebugMessageCallbackARB( TEB *teb, GLDEBUGPROCARB callback, const void *userParam )
+static void wrap_glDebugMessageCallbackARB( TEB *teb, GLDEBUGPROCARB callback, const void *user )
 {
     struct wgl_handle *ptr = get_current_context_ptr( teb );
     const struct opengl_funcs *funcs = teb->glTable;
 
     if (!funcs->ext.p_glDebugMessageCallbackARB) return;
 
-    ptr->u.context->debug_callback = callback;
-    ptr->u.context->debug_user     = userParam;
+    ptr->u.context->debug_callback = (UINT_PTR)callback;
+    ptr->u.context->debug_user     = (UINT_PTR)user;
     funcs->ext.p_glDebugMessageCallbackARB( gl_debug_message_callback, ptr );
 }
 
@@ -1059,6 +1167,13 @@ NTSTATUS ext_wglSetPbufferAttribARB( void *args )
     return STATUS_SUCCESS;
 }
 
+NTSTATUS process_attach( void *args )
+{
+    struct process_attach_params *params = args;
+    call_gl_debug_message_callback = params->call_gl_debug_message_callback;
+    return STATUS_SUCCESS;
+}
+
 NTSTATUS thread_attach( void *args )
 {
     TEB *teb = args;
@@ -1071,38 +1186,48 @@ NTSTATUS process_detach( void *args )
     return STATUS_SUCCESS;
 }
 
+NTSTATUS get_pixel_formats( void *args )
+{
+    struct get_pixel_formats_params *params = args;
+    const struct opengl_funcs *funcs = get_dc_funcs( params->hdc );
+    if (!funcs || !funcs->wgl.p_get_pixel_formats) return STATUS_NOT_IMPLEMENTED;
+    funcs->wgl.p_get_pixel_formats( params->formats, params->max_formats,
+                                    &params->num_formats, &params->num_onscreen_formats );
+    return STATUS_SUCCESS;
+}
+
 #ifdef _WIN64
 
 typedef ULONG PTR32;
 
-extern NTSTATUS ext_glClientWaitSync( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glDeleteSync( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glFenceSync( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glGetBufferPointerv( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glGetBufferPointervARB( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glGetNamedBufferPointerv( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glGetNamedBufferPointervEXT( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glGetSynciv( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glIsSync( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glMapBuffer( void *args ) DECLSPEC_HIDDEN;
+extern NTSTATUS ext_glClientWaitSync( void *args );
+extern NTSTATUS ext_glDeleteSync( void *args );
+extern NTSTATUS ext_glFenceSync( void *args );
+extern NTSTATUS ext_glGetBufferPointerv( void *args );
+extern NTSTATUS ext_glGetBufferPointervARB( void *args );
+extern NTSTATUS ext_glGetNamedBufferPointerv( void *args );
+extern NTSTATUS ext_glGetNamedBufferPointervEXT( void *args );
+extern NTSTATUS ext_glGetSynciv( void *args );
+extern NTSTATUS ext_glIsSync( void *args );
+extern NTSTATUS ext_glMapBuffer( void *args );
 
-extern NTSTATUS ext_glUnmapBuffer( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glUnmapBufferARB( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glUnmapNamedBuffer( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glUnmapNamedBufferEXT( void *args ) DECLSPEC_HIDDEN;
+extern NTSTATUS ext_glUnmapBuffer( void *args );
+extern NTSTATUS ext_glUnmapBufferARB( void *args );
+extern NTSTATUS ext_glUnmapNamedBuffer( void *args );
+extern NTSTATUS ext_glUnmapNamedBufferEXT( void *args );
 
-extern NTSTATUS ext_glMapBufferARB( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glMapBufferRange( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glMapNamedBuffer( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glMapNamedBufferEXT( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glMapNamedBufferRange( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glMapNamedBufferRangeEXT( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glPathGlyphIndexRangeNV( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_glWaitSync( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_wglGetExtensionsStringARB( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_wglGetExtensionsStringEXT( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_wglQueryCurrentRendererStringWINE( void *args ) DECLSPEC_HIDDEN;
-extern NTSTATUS ext_wglQueryRendererStringWINE( void *args ) DECLSPEC_HIDDEN;
+extern NTSTATUS ext_glMapBufferARB( void *args );
+extern NTSTATUS ext_glMapBufferRange( void *args );
+extern NTSTATUS ext_glMapNamedBuffer( void *args );
+extern NTSTATUS ext_glMapNamedBufferEXT( void *args );
+extern NTSTATUS ext_glMapNamedBufferRange( void *args );
+extern NTSTATUS ext_glMapNamedBufferRangeEXT( void *args );
+extern NTSTATUS ext_glPathGlyphIndexRangeNV( void *args );
+extern NTSTATUS ext_glWaitSync( void *args );
+extern NTSTATUS ext_wglGetExtensionsStringARB( void *args );
+extern NTSTATUS ext_wglGetExtensionsStringEXT( void *args );
+extern NTSTATUS ext_wglQueryCurrentRendererStringWINE( void *args );
+extern NTSTATUS ext_wglQueryRendererStringWINE( void *args );
 
 struct wow64_string_entry
 {
@@ -1769,10 +1894,15 @@ static void unmap_named_buffer( TEB *teb, GLint buffer )
 static NTSTATUS wow64_map_buffer( TEB *teb, GLint buffer, GLenum target, void *ptr, SIZE_T size,
                                   GLbitfield access, PTR32 *ret )
 {
+    static unsigned int once;
+
     if (*ret)  /* wow64 pointer provided, map buffer to it */
     {
-        if (access & GL_MAP_READ_BIT)
+        if (!(access & (GL_MAP_INVALIDATE_RANGE_BIT | GL_MAP_INVALIDATE_BUFFER_BIT)))
         {
+            if (!once++)
+                FIXME( "Doing a copy of a mapped buffer (expect performance issues)\n" );
+
             TRACE( "Copying %#zx from buffer at %p to wow64 buffer %p\n", size, ptr, UlongToPtr(*ret) );
             memcpy( UlongToPtr(*ret), ptr, size );
         }
@@ -1801,6 +1931,19 @@ static NTSTATUS wow64_map_buffer( TEB *teb, GLint buffer, GLenum target, void *p
     return STATUS_INVALID_ADDRESS;
 }
 
+static GLbitfield map_range_flags_from_map_flags( GLenum flags )
+{
+    switch (flags)
+    {
+        case GL_READ_ONLY: return GL_MAP_READ_BIT;
+        case GL_WRITE_ONLY: return GL_MAP_WRITE_BIT;
+        case GL_READ_WRITE: return GL_MAP_READ_BIT | GL_MAP_WRITE_BIT;
+        default:
+            ERR( "invalid map flags %#x\n", flags );
+            return GL_MAP_READ_BIT | GL_MAP_WRITE_BIT;
+    }
+}
+
 static NTSTATUS wow64_unmap_buffer( void *ptr, SIZE_T size, GLbitfield access )
 {
     void *wow_ptr;
@@ -1912,7 +2055,8 @@ static NTSTATUS wow64_gl_map_buffer( void *args, NTSTATUS (*gl_map_buffer64)(voi
     if (params32->ret) params.ret = get_buffer_pointer( params.teb, params.target );
     else if ((status = gl_map_buffer64( &params ))) return status;
 
-    status = wow64_map_buffer( params.teb, 0, params.target, params.ret, 0, params.access, &params32->ret );
+    status = wow64_map_buffer( params.teb, 0, params.target, params.ret, 0,
+                               map_range_flags_from_map_flags( params.access ), &params32->ret );
     if (!status || status == STATUS_INVALID_ADDRESS) return status;
 
     unmap_buffer( params.teb, params.target );
@@ -1982,7 +2126,8 @@ static NTSTATUS wow64_gl_map_named_buffer( void *args, NTSTATUS (*gl_map_named_b
     if (params32->ret) params.ret = get_named_buffer_pointer( params.teb, params.buffer );
     else if ((status = gl_map_named_buffer64( &params ))) return status;
 
-    status = wow64_map_buffer( params.teb, params.buffer, 0, params.ret, 0, params.access, &params32->ret );
+    status = wow64_map_buffer( params.teb, params.buffer, 0, params.ret, 0,
+                               map_range_flags_from_map_flags( params.access ), &params32->ret );
     if (!status || status == STATUS_INVALID_ADDRESS) return status;
 
     unmap_named_buffer( params.teb, params.buffer );
@@ -2050,14 +2195,20 @@ static NTSTATUS wow64_gl_unmap_buffer( void *args, NTSTATUS (*gl_unmap_buffer64)
         GLenum target;
         GLboolean ret;
     } *params32 = args;
+    struct glUnmapBuffer_params params =
+    {
+        .teb = get_teb64(params32->teb),
+        .target = params32->target,
+        .ret = TRUE,
+    };
     NTSTATUS status;
-    TEB *teb = get_teb64( params32->teb );
 
-    if (!(ptr = get_buffer_pointer( teb, params32->target ))) return STATUS_SUCCESS;
+    if (!(ptr = get_buffer_pointer( params.teb, params.target ))) return STATUS_SUCCESS;
 
-    status = wow64_unmap_buffer( ptr, get_buffer_param( teb, params32->target, GL_BUFFER_MAP_LENGTH ),
-                                 get_buffer_param( teb, params32->target, GL_BUFFER_ACCESS_FLAGS ) );
-    gl_unmap_buffer64( args );
+    status = wow64_unmap_buffer( ptr, get_buffer_param( params.teb, params.target, GL_BUFFER_MAP_LENGTH ),
+                                 get_buffer_param( params.teb, params.target, GL_BUFFER_ACCESS_FLAGS ) );
+    gl_unmap_buffer64( &params );
+    params32->ret = params.ret;
 
     return status;
 }
@@ -2081,14 +2232,20 @@ static NTSTATUS wow64_gl_unmap_named_buffer( void *args, NTSTATUS (*gl_unmap_nam
         GLint buffer;
         GLboolean ret;
     } *params32 = args;
+    struct glUnmapNamedBuffer_params params =
+    {
+        .teb = get_teb64(params32->teb),
+        .buffer = params32->buffer,
+        .ret = TRUE,
+    };
     NTSTATUS status;
-    TEB *teb = get_teb64( params32->teb );
 
-    if (!(ptr = get_named_buffer_pointer( teb, params32->buffer ))) return STATUS_SUCCESS;
+    if (!(ptr = get_named_buffer_pointer( params.teb, params.buffer ))) return STATUS_SUCCESS;
 
-    status = wow64_unmap_buffer( ptr, get_named_buffer_param( teb, params32->buffer, GL_BUFFER_MAP_LENGTH ),
-                                 get_named_buffer_param( teb, params32->buffer, GL_BUFFER_ACCESS_FLAGS ) );
-    gl_unmap_named_buffer64( args );
+    status = wow64_unmap_buffer( ptr, get_named_buffer_param( params.teb, params.buffer, GL_BUFFER_MAP_LENGTH ),
+                                 get_named_buffer_param( params.teb, params.buffer, GL_BUFFER_ACCESS_FLAGS ) );
+    gl_unmap_named_buffer64( &params );
+    params32->ret = params.ret;
 
     return status;
 }
@@ -2121,4 +2278,29 @@ NTSTATUS wow64_process_detach( void *args )
     return STATUS_SUCCESS;
 }
 
+NTSTATUS wow64_get_pixel_formats( void *args )
+{
+    struct
+    {
+        PTR32 teb;
+        PTR32 hdc;
+        PTR32 formats;
+        UINT max_formats;
+        UINT num_formats;
+        UINT num_onscreen_formats;
+    } *params32 = args;
+    struct get_pixel_formats_params params =
+    {
+        .teb = get_teb64(params32->teb),
+        .hdc = ULongToPtr(params32->hdc),
+        .formats = ULongToPtr(params32->formats),
+        .max_formats = params32->max_formats,
+    };
+    NTSTATUS status;
+    status = get_pixel_formats( &params );
+    params32->num_formats = params.num_formats;
+    params32->num_onscreen_formats = params.num_onscreen_formats;
+    return status;
+}
+
 #endif
