@@ -23,7 +23,6 @@
 #include <stdlib.h>
 #include <errno.h>
 
-#define NONAMELESSUNION
 #include "windef.h"
 #include "winbase.h"
 #include "winternl.h"
@@ -63,7 +62,9 @@ struct schan_context
     ULONG req_ctx_attr;
     const CERT_CONTEXT *cert;
     SIZE_T header_size;
-    BOOL shutdown_requested;
+    enum control_token control_token;
+    unsigned int alert_type;
+    unsigned int alert_number;
     BOOL rehandshake_requested;
 };
 
@@ -332,7 +333,7 @@ static SECURITY_STATUS SEC_ENTRY schan_QueryCredentialsAttributesA(
     {
     case SECPKG_CRED_ATTR_NAMES:
         FIXME("SECPKG_CRED_ATTR_NAMES: stub\n");
-        ret = SEC_E_UNSUPPORTED_FUNCTION;
+        ret = SEC_E_NO_CREDENTIALS;
         break;
     default:
         ret = schan_QueryCredentialsAttributes(phCredential, ulAttribute,
@@ -414,7 +415,7 @@ static SECURITY_STATUS get_cert(const void *credentials, CERT_CONTEXT const **ce
 
     default:
         FIXME("unhandled version %lu\n", cred->dwVersion);
-        return SEC_E_INTERNAL_ERROR;
+        return SEC_E_UNKNOWN_CREDENTIALS;
     }
 
     if (!cert_count) status = SEC_E_NO_CREDENTIALS;
@@ -565,7 +566,7 @@ static SECURITY_STATUS acquire_credentials_handle(ULONG fCredentialUse,
  const SCHANNEL_CRED *schanCred, PCredHandle phCredential, PTimeStamp ptsExpiry)
 {
     struct schan_credentials *creds;
-    DWORD enabled_protocols, cred_enabled_protocols;
+    DWORD enabled_protocols, cred_enabled_protocols = 0;
     ULONG_PTR handle;
     SECURITY_STATUS status = SEC_E_OK;
     const CERT_CONTEXT *cert = NULL;
@@ -590,13 +591,13 @@ static SECURITY_STATUS acquire_credentials_handle(ULONG fCredentialUse,
 
         status = SEC_E_OK;
     }
-    else if (fCredentialUse & SECPKG_CRED_INBOUND)
+    else if (!fCredentialUse || (fCredentialUse & SECPKG_CRED_INBOUND))
     {
         return SEC_E_NO_CREDENTIALS;
     }
 
     read_config();
-    if(schanCred && cred_enabled_protocols)
+    if (cred_enabled_protocols)
         enabled_protocols = cred_enabled_protocols & config_enabled_protocols;
     else
         enabled_protocols = config_enabled_protocols & ~config_default_disabled_protocols;
@@ -604,7 +605,8 @@ static SECURITY_STATUS acquire_credentials_handle(ULONG fCredentialUse,
         enabled_protocols &= ~SP_PROT_X_CLIENTS;
     if (!(fCredentialUse & SECPKG_CRED_INBOUND))
         enabled_protocols &= ~SP_PROT_X_SERVERS;
-    if(!enabled_protocols) {
+    if (!enabled_protocols)
+    {
         ERR("Could not find matching protocol\n");
         return SEC_E_ALGORITHM_MISMATCH;
     }
@@ -887,9 +889,9 @@ static SECURITY_STATUS establish_context(
         unsigned char *ptr;
 
         if (phContext && !(ctx = schan_get_object(phContext->dwLower, SCHAN_HANDLE_CTX))) return SEC_E_INVALID_HANDLE;
-        if (!pInput && !ctx->shutdown_requested && !is_dtls_context(ctx)) return SEC_E_INCOMPLETE_MESSAGE;
+        if (!pInput && !ctx->control_token && !is_dtls_context(ctx)) return SEC_E_INCOMPLETE_MESSAGE;
 
-        if (!ctx->shutdown_requested && pInput)
+        if (!ctx->control_token && pInput)
         {
             if (!validate_input_buffers(pInput)) return SEC_E_INVALID_TOKEN;
             if ((idx = schan_find_sec_buffer_idx(pInput, 0, SECBUFFER_TOKEN)) == -1) return SEC_E_INCOMPLETE_MESSAGE;
@@ -962,8 +964,10 @@ static SECURITY_STATUS establish_context(
     params.input_offset = &input_offset;
     params.output_buffer_idx = &output_buffer_idx;
     params.output_offset = &output_offset;
-    params.control_token = ctx->shutdown_requested ? control_token_shutdown : control_token_none;
-    ctx->shutdown_requested = FALSE;
+    params.control_token = ctx->control_token;
+    params.alert_type = ctx->alert_type;
+    params.alert_number = ctx->alert_number;
+    ctx->control_token = CONTROL_TOKEN_NONE;
     ctx->rehandshake_requested = FALSE;
     ret = GNUTLS_CALL( handshake, &params );
 
@@ -1053,7 +1057,7 @@ static SECURITY_STATUS SEC_ENTRY schan_InitializeSecurityContextA(
     SECURITY_STATUS ret;
     SEC_WCHAR *target_name = NULL;
 
-    TRACE("%p %p %s %ld %ld %ld %p %ld %p %p %p %p\n", phCredential, phContext,
+    TRACE("%p %p %s 0x%08lx %ld %ld %p %ld %p %p %p %p\n", phCredential, phContext,
      debugstr_a(pszTargetName), fContextReq, Reserved1, TargetDataRep, pInput,
      Reserved1, phNewContext, pOutput, pfContextAttr, ptsExpiry);
 
@@ -1169,197 +1173,178 @@ static SECURITY_STATUS SEC_ENTRY schan_QueryContextAttributesW(
     struct schan_context *ctx;
     SECURITY_STATUS status;
 
-    TRACE("context_handle %p, attribute %#lx, buffer %p\n",
-            context_handle, attribute, buffer);
+    TRACE("context_handle %p, attribute %#lx, buffer %p\n", context_handle, attribute, buffer);
 
-    if (!context_handle) return SEC_E_INVALID_HANDLE;
-    ctx = schan_get_object(context_handle->dwLower, SCHAN_HANDLE_CTX);
+    if (!context_handle || !(ctx = schan_get_object(context_handle->dwLower, SCHAN_HANDLE_CTX)))
+        return SEC_E_INVALID_HANDLE;
 
-    switch(attribute)
+    switch (attribute)
     {
-        case SECPKG_ATTR_STREAM_SIZES:
+    case SECPKG_ATTR_STREAM_SIZES:
+    {
+        SecPkgContext_ConnectionInfo info;
+        struct get_connection_info_params params = { ctx->session, &info };
+        status = GNUTLS_CALL( get_connection_info, &params );
+        if (status == SEC_E_OK)
         {
-            SecPkgContext_ConnectionInfo info;
-            struct get_connection_info_params params = { ctx->session, &info };
-            status = GNUTLS_CALL( get_connection_info, &params );
-            if (status == SEC_E_OK)
-            {
-                struct session_params params = { ctx->session };
-                SecPkgContext_StreamSizes *stream_sizes = buffer;
-                SIZE_T mac_size = info.dwHashStrength;
-                unsigned int block_size = GNUTLS_CALL( get_session_cipher_block_size, &params );
-                unsigned int message_size = GNUTLS_CALL( get_max_message_size, &params );
+            struct session_params params = { ctx->session };
+            SecPkgContext_StreamSizes *stream_sizes = buffer;
+            SIZE_T mac_size = info.dwHashStrength;
+            unsigned int block_size = GNUTLS_CALL( get_session_cipher_block_size, &params );
+            unsigned int message_size = GNUTLS_CALL( get_max_message_size, &params );
 
-                TRACE("Using header size %Iu mac bytes %Iu, message size %u, block size %u\n",
-                      ctx->header_size, mac_size, message_size, block_size);
+            TRACE("Using header size %Iu mac bytes %Iu, message size %u, block size %u\n",
+                  ctx->header_size, mac_size, message_size, block_size);
 
-                /* These are defined by the TLS RFC */
-                stream_sizes->cbHeader = ctx->header_size;
-                stream_sizes->cbTrailer = mac_size + 256; /* Max 255 bytes padding + 1 for padding size */
-                stream_sizes->cbMaximumMessage = message_size;
-                stream_sizes->cBuffers = 4;
-                stream_sizes->cbBlockSize = block_size;
-            }
-
-            return status;
-        }
-        case SECPKG_ATTR_KEY_INFO:
-        {
-            SecPkgContext_ConnectionInfo conn_info;
-            struct get_connection_info_params params = { ctx->session, &conn_info };
-            status = GNUTLS_CALL( get_connection_info, &params );
-            if (status == SEC_E_OK)
-            {
-                struct session_params params = { ctx->session };
-                SecPkgContext_KeyInfoW *info = buffer;
-                info->KeySize = conn_info.dwCipherStrength;
-                info->SignatureAlgorithm = GNUTLS_CALL( get_key_signature_algorithm, &params );
-                info->EncryptAlgorithm = conn_info.aiCipher;
-                info->sSignatureAlgorithmName = get_alg_name(info->SignatureAlgorithm, TRUE);
-                info->sEncryptAlgorithmName = get_alg_name(info->EncryptAlgorithm, TRUE);
-            }
-            return status;
-        }
-        case SECPKG_ATTR_REMOTE_CERT_CONTEXT:
-        {
-            PCCERT_CONTEXT *cert = buffer;
-
-            status = ensure_remote_cert(ctx);
-            if(status != SEC_E_OK)
-                return status;
-
-            *cert = CertDuplicateCertificateContext(ctx->cert);
-            return SEC_E_OK;
-        }
-        case SECPKG_ATTR_CONNECTION_INFO:
-        {
-            SecPkgContext_ConnectionInfo *info = buffer;
-            struct get_connection_info_params params = { ctx->session, info };
-            return GNUTLS_CALL( get_connection_info, &params );
-        }
-        case SECPKG_ATTR_ENDPOINT_BINDINGS:
-        {
-            SecPkgContext_Bindings *bindings = buffer;
-            CCRYPT_OID_INFO *info;
-            ALG_ID hash_alg = CALG_SHA_256;
-            BYTE hash[1024];
-            DWORD hash_size;
-            char *p;
-            BOOL r;
-
-            static const char prefix[] = "tls-server-end-point:";
-
-            status = ensure_remote_cert(ctx);
-            if(status != SEC_E_OK)
-                return status;
-
-            /* RFC 5929 */
-            info = CryptFindOIDInfo(CRYPT_OID_INFO_OID_KEY, ctx->cert->pCertInfo->SignatureAlgorithm.pszObjId, 0);
-            if(info && info->u.Algid != CALG_SHA1 && info->u.Algid != CALG_MD5)
-                hash_alg = info->u.Algid;
-
-            hash_size = sizeof(hash);
-            r = CryptHashCertificate(0, hash_alg, 0, ctx->cert->pbCertEncoded, ctx->cert->cbCertEncoded, hash, &hash_size);
-            if(!r)
-                return GetLastError();
-
-            bindings->BindingsLength = sizeof(*bindings->Bindings) + sizeof(prefix)-1 + hash_size;
-            /* freed with FreeContextBuffer */
-            bindings->Bindings = RtlAllocateHeap(GetProcessHeap(), HEAP_ZERO_MEMORY, bindings->BindingsLength);
-            if(!bindings->Bindings)
-                return SEC_E_INSUFFICIENT_MEMORY;
-
-            bindings->Bindings->cbApplicationDataLength = sizeof(prefix)-1 + hash_size;
-            bindings->Bindings->dwApplicationDataOffset = sizeof(*bindings->Bindings);
-
-            p = (char*)(bindings->Bindings+1);
-            memcpy(p, prefix, sizeof(prefix)-1);
-            p += sizeof(prefix)-1;
-            memcpy(p, hash, hash_size);
-            return SEC_E_OK;
-        }
-        case SECPKG_ATTR_UNIQUE_BINDINGS:
-        {
-            static const char prefix[] = "tls-unique:";
-            SecPkgContext_Bindings *bindings = buffer;
-            ULONG size;
-            char *p;
-            struct get_unique_channel_binding_params params = { ctx->session, NULL, &size };
-
-            if (GNUTLS_CALL( get_unique_channel_binding, &params ) != SEC_E_BUFFER_TOO_SMALL)
-                return SEC_E_INTERNAL_ERROR;
-
-            bindings->BindingsLength = sizeof(*bindings->Bindings) + sizeof(prefix)-1 + size;
-            /* freed with FreeContextBuffer */
-            bindings->Bindings = RtlAllocateHeap(GetProcessHeap(), HEAP_ZERO_MEMORY, bindings->BindingsLength);
-            if(!bindings->Bindings)
-                return SEC_E_INSUFFICIENT_MEMORY;
-
-            bindings->Bindings->cbApplicationDataLength = sizeof(prefix)-1 + size;
-            bindings->Bindings->dwApplicationDataOffset = sizeof(*bindings->Bindings);
-
-            p = (char*)(bindings->Bindings+1);
-            memcpy(p, prefix, sizeof(prefix)-1);
-            p += sizeof(prefix)-1;
-            params.buffer = p;
-            return GNUTLS_CALL( get_unique_channel_binding, &params );
-        }
-        case SECPKG_ATTR_APPLICATION_PROTOCOL:
-        {
-            SecPkgContext_ApplicationProtocol *protocol = buffer;
-            struct get_application_protocol_params params = { ctx->session, protocol };
-            return GNUTLS_CALL( get_application_protocol, &params );
-        }
-        case SECPKG_ATTR_CIPHER_INFO:
-        {
-            SecPkgContext_CipherInfo *info = buffer;
-            struct get_cipher_info_params params = { ctx->session, info };
-            return GNUTLS_CALL( get_cipher_info, &params );
+            /* These are defined by the TLS RFC */
+            stream_sizes->cbHeader = ctx->header_size;
+            stream_sizes->cbTrailer = mac_size + 256; /* Max 255 bytes padding + 1 for padding size */
+            stream_sizes->cbMaximumMessage = message_size;
+            stream_sizes->cBuffers = 4;
+            stream_sizes->cbBlockSize = block_size;
         }
 
-        default:
-            FIXME("Unhandled attribute %#lx\n", attribute);
-            return SEC_E_UNSUPPORTED_FUNCTION;
+        return status;
+    }
+    case SECPKG_ATTR_KEY_INFO:
+    {
+        SecPkgContext_ConnectionInfo conn_info;
+        struct get_connection_info_params params = { ctx->session, &conn_info };
+        status = GNUTLS_CALL( get_connection_info, &params );
+        if (status == SEC_E_OK)
+        {
+            struct session_params params = { ctx->session };
+            SecPkgContext_KeyInfoW *info = buffer;
+            info->KeySize = conn_info.dwCipherStrength;
+            info->SignatureAlgorithm = GNUTLS_CALL( get_key_signature_algorithm, &params );
+            info->EncryptAlgorithm = conn_info.aiCipher;
+            info->sSignatureAlgorithmName = get_alg_name(info->SignatureAlgorithm, TRUE);
+            info->sEncryptAlgorithmName = get_alg_name(info->EncryptAlgorithm, TRUE);
+        }
+        return status;
+    }
+    case SECPKG_ATTR_REMOTE_CERT_CONTEXT:
+    {
+        PCCERT_CONTEXT *cert = buffer;
+
+        if ((status = ensure_remote_cert(ctx)) != SEC_E_OK) return status;
+        *cert = CertDuplicateCertificateContext(ctx->cert);
+        return SEC_E_OK;
+    }
+    case SECPKG_ATTR_CONNECTION_INFO:
+    {
+        SecPkgContext_ConnectionInfo *info = buffer;
+        struct get_connection_info_params params = { ctx->session, info };
+        return GNUTLS_CALL( get_connection_info, &params );
+    }
+    case SECPKG_ATTR_ENDPOINT_BINDINGS:
+    {
+        static const char prefix[] = "tls-server-end-point:";
+        SecPkgContext_Bindings *bindings = buffer;
+        CCRYPT_OID_INFO *info;
+        ALG_ID hash_alg = CALG_SHA_256;
+        BYTE hash[1024];
+        DWORD hash_size;
+        char *p;
+        BOOL ret;
+
+        if ((status = ensure_remote_cert(ctx)) != SEC_E_OK) return status;
+
+        /* RFC 5929 */
+        info = CryptFindOIDInfo(CRYPT_OID_INFO_OID_KEY, ctx->cert->pCertInfo->SignatureAlgorithm.pszObjId, 0);
+        if (info && info->Algid != CALG_SHA1 && info->Algid != CALG_MD5) hash_alg = info->Algid;
+
+        hash_size = sizeof(hash);
+        ret = CryptHashCertificate(0, hash_alg, 0, ctx->cert->pbCertEncoded, ctx->cert->cbCertEncoded, hash, &hash_size);
+        if (!ret) return GetLastError();
+
+        bindings->BindingsLength = sizeof(*bindings->Bindings) + sizeof(prefix) - 1 + hash_size;
+        /* freed with FreeContextBuffer */
+        bindings->Bindings = RtlAllocateHeap(GetProcessHeap(), HEAP_ZERO_MEMORY, bindings->BindingsLength);
+        if (!bindings->Bindings) return SEC_E_INSUFFICIENT_MEMORY;
+
+        bindings->Bindings->cbApplicationDataLength = sizeof(prefix) - 1 + hash_size;
+        bindings->Bindings->dwApplicationDataOffset = sizeof(*bindings->Bindings);
+
+        p = (char *)(bindings->Bindings + 1);
+        memcpy(p, prefix, sizeof(prefix) - 1);
+        p += sizeof(prefix) - 1;
+        memcpy(p, hash, hash_size);
+        return SEC_E_OK;
+    }
+    case SECPKG_ATTR_UNIQUE_BINDINGS:
+    {
+        static const char prefix[] = "tls-unique:";
+        SecPkgContext_Bindings *bindings = buffer;
+        ULONG size;
+        char *p;
+        struct get_unique_channel_binding_params params = { ctx->session, NULL, &size };
+
+        if (GNUTLS_CALL( get_unique_channel_binding, &params ) != SEC_E_BUFFER_TOO_SMALL)
+            return SEC_E_INTERNAL_ERROR;
+
+        bindings->BindingsLength = sizeof(*bindings->Bindings) + sizeof(prefix) - 1 + size;
+        /* freed with FreeContextBuffer */
+        bindings->Bindings = RtlAllocateHeap(GetProcessHeap(), HEAP_ZERO_MEMORY, bindings->BindingsLength);
+        if (!bindings->Bindings) return SEC_E_INSUFFICIENT_MEMORY;
+
+        bindings->Bindings->cbApplicationDataLength = sizeof(prefix) - 1 + size;
+        bindings->Bindings->dwApplicationDataOffset = sizeof(*bindings->Bindings);
+
+        p = (char *)(bindings->Bindings + 1);
+        memcpy(p, prefix, sizeof(prefix) - 1);
+        p += sizeof(prefix) - 1;
+        params.buffer = p;
+        return GNUTLS_CALL( get_unique_channel_binding, &params );
+    }
+    case SECPKG_ATTR_APPLICATION_PROTOCOL:
+    {
+        SecPkgContext_ApplicationProtocol *protocol = buffer;
+        struct get_application_protocol_params params = { ctx->session, protocol };
+        return GNUTLS_CALL( get_application_protocol, &params );
+    }
+    case SECPKG_ATTR_CIPHER_INFO:
+    {
+        SecPkgContext_CipherInfo *info = buffer;
+        struct get_cipher_info_params params = { ctx->session, info };
+        return GNUTLS_CALL( get_cipher_info, &params );
+    }
+    default:
+        FIXME("Unhandled attribute %#lx\n", attribute);
+        return SEC_E_UNSUPPORTED_FUNCTION;
     }
 }
 
 static SECURITY_STATUS SEC_ENTRY schan_QueryContextAttributesA(
         PCtxtHandle context_handle, ULONG attribute, PVOID buffer)
 {
-    TRACE("context_handle %p, attribute %#lx, buffer %p\n",
-            context_handle, attribute, buffer);
+    TRACE("context_handle %p, attribute %#lx, buffer %p\n", context_handle, attribute, buffer);
 
     switch(attribute)
     {
-        case SECPKG_ATTR_STREAM_SIZES:
-            return schan_QueryContextAttributesW(context_handle, attribute, buffer);
-        case SECPKG_ATTR_KEY_INFO:
+    case SECPKG_ATTR_KEY_INFO:
+    {
+        SECURITY_STATUS status = schan_QueryContextAttributesW(context_handle, attribute, buffer);
+        if (status == SEC_E_OK)
         {
-            SECURITY_STATUS status = schan_QueryContextAttributesW(context_handle, attribute, buffer);
-            if (status == SEC_E_OK)
-            {
-                SecPkgContext_KeyInfoA *info = buffer;
-                info->sSignatureAlgorithmName = get_alg_name(info->SignatureAlgorithm, FALSE);
-                info->sEncryptAlgorithmName = get_alg_name(info->EncryptAlgorithm, FALSE);
-            }
-            return status;
+            SecPkgContext_KeyInfoA *info = buffer;
+            info->sSignatureAlgorithmName = get_alg_name(info->SignatureAlgorithm, FALSE);
+            info->sEncryptAlgorithmName = get_alg_name(info->EncryptAlgorithm, FALSE);
         }
-        case SECPKG_ATTR_REMOTE_CERT_CONTEXT:
-            return schan_QueryContextAttributesW(context_handle, attribute, buffer);
-        case SECPKG_ATTR_CONNECTION_INFO:
-            return schan_QueryContextAttributesW(context_handle, attribute, buffer);
-        case SECPKG_ATTR_ENDPOINT_BINDINGS:
-            return schan_QueryContextAttributesW(context_handle, attribute, buffer);
-        case SECPKG_ATTR_UNIQUE_BINDINGS:
-            return schan_QueryContextAttributesW(context_handle, attribute, buffer);
-        case SECPKG_ATTR_APPLICATION_PROTOCOL:
-            return schan_QueryContextAttributesW(context_handle, attribute, buffer);
-        case SECPKG_ATTR_CIPHER_INFO:
-            return schan_QueryContextAttributesW(context_handle, attribute, buffer);
+        return status;
+    }
+    case SECPKG_ATTR_STREAM_SIZES:
+    case SECPKG_ATTR_REMOTE_CERT_CONTEXT:
+    case SECPKG_ATTR_CONNECTION_INFO:
+    case SECPKG_ATTR_ENDPOINT_BINDINGS:
+    case SECPKG_ATTR_UNIQUE_BINDINGS:
+    case SECPKG_ATTR_APPLICATION_PROTOCOL:
+    case SECPKG_ATTR_CIPHER_INFO:
+        return schan_QueryContextAttributesW(context_handle, attribute, buffer);
 
-        default:
-            FIXME("Unhandled attribute %#lx\n", attribute);
-            return SEC_E_UNSUPPORTED_FUNCTION;
+    default:
+        FIXME("Unhandled attribute %#lx\n", attribute);
+        return SEC_E_UNSUPPORTED_FUNCTION;
     }
 }
 
@@ -1612,23 +1597,43 @@ static SECURITY_STATUS SEC_ENTRY schan_DeleteSecurityContext(PCtxtHandle context
 static SECURITY_STATUS SEC_ENTRY schan_ApplyControlToken(PCtxtHandle context_handle, PSecBufferDesc input)
 {
     struct schan_context *ctx;
+    DWORD type;
 
     TRACE("%p %p\n", context_handle, input);
 
     dump_buffer_desc(input);
 
-    if (!context_handle) return SEC_E_INVALID_HANDLE;
+    if (!context_handle || !(ctx = schan_get_object(context_handle->dwLower, SCHAN_HANDLE_CTX)))
+        return SEC_E_INVALID_HANDLE;
     if (!input) return SEC_E_INTERNAL_ERROR;
 
     if (input->cBuffers != 1) return SEC_E_INVALID_TOKEN;
     if (input->pBuffers[0].BufferType != SECBUFFER_TOKEN) return SEC_E_INVALID_TOKEN;
-    if (input->pBuffers[0].cbBuffer < sizeof(DWORD)) return SEC_E_UNSUPPORTED_FUNCTION;
-    if (*(DWORD *)input->pBuffers[0].pvBuffer != SCHANNEL_SHUTDOWN) return SEC_E_UNSUPPORTED_FUNCTION;
+    if (input->pBuffers[0].cbBuffer < sizeof(type)) return SEC_E_UNSUPPORTED_FUNCTION;
+    type = *(DWORD *)input->pBuffers[0].pvBuffer;
 
-    ctx = schan_get_object(context_handle->dwLower, SCHAN_HANDLE_CTX);
-    ctx->shutdown_requested = TRUE;
+    switch (type)
+    {
+    case SCHANNEL_SHUTDOWN:
+        ctx->control_token = CONTROL_TOKEN_SHUTDOWN;
+        ctx->alert_type = TLS1_ALERT_WARNING;
+        ctx->alert_number = TLS1_ALERT_CLOSE_NOTIFY;
+        return SEC_E_OK;
 
-    return SEC_E_OK;
+    case SCHANNEL_ALERT:
+    {
+        SCHANNEL_ALERT_TOKEN *alert = input->pBuffers[0].pvBuffer;
+        if (input->pBuffers[0].cbBuffer < sizeof(*alert)) return SEC_E_INVALID_TOKEN;
+        ctx->control_token = CONTROL_TOKEN_ALERT;
+        ctx->alert_type = alert->dwAlertType;
+        ctx->alert_number = alert->dwAlertNumber;
+        return SEC_E_OK;
+    }
+
+    default:
+        FIXME("token type %lu not supported\n", type);
+        return SEC_E_UNSUPPORTED_FUNCTION;
+    }
 }
 
 static const SecurityFunctionTableA schanTableA = {
