@@ -117,13 +117,9 @@ static inline struct base_device *impl_from_unix_device(struct unix_device *ifac
     return CONTAINING_RECORD(iface, struct base_device, unix_device);
 }
 
-#define QUIRK_DS4_BT 0x1
-#define QUIRK_DUALSENSE_BT 0x2
-
 struct hidraw_device
 {
     struct base_device base;
-    DWORD quirks;
 };
 
 static inline struct hidraw_device *hidraw_impl_from_unix_device(struct unix_device *iface)
@@ -193,7 +189,11 @@ struct lnxev_device
     BYTE hat_map[8];
     BYTE button_map[KEY_MAX];
 
-    int haptic_effect_id;
+    int haptics_state;
+    pthread_cond_t haptics_cond;
+    pthread_t haptics_thread;
+    struct ff_effect haptics;
+
     int effect_ids[256];
     LONG effect_flags;
 };
@@ -340,77 +340,7 @@ static void hidraw_device_read_report(struct unix_device *iface)
     else if (size == 0)
         TRACE("Failed to read report\n");
     else
-    {
-        /* As described in the Linux kernel driver, when connected over bluetooth, DS4 controllers
-         * start sending input through report #17 as soon as they receive a feature report #2, which
-         * the kernel sends anyway for calibration.
-         *
-         * Input report #17 is the same as the default input report #1, with additional gyro data and
-         * two additional bytes in front, but is only described as vendor specific in the report descriptor,
-         * and applications aren't expecting it.
-         *
-         * We have to translate it to input report #1, like native driver does.
-         */
-        if ((impl->quirks & QUIRK_DS4_BT) && report_buffer[0] == 0x11 && size >= 12)
-        {
-            size = 10;
-            buff += 2;
-            buff[0] = 1;
-        }
-
-        /* The behavior of DualSense is very similar to DS4 described above with a few exceptions.
-         *
-         * The report number #41 is used for the extended bluetooth input report. The report comes
-         * with only one extra byte in front and the format is not exactly the same as the one used
-         * for the report #1 so we need to shuffle a few bytes around.
-         *
-         * Basic #1 report:
-         *   X  Y  Z  RZ  Buttons[3]  TriggerLeft  TriggerRight
-         *
-         * Extended #41 report:
-         *   Prefix X  Y  Z  Rz  TriggerLeft  TriggerRight  Counter  Buttons[3] ...
-         */
-        if ((impl->quirks & QUIRK_DUALSENSE_BT) && report_buffer[0] == 0x31 && size >= 11)
-        {
-            BYTE trigger[2];
-            size = 10;
-            buff += 1;
-
-            buff[0] = 1; /* fake report #1 */
-
-            trigger[0] = buff[5]; /* TriggerLeft*/
-            trigger[1] = buff[6]; /* TriggerRight */
-
-            buff[5] = buff[8];    /* Buttons[0] */
-            buff[6] = buff[9];    /* Buttons[1] */
-            buff[7] = buff[10];   /* Buttons[2] */
-            buff[8] = trigger[0]; /* TriggerLeft */
-            buff[9] = trigger[1]; /* TirggerRight */
-        }
-
         bus_event_queue_input_report(&event_queue, iface, buff, size);
-    }
-}
-
-static void hidraw_disable_sony_quirks(struct unix_device *iface)
-{
-    struct hidraw_device *impl = hidraw_impl_from_unix_device(iface);
-
-    /* FIXME: we may want to validate CRC at the end of the outbound HID reports,
-     * as controllers do not switch modes if it is incorrect.
-     */
-
-    if ((impl->quirks & QUIRK_DS4_BT))
-    {
-        TRACE("Disabling report quirk for Bluetooth DualShock4 controller iface %p\n", iface);
-        impl->quirks &= ~QUIRK_DS4_BT;
-    }
-
-    if ((impl->quirks & QUIRK_DUALSENSE_BT))
-    {
-        TRACE("Disabling report quirk for Bluetooth DualSense controller iface %p\n", iface);
-        impl->quirks &= ~QUIRK_DUALSENSE_BT;
-    }
 }
 
 static void hidraw_device_set_output_report(struct unix_device *iface, HID_XFER_PACKET *packet, IO_STATUS_BLOCK *io)
@@ -432,7 +362,6 @@ static void hidraw_device_set_output_report(struct unix_device *iface, HID_XFER_
 
     if (count > 0)
     {
-        hidraw_disable_sony_quirks(iface);
         io->Information = count;
         io->Status = STATUS_SUCCESS;
     }
@@ -465,7 +394,6 @@ static void hidraw_device_get_feature_report(struct unix_device *iface, HID_XFER
 
     if (count > 0)
     {
-        hidraw_disable_sony_quirks(iface);
         io->Information = count;
         io->Status = STATUS_SUCCESS;
     }
@@ -502,7 +430,6 @@ static void hidraw_device_set_feature_report(struct unix_device *iface, HID_XFER
 
     if (count > 0)
     {
-        hidraw_disable_sony_quirks(iface);
         io->Information = count;
         io->Status = STATUS_SUCCESS;
     }
@@ -531,29 +458,6 @@ static const struct raw_device_vtbl hidraw_device_vtbl =
 
 #ifdef HAS_PROPER_INPUT_HEADER
 
-static const char *get_device_syspath(struct udev_device *dev)
-{
-    struct udev_device *parent;
-
-    if ((parent = udev_device_get_parent_with_subsystem_devtype(dev, "hid", NULL)))
-        return udev_device_get_syspath(parent);
-
-    if ((parent = udev_device_get_parent_with_subsystem_devtype(dev, "usb", "usb_device")))
-        return udev_device_get_syspath(parent);
-
-    return "";
-}
-
-static struct base_device *find_device_from_syspath(const char *path)
-{
-    struct base_device *impl;
-
-    LIST_FOR_EACH_ENTRY(impl, &device_list, struct base_device, unix_device.entry)
-        if (!strcmp(get_device_syspath(impl->udev_device), path)) return impl;
-
-    return NULL;
-}
-
 #define test_bit(arr,bit) (((BYTE*)(arr))[(bit)>>3]&(1<<((bit)&7)))
 
 /* Minimal compatibility with code taken from steam-runtime-tools */
@@ -919,6 +823,24 @@ static const USAGE_AND_PAGE *what_am_I(struct udev_device *dev, int fd)
 
 static INT count_buttons(int device_fd, BYTE *map)
 {
+    static const UINT gamepad_buttons[] =
+    {
+        BTN_A,
+        BTN_B,
+        BTN_X,
+        BTN_Y,
+        BTN_TL,
+        BTN_TR,
+        BTN_SELECT,
+        BTN_START,
+        BTN_THUMBL,
+        BTN_THUMBR,
+        BTN_MODE,
+        BTN_C,
+        BTN_Z,
+        BTN_TL2,
+        BTN_TR2,
+    };
     int i;
     int button_count = 0;
     BYTE keybits[(KEY_MAX+7)/8];
@@ -929,7 +851,16 @@ static INT count_buttons(int device_fd, BYTE *map)
         return FALSE;
     }
 
-    for (i = BTN_MISC; i < KEY_MAX; i++)
+    for (i = 0; i < ARRAY_SIZE(gamepad_buttons); i++)
+    {
+        if (test_bit(keybits, gamepad_buttons[i]))
+        {
+            if (map) map[gamepad_buttons[i]] = button_count;
+            button_count++;
+        }
+    }
+
+    for (i = BTN_DIGI; i < KEY_MAX; i++)
     {
         if (test_bit(keybits, i))
         {
@@ -963,7 +894,6 @@ static NTSTATUS build_report_descriptor(struct unix_device *iface, struct udev_d
     BYTE absbits[(ABS_MAX+7)/8];
     BYTE relbits[(REL_MAX+7)/8];
     BYTE ffbits[(FF_MAX+7)/8];
-    struct ff_effect effect;
     USAGE_AND_PAGE usage;
     USHORT count = 0;
     USAGE usages[16];
@@ -1039,24 +969,11 @@ static NTSTATUS build_report_descriptor(struct unix_device *iface, struct udev_d
     if (!hid_device_end_input_report(iface))
         return STATUS_NO_MEMORY;
 
-    impl->haptic_effect_id = -1;
+    impl->haptics.id = -1;
     for (i = 0; i < ARRAY_SIZE(impl->effect_ids); ++i) impl->effect_ids[i] = -1;
 
-    if (test_bit(ffbits, FF_RUMBLE))
-    {
-        effect.id = -1;
-        effect.type = FF_RUMBLE;
-        effect.replay.length = 0;
-        effect.u.rumble.strong_magnitude = 0;
-        effect.u.rumble.weak_magnitude = 0;
-
-        if (ioctl(impl->base.device_fd, EVIOCSFF, &effect) == -1)
-            WARN("couldn't allocate rumble effect for haptics: %d %s\n", errno, strerror(errno));
-        else if (!hid_device_add_haptics(iface))
-            return FALSE;
-        else
-            impl->haptic_effect_id = effect.id;
-    }
+    if (test_bit(ffbits, FF_RUMBLE) && !hid_device_add_haptics(iface))
+        return STATUS_NO_MEMORY;
 
     for (i = 0; i < FF_MAX; ++i) if (test_bit(ffbits, i)) break;
     if (i != FF_MAX)
@@ -1151,17 +1068,54 @@ static void lnxev_device_destroy(struct unix_device *iface)
     udev_device_unref(impl->base.udev_device);
 }
 
+static void *lnxev_device_haptics_thread(void *args)
+{
+    struct lnxev_device *impl = lnxev_impl_from_unix_device(args);
+
+    pthread_mutex_lock(&udev_cs);
+
+    for (;;)
+    {
+        struct ff_effect effect;
+
+        while (impl->haptics_state == 1) pthread_cond_wait(&impl->haptics_cond, &udev_cs);
+        if (!impl->haptics_state) break;
+
+        effect = impl->haptics;
+        impl->haptics_state = 1;
+        pthread_mutex_unlock(&udev_cs);
+
+        if (effect.type && (effect.id == -1 || ioctl(impl->base.device_fd, EVIOCSFF, &effect) == -1))
+        {
+            effect.id = -1;
+            ioctl(impl->base.device_fd, EVIOCSFF, &effect);
+        }
+
+        if (effect.id != -1)
+        {
+            struct input_event event = {.type = EV_FF, .code = effect.id, .value = !!effect.type};
+            write(impl->base.device_fd, &event, sizeof(event));
+        }
+
+        pthread_mutex_lock(&udev_cs);
+        impl->haptics.id = effect.id;
+    }
+
+    pthread_mutex_unlock(&udev_cs);
+    return NULL;
+}
+
 static NTSTATUS lnxev_device_start(struct unix_device *iface)
 {
     struct lnxev_device *impl = lnxev_impl_from_unix_device(iface);
-    NTSTATUS status;
-
-    if ((status = build_report_descriptor(iface, impl->base.udev_device)))
-        return status;
 
     pthread_mutex_lock(&udev_cs);
     start_polling_device(iface);
+    impl->haptics_state = 1;
     pthread_mutex_unlock(&udev_cs);
+
+    pthread_cond_init(&impl->haptics_cond, NULL);
+    pthread_create(&impl->haptics_thread, NULL, lnxev_device_haptics_thread, iface);
     return STATUS_SUCCESS;
 }
 
@@ -1172,7 +1126,12 @@ static void lnxev_device_stop(struct unix_device *iface)
     pthread_mutex_lock(&udev_cs);
     stop_polling_device(iface);
     list_remove(&impl->base.unix_device.entry);
+    impl->haptics_state = 0;
     pthread_mutex_unlock(&udev_cs);
+    pthread_cond_signal(&impl->haptics_cond);
+
+    pthread_join(impl->haptics_thread, NULL);
+    pthread_cond_destroy(&impl->haptics_cond);
 }
 
 static void lnxev_device_read_report(struct unix_device *iface)
@@ -1196,39 +1155,18 @@ static NTSTATUS lnxev_device_haptics_start(struct unix_device *iface, UINT durat
                                            USHORT left_intensity, USHORT right_intensity)
 {
     struct lnxev_device *impl = lnxev_impl_from_unix_device(iface);
-    struct ff_effect effect =
-    {
-        .id = impl->haptic_effect_id,
-        .type = FF_RUMBLE,
-    };
-    struct input_event event;
 
     TRACE("iface %p, duration_ms %u, rumble_intensity %u, buzz_intensity %u, left_intensity %u, right_intensity %u.\n",
           iface, duration_ms, rumble_intensity, buzz_intensity, left_intensity, right_intensity);
 
-    effect.replay.length = duration_ms;
-    effect.u.rumble.strong_magnitude = rumble_intensity;
-    effect.u.rumble.weak_magnitude = buzz_intensity;
-
-    if (ioctl(impl->base.device_fd, EVIOCSFF, &effect) == -1)
-    {
-        effect.id = -1;
-        if (ioctl(impl->base.device_fd, EVIOCSFF, &effect) == 1)
-        {
-            WARN("couldn't re-allocate rumble effect for haptics: %d %s\n", errno, strerror(errno));
-            return STATUS_UNSUCCESSFUL;
-        }
-        impl->haptic_effect_id = effect.id;
-    }
-
-    event.type = EV_FF;
-    event.code = effect.id;
-    event.value = 1;
-    if (write(impl->base.device_fd, &event, sizeof(event)) == -1)
-    {
-        WARN("couldn't start haptics rumble effect: %d %s\n", errno, strerror(errno));
-        return STATUS_UNSUCCESSFUL;
-    }
+    pthread_mutex_lock(&udev_cs);
+    impl->haptics.type = FF_RUMBLE;
+    impl->haptics.replay.length = duration_ms;
+    impl->haptics.u.rumble.strong_magnitude = rumble_intensity;
+    impl->haptics.u.rumble.weak_magnitude = buzz_intensity;
+    impl->haptics_state = 2;
+    pthread_mutex_unlock(&udev_cs);
+    pthread_cond_signal(&impl->haptics_cond);
 
     return STATUS_SUCCESS;
 }
@@ -1236,22 +1174,17 @@ static NTSTATUS lnxev_device_haptics_start(struct unix_device *iface, UINT durat
 static NTSTATUS lnxev_device_haptics_stop(struct unix_device *iface)
 {
     struct lnxev_device *impl = lnxev_impl_from_unix_device(iface);
-    struct ff_effect effect =
-    {
-        .id = impl->haptic_effect_id,
-        .type = FF_RUMBLE,
-    };
-    struct input_event event;
 
     TRACE("iface %p.\n", iface);
 
-    if (effect.id == -1) return STATUS_SUCCESS;
-
-    event.type = EV_FF;
-    event.code = effect.id;
-    event.value = 0;
-    if (write(impl->base.device_fd, &event, sizeof(event)) == -1)
-        WARN("couldn't stop haptics rumble effect: %d %s\n", errno, strerror(errno));
+    pthread_mutex_lock(&udev_cs);
+    impl->haptics.type = 0;
+    impl->haptics.replay.length = 0;
+    impl->haptics.u.rumble.strong_magnitude = 0;
+    impl->haptics.u.rumble.weak_magnitude = 0;
+    impl->haptics_state = 2;
+    pthread_mutex_unlock(&udev_cs);
+    pthread_cond_signal(&impl->haptics_cond);
 
     return STATUS_SUCCESS;
 }
@@ -1277,6 +1210,24 @@ static NTSTATUS lnxev_device_physical_effect_run(struct lnxev_device *impl, BYTE
     return STATUS_SUCCESS;
 }
 
+static NTSTATUS lnxev_device_physical_device_set_autocenter(struct unix_device *iface, BYTE percent)
+{
+    struct lnxev_device *impl = lnxev_impl_from_unix_device(iface);
+    struct input_event ie =
+    {
+        .type = EV_FF,
+        .code = FF_AUTOCENTER,
+        .value = 0xffff * percent / 100,
+    };
+
+    TRACE("iface %p, percent %#x.\n", iface, percent);
+
+    if (write(impl->base.device_fd, &ie, sizeof(ie)) == -1)
+        WARN("write failed %d %s\n", errno, strerror(errno));
+
+    return STATUS_SUCCESS;
+}
+
 static NTSTATUS lnxev_device_physical_device_control(struct unix_device *iface, USAGE control)
 {
     struct lnxev_device *impl = lnxev_impl_from_unix_device(iface);
@@ -1320,6 +1271,7 @@ static NTSTATUS lnxev_device_physical_device_control(struct unix_device *iface,
             if (impl->effect_ids[i] < 0) continue;
             lnxev_device_physical_effect_run(impl, i, 0);
         }
+        lnxev_device_physical_device_set_autocenter(iface, 0);
         return STATUS_SUCCESS;
     case PID_USAGE_DC_DEVICE_RESET:
         for (i = 0; i < ARRAY_SIZE(impl->effect_ids); ++i)
@@ -1329,6 +1281,7 @@ static NTSTATUS lnxev_device_physical_device_control(struct unix_device *iface,
                 WARN("couldn't free effect, EVIOCRMFF ioctl failed: %d %s\n", errno, strerror(errno));
             impl->effect_ids[i] = -1;
         }
+        lnxev_device_physical_device_set_autocenter(iface, 100);
         return STATUS_SUCCESS;
     case PID_USAGE_DC_DEVICE_PAUSE:
         WARN("device pause not supported\n");
@@ -1543,14 +1496,14 @@ static const struct hid_device_vtbl lnxev_device_vtbl =
 };
 #endif /* HAS_PROPER_INPUT_HEADER */
 
-static void get_device_subsystem_info(struct udev_device *dev, char const *subsystem, struct device_desc *desc,
-                                      int *bus)
+static void get_device_subsystem_info(struct udev_device *dev, const char *subsystem, const char *devtype,
+                                      struct device_desc *desc, int *bus)
 {
     struct udev_device *parent = NULL;
     const char *ptr, *next, *tmp;
     char buffer[MAX_PATH];
 
-    if (!(parent = udev_device_get_parent_with_subsystem_devtype(dev, subsystem, NULL))) return;
+    if (!(parent = udev_device_get_parent_with_subsystem_devtype(dev, subsystem, devtype))) return;
 
     if ((next = udev_device_get_sysattr_value(parent, "uevent")))
     {
@@ -1582,34 +1535,32 @@ static void get_device_subsystem_info(struct udev_device *dev, char const *subsy
                 if (*bus || desc->vid || desc->pid) continue;
                 sscanf(ptr, "HID_ID=%x:%x:%x\n", bus, &desc->vid, &desc->pid);
             }
-            if (!strncmp(ptr, "PRODUCT=", 8) && *bus != BUS_BLUETOOTH)
+
+            if (!strcmp(subsystem, "input"))
             {
-                if (desc->version) continue;
-                if (!strcmp(subsystem, "usb"))
-                    sscanf(ptr, "PRODUCT=%x/%x/%x\n", &desc->vid, &desc->pid, &desc->version);
-                else
+                if (!strncmp(ptr, "PRODUCT=", 8))
                     sscanf(ptr, "PRODUCT=%x/%x/%x/%x\n", bus, &desc->vid, &desc->pid, &desc->version);
             }
+
+            if (!strcmp(subsystem, "usb") && *bus != BUS_BLUETOOTH)
+            {
+                if (!strncmp(ptr, "PRODUCT=", 8))
+                    sscanf(ptr, "PRODUCT=%x/%x/%x\n", &desc->vid, &desc->pid, &desc->version);
+            }
         }
     }
 
-    if (!desc->manufacturer[0] && (tmp = udev_device_get_sysattr_value(dev, "manufacturer")))
-        ntdll_umbstowcs(tmp, strlen(tmp) + 1, desc->manufacturer, ARRAY_SIZE(desc->manufacturer));
+    if (!strcmp(subsystem, "usb") && *bus != BUS_BLUETOOTH)
+    {
+        if ((tmp = udev_device_get_sysattr_value(parent, "manufacturer")))
+            ntdll_umbstowcs(tmp, strlen(tmp) + 1, desc->manufacturer, ARRAY_SIZE(desc->manufacturer));
 
-    if (!desc->product[0] && (tmp = udev_device_get_sysattr_value(dev, "product")))
-        ntdll_umbstowcs(tmp, strlen(tmp) + 1, desc->product, ARRAY_SIZE(desc->product));
+        if ((tmp = udev_device_get_sysattr_value(parent, "product")))
+            ntdll_umbstowcs(tmp, strlen(tmp) + 1, desc->product, ARRAY_SIZE(desc->product));
 
-    if (!desc->serialnumber[0] && (tmp = udev_device_get_sysattr_value(dev, "serial")))
-        ntdll_umbstowcs(tmp, strlen(tmp) + 1, desc->serialnumber, ARRAY_SIZE(desc->serialnumber));
-}
-
-static void hidraw_set_quirks(struct hidraw_device *impl, DWORD bus_type, WORD vid, WORD pid)
-{
-    if (bus_type == BUS_BLUETOOTH && is_dualshock4_gamepad(vid, pid))
-        impl->quirks |= QUIRK_DS4_BT;
-
-    if (bus_type == BUS_BLUETOOTH && is_dualsense_gamepad(vid, pid))
-        impl->quirks |= QUIRK_DUALSENSE_BT;
+        if ((tmp = udev_device_get_sysattr_value(parent, "serial")))
+            ntdll_umbstowcs(tmp, strlen(tmp) + 1, desc->serialnumber, ARRAY_SIZE(desc->serialnumber));
+    }
 }
 
 static void udev_add_device(struct udev_device *dev, int fd)
@@ -1622,7 +1573,6 @@ static void udev_add_device(struct udev_device *dev, int fd)
     const char *subsystem;
     const char *devnode;
     int bus = 0;
-    int axes = -1, buttons = -1;
 
     if (!(devnode = udev_device_get_devnode(dev)))
     {
@@ -1638,21 +1588,10 @@ static void udev_add_device(struct udev_device *dev, int fd)
 
     TRACE("udev %s syspath %s\n", debugstr_a(devnode), udev_device_get_syspath(dev));
 
-#ifdef HAS_PROPER_INPUT_HEADER
-    if ((impl = find_device_from_syspath(get_device_syspath(dev))))
-    {
-        TRACE("duplicate device found, not adding the new one\n");
-        close(fd);
-        return;
-    }
-
-    axes = count_abs_axis(fd);
-    buttons = count_buttons(fd, NULL);
-#endif
-
-    get_device_subsystem_info(dev, "hid", &desc, &bus);
-    get_device_subsystem_info(dev, "input", &desc, &bus);
-    get_device_subsystem_info(dev, "usb", &desc, &bus);
+    get_device_subsystem_info(dev, "hid", NULL, &desc, &bus);
+    get_device_subsystem_info(dev, "input", NULL, &desc, &bus);
+    get_device_subsystem_info(dev, "usb", "usb_device", &desc, &bus);
+    if (bus == BUS_BLUETOOTH) desc.is_bluetooth = TRUE;
 
     subsystem = udev_device_get_subsystem(dev);
     if (!strcmp(subsystem, "hidraw"))
@@ -1663,6 +1602,7 @@ static void udev_add_device(struct udev_device *dev, int fd)
 #endif
 
         if (!desc.manufacturer[0]) memcpy(desc.manufacturer, hidraw, sizeof(hidraw));
+        desc.is_hidraw = TRUE;
 
 #ifdef HAVE_LINUX_HIDRAW_H
         if (!desc.product[0] && ioctl(fd, HIDIOCGRAWNAME(sizeof(product) - 1), product) >= 0)
@@ -1688,7 +1628,10 @@ static void udev_add_device(struct udev_device *dev, int fd)
         if (!desc.manufacturer[0]) memcpy(desc.manufacturer, evdev, sizeof(evdev));
 
         if (!desc.product[0] && ioctl(fd, EVIOCGNAME(sizeof(buffer) - 1), buffer) > 0)
+        {
+            if (sscanf(buffer, "Microsoft X-Box 360 pad %u", &desc.input) != 1) desc.input = -1;
             ntdll_umbstowcs(buffer, strlen(buffer) + 1, desc.product, ARRAY_SIZE(desc.product));
+        }
 
         if (!desc.serialnumber[0] && ioctl(fd, EVIOCGUNIQ(sizeof(buffer)), buffer) >= 0)
             ntdll_umbstowcs(buffer, strlen(buffer) + 1, desc.serialnumber, ARRAY_SIZE(desc.serialnumber));
@@ -1701,22 +1644,36 @@ static void udev_add_device(struct udev_device *dev, int fd)
         memcpy(desc.serialnumber, zeros, sizeof(zeros));
     }
 
-    if (!is_hidraw_enabled(desc.vid, desc.pid, axes, buttons))
+    if (desc.vid == 0x28de && desc.pid == 0x11ff && !strcmp(subsystem, "input"))
     {
-        TRACE("hidraw %s: deferring %s to a different backend\n", debugstr_a(devnode), debugstr_device_desc(&desc));
+        TRACE("evdev %s: detected steam input virtual controller\n", debugstr_a(devnode));
+        desc.is_gamepad = TRUE;
+        desc.version = 0; /* keep version fixed as 0 so we can hardcode it in ntdll rawinput pipe redirection */
+    }
+    else if (is_sdl_ignored_device(desc.vid, desc.pid))
+    {
+        TRACE("evdev %s: ignoring %s, in SDL ignore list\n", debugstr_a(devnode), debugstr_device_desc(&desc));
         close(fd);
         return;
     }
-    if (is_sdl_blacklisted(desc.vid, desc.pid))
+    else if (!strcmp(subsystem, "input"))
     {
-        /* this device is blacklisted */
-        TRACE("hidraw %s: ignoring %s, in SDL blacklist\n", debugstr_a(devnode), debugstr_device_desc(&desc));
+        TRACE("evdev %s: deferring %s to a different backend\n", debugstr_a(devnode), debugstr_device_desc(&desc));
         close(fd);
         return;
     }
+    else if (is_xbox_gamepad(desc.vid, desc.pid))
+    {
+        desc.is_gamepad = TRUE;
+    }
 #ifdef HAS_PROPER_INPUT_HEADER
-    else
+    else if (!strcmp(subsystem, "input"))
+    {
+        int axes=0, buttons=0;
+        axes = count_abs_axis(fd);
+        buttons = count_buttons(fd, NULL);
         desc.is_gamepad = (axes == 6 && buttons >= 14);
+    }
 #endif
 
     TRACE("dev %p, node %s, desc %s.\n", dev, debugstr_a(devnode), debugstr_device_desc(&desc));
@@ -1729,7 +1686,6 @@ static void udev_add_device(struct udev_device *dev, int fd)
         impl->udev_device = udev_device_ref(dev);
         strcpy(impl->devnode, devnode);
         impl->device_fd = fd;
-        hidraw_set_quirks((struct hidraw_device *)impl, bus, desc.vid, desc.pid);
 
         bus_event_queue_device_created(&event_queue, &impl->unix_device, &desc);
     }
@@ -1743,6 +1699,13 @@ static void udev_add_device(struct udev_device *dev, int fd)
         strcpy(impl->devnode, devnode);
         impl->device_fd = fd;
 
+        if (build_report_descriptor(&impl->unix_device, impl->udev_device))
+        {
+            list_remove(&impl->unix_device.entry);
+            impl->unix_device.vtbl->destroy(&impl->unix_device);
+            return;
+        }
+
         bus_event_queue_device_created(&event_queue, &impl->unix_device, &desc);
     }
 #endif
@@ -2053,26 +2016,38 @@ static void process_monitor_event(struct udev_monitor *monitor)
 {
     struct base_device *impl;
     struct udev_device *dev;
-    const char *action;
+    const char *action, *devnode, *syspath;
 
     dev = udev_monitor_receive_device(monitor);
     if (!dev)
     {
-        FIXME("Failed to get device that has changed\n");
+        ERR("Failed to get device that has changed\n");
         return;
     }
 
     action = udev_device_get_action(dev);
-    TRACE("Received action %s for udev device %s\n", debugstr_a(action),
-          debugstr_a(udev_device_get_devnode(dev)));
+    syspath = udev_device_get_syspath(dev);
+    devnode = udev_device_get_devnode(dev);
+    TRACE("Received action %s for udev device %s (%p) devnode %s\n",
+          debugstr_a(action), debugstr_a(syspath), dev, debugstr_a(devnode));
 
-    if (!action)
-        WARN("No action received\n");
+    if (!syspath)
+        ERR("udev device %p does not have syspath!\n", dev);
+    else if (!action)
+        ERR("event for udev device %s does not have any action!\n", syspath);
+    else if (!devnode)
+    {
+        /* Pretty normal case, not all devices have associated
+         * devnodes. For example root input devices do not, but
+         * related/child mouse and event devices do.
+         */
+        TRACE("udev device %s does not have devnode, ignoring\n", syspath);
+    }
     else if (strcmp(action, "remove"))
         udev_add_device(dev, -1);
     else
     {
-        impl = find_device_from_devnode(udev_device_get_devnode(dev));
+        impl = find_device_from_devnode(devnode);
         if (impl) bus_event_queue_device_removed(&event_queue, &impl->unix_device);
         else WARN("failed to find device for udev device %p\n", dev);
     }
