@@ -44,6 +44,8 @@
 
 #include "../mmdevapi/unixlib.h"
 
+#include "mult.h"
+
 WINE_DEFAULT_DEBUG_CHANNEL(pulse);
 
 enum phys_device_bus_type {
@@ -67,6 +69,7 @@ struct pulse_stream
     float vol[PA_CHANNELS_MAX];
 
     REFERENCE_TIME def_period;
+    REFERENCE_TIME duration;
 
     INT32 locked;
     BOOL started;
@@ -111,14 +114,15 @@ static pa_mainloop *pulse_ml;
 static struct list g_phys_speakers = LIST_INIT(g_phys_speakers);
 static struct list g_phys_sources = LIST_INIT(g_phys_sources);
 
-static const REFERENCE_TIME MinimumPeriod = 30000;
-static const REFERENCE_TIME DefaultPeriod = 100000;
-
 static pthread_mutex_t pulse_mutex;
 static pthread_cond_t pulse_cond = PTHREAD_COND_INITIALIZER;
 
-UINT8 mult_alaw_sample(UINT8, float);
-UINT8 mult_ulaw_sample(UINT8, float);
+static ULONG_PTR zero_bits = 0;
+
+static NTSTATUS pulse_not_implemented(void *args)
+{
+    return STATUS_SUCCESS;
+}
 
 static void pulse_lock(void)
 {
@@ -192,6 +196,25 @@ static int muldiv(int a, int b, int c)
     return ret;
 }
 
+static char *wstr_to_str(const WCHAR *wstr)
+{
+    const int len = wcslen(wstr);
+    char *str = malloc(len * 3 + 1);
+    ntdll_wcstoumbs(wstr, len + 1, str, len * 3 + 1, FALSE);
+    return str;
+}
+
+static BOOL wait_pa_operation_complete(pa_operation *o)
+{
+    if (!o)
+        return FALSE;
+
+    while (pa_operation_get_state(o) == PA_OPERATION_RUNNING)
+        pulse_cond_wait();
+    pa_operation_unref(o);
+    return TRUE;
+}
+
 /* Following pulseaudio design here, mainloop has the lock taken whenever
  * it is handling something for pulse, and the lock is required whenever
  * doing any pa_* call that can affect the state in any way
@@ -225,6 +248,16 @@ static NTSTATUS pulse_process_attach(void *args)
     if (pthread_mutex_init(&pulse_mutex, &attr) != 0)
         pthread_mutex_init(&pulse_mutex, NULL);
 
+#ifdef _WIN64
+    if (NtCurrentTeb()->WowTebOffset)
+    {
+        SYSTEM_BASIC_INFORMATION info;
+
+        NtQuerySystemInformation(SystemEmulationBasicInformation, &info, sizeof(info), NULL);
+        zero_bits = (ULONG_PTR)info.HighestUserAddress | 0x7fffffff;
+    }
+#endif
+
     return STATUS_SUCCESS;
 }
 
@@ -700,7 +733,7 @@ static void pulse_probe_settings(int render, const char *pulse_name, WAVEFORMATE
         ret = -1;
     else if (render)
         ret = pa_stream_connect_playback(stream, pulse_name, &attr,
-        PA_STREAM_START_CORKED|PA_STREAM_FIX_RATE|PA_STREAM_FIX_CHANNELS|PA_STREAM_EARLY_REQUESTS|PA_STREAM_VARIABLE_RATE, NULL, NULL);
+        PA_STREAM_START_CORKED|PA_STREAM_FIX_RATE|PA_STREAM_FIX_CHANNELS|PA_STREAM_EARLY_REQUESTS, NULL, NULL);
     else
         ret = pa_stream_connect_record(stream, pulse_name, &attr, PA_STREAM_START_CORKED|PA_STREAM_FIX_RATE|PA_STREAM_FIX_CHANNELS|PA_STREAM_EARLY_REQUESTS);
     if (ret >= 0) {
@@ -727,12 +760,6 @@ static void pulse_probe_settings(int render, const char *pulse_name, WAVEFORMATE
     if (length)
         *def_period = *min_period = pa_bytes_to_usec(10 * length, &ss);
 
-    if (*min_period < MinimumPeriod)
-        *min_period = MinimumPeriod;
-
-    if (*def_period < DefaultPeriod)
-        *def_period = DefaultPeriod;
-
     wfx->wFormatTag = WAVE_FORMAT_EXTENSIBLE;
     wfx->cbSize = sizeof(WAVEFORMATEXTENSIBLE) - sizeof(WAVEFORMATEX);
 
@@ -762,13 +789,17 @@ static NTSTATUS pulse_test_connect(void *args)
     PhysDevice *dev;
     pa_operation *o;
     int ret;
+    char *name = wstr_to_str(params->name);
 
     pulse_lock();
     pulse_ml = pa_mainloop_new();
 
     pa_mainloop_set_poll_func(pulse_ml, pulse_poll_func, NULL);
 
-    pulse_ctx = pa_context_new(pa_mainloop_get_api(pulse_ml), params->name);
+    pulse_ctx = pa_context_new(pa_mainloop_get_api(pulse_ml), name);
+
+    free(name);
+
     if (!pulse_ctx) {
         ERR("Failed to create context\n");
         pa_mainloop_free(pulse_ml);
@@ -1043,10 +1074,8 @@ static HRESULT pulse_stream_connect(struct pulse_stream *stream, const char *pul
     else
         pulse_name = NULL;  /* use default */
 
-    if (stream->dataflow == eRender) flags |= PA_STREAM_VARIABLE_RATE;
-
     if (stream->dataflow == eRender)
-        ret = pa_stream_connect_playback(stream->stream, pulse_name, &attr, flags, NULL, NULL);
+        ret = pa_stream_connect_playback(stream->stream, pulse_name, &attr, flags|PA_STREAM_VARIABLE_RATE, NULL, NULL);
     else
         ret = pa_stream_connect_record(stream->stream, pulse_name, &attr, flags);
     if (ret < 0) {
@@ -1065,15 +1094,6 @@ static HRESULT pulse_stream_connect(struct pulse_stream *stream, const char *pul
     return S_OK;
 }
 
-static ULONG_PTR zero_bits(void)
-{
-#ifdef _WIN64
-    return !NtCurrentTeb()->WowTebOffset ? 0 : 0x7fffffff;
-#else
-    return 0;
-#endif
-}
-
 static HRESULT get_device_period_helper(EDataFlow flow, const char *pulse_name, REFERENCE_TIME *def, REFERENCE_TIME *min)
 {
     struct list *list = (flow == eRender) ? &g_phys_speakers : &g_phys_sources;
@@ -1100,14 +1120,23 @@ static HRESULT get_device_period_helper(EDataFlow flow, const char *pulse_name,
 static NTSTATUS pulse_create_stream(void *args)
 {
     struct create_stream_params *params = args;
-    REFERENCE_TIME period, duration = params->duration;
     struct pulse_stream *stream;
     unsigned int i, bufsize_bytes;
     HRESULT hr;
+    char *name;
+
+    if (params->share == AUDCLNT_SHAREMODE_EXCLUSIVE) {
+        params->result = AUDCLNT_E_EXCLUSIVE_MODE_NOT_ALLOWED;
+        return STATUS_SUCCESS;
+    }
 
     pulse_lock();
 
-    if (FAILED(params->result = pulse_connect(params->name)))
+    name = wstr_to_str(params->name);
+    params->result = pulse_connect(name);
+    free(name);
+
+    if (FAILED(params->result))
     {
         pulse_unlock();
         return STATUS_SUCCESS;
@@ -1130,21 +1159,16 @@ static NTSTATUS pulse_create_stream(void *args)
     if (FAILED(hr))
         goto exit;
 
-    period = 0;
-    hr = get_device_period_helper(params->flow, params->device, &period, NULL);
-    if (FAILED(hr))
-        goto exit;
+    stream->def_period = params->period;
+    stream->duration = params->duration;
 
-    if (duration < 3 * period)
-        duration = 3 * period;
+    stream->period_bytes = pa_frame_size(&stream->ss) * muldiv(params->period,
+                                                               stream->ss.rate,
+                                                               10000000);
 
-    stream->def_period = period;
-
-    stream->period_bytes = pa_frame_size(&stream->ss) * muldiv(period, stream->ss.rate, 10000000);
-
-    stream->bufsize_frames = ceil((duration / 10000000.) * params->fmt->nSamplesPerSec);
+    stream->bufsize_frames = ceil((params->duration / 10000000.) * params->fmt->nSamplesPerSec);
     bufsize_bytes = stream->bufsize_frames * pa_frame_size(&stream->ss);
-    stream->mmdev_period_usec = period / 10;
+    stream->mmdev_period_usec = params->period / 10;
 
     stream->share = params->share;
     stream->flags = params->flags;
@@ -1161,7 +1185,7 @@ static NTSTATUS pulse_create_stream(void *args)
             size = stream->real_bufsize_bytes =
                 stream->bufsize_frames * 2 * pa_frame_size(&stream->ss);
             if (NtAllocateVirtualMemory(GetCurrentProcess(), (void **)&stream->local_buffer,
-                                        zero_bits(), &size, MEM_COMMIT, PAGE_READWRITE))
+                                        zero_bits, &size, MEM_COMMIT, PAGE_READWRITE))
                 hr = E_OUTOFMEMORY;
         } else {
             UINT32 i, capture_packets;
@@ -1175,7 +1199,7 @@ static NTSTATUS pulse_create_stream(void *args)
 
             size = stream->real_bufsize_bytes + capture_packets * sizeof(ACPacket);
             if (NtAllocateVirtualMemory(GetCurrentProcess(), (void **)&stream->local_buffer,
-                                        zero_bits(), &size, MEM_COMMIT, PAGE_READWRITE))
+                                        zero_bits, &size, MEM_COMMIT, PAGE_READWRITE))
                 hr = E_OUTOFMEMORY;
             else {
                 ACPacket *cur_packet = (ACPacket*)((char*)stream->local_buffer + stream->real_bufsize_bytes);
@@ -1538,7 +1562,6 @@ static NTSTATUS pulse_timer_loop(void *args)
     pa_usec_t last_time;
     UINT32 adv_bytes;
     int success;
-    pa_operation *o;
 
     pulse_lock();
     delay.QuadPart = -stream->mmdev_period_usec * 10;
@@ -1556,13 +1579,7 @@ static NTSTATUS pulse_timer_loop(void *args)
 
         delay.QuadPart = -stream->mmdev_period_usec * 10;
 
-        o = pa_stream_update_timing_info(stream->stream, pulse_op_cb, &success);
-        if (o)
-        {
-            while (pa_operation_get_state(o) == PA_OPERATION_RUNNING)
-                pulse_cond_wait();
-            pa_operation_unref(o);
-        }
+        wait_pa_operation_complete(pa_stream_update_timing_info(stream->stream, pulse_op_cb, &success));
         err = pa_stream_get_time(stream->stream, &now);
         if (err == 0)
         {
@@ -1643,7 +1660,6 @@ static NTSTATUS pulse_start(void *args)
     struct start_params *params = args;
     struct pulse_stream *stream = handle_get_stream(params->stream);
     int success;
-    pa_operation *o;
 
     params->result = S_OK;
     pulse_lock();
@@ -1672,14 +1688,7 @@ static NTSTATUS pulse_start(void *args)
 
     if (pa_stream_is_corked(stream->stream))
     {
-        o = pa_stream_cork(stream->stream, 0, pulse_op_cb, &success);
-        if (o)
-        {
-            while(pa_operation_get_state(o) == PA_OPERATION_RUNNING)
-                pulse_cond_wait();
-            pa_operation_unref(o);
-        }
-        else
+        if (!wait_pa_operation_complete(pa_stream_cork(stream->stream, 0, pulse_op_cb, &success)))
             success = 0;
         if (!success)
             params->result = E_FAIL;
@@ -1698,7 +1707,6 @@ static NTSTATUS pulse_stop(void *args)
 {
     struct stop_params *params = args;
     struct pulse_stream *stream = handle_get_stream(params->stream);
-    pa_operation *o;
     int success;
 
     pulse_lock();
@@ -1719,14 +1727,7 @@ static NTSTATUS pulse_stop(void *args)
     params->result = S_OK;
     if (stream->dataflow == eRender)
     {
-        o = pa_stream_cork(stream->stream, 1, pulse_op_cb, &success);
-        if (o)
-        {
-            while(pa_operation_get_state(o) == PA_OPERATION_RUNNING)
-                pulse_cond_wait();
-            pa_operation_unref(o);
-        }
-        else
+        if (!wait_pa_operation_complete(pa_stream_cork(stream->stream, 1, pulse_op_cb, &success)))
             success = 0;
         if (!success)
             params->result = E_FAIL;
@@ -1769,15 +1770,8 @@ static NTSTATUS pulse_reset(void *args)
         /* If there is still data in the render buffer it needs to be removed from the server */
         int success = 0;
         if (stream->held_bytes)
-        {
-            pa_operation *o = pa_stream_flush(stream->stream, pulse_op_cb, &success);
-            if (o)
-            {
-                while (pa_operation_get_state(o) == PA_OPERATION_RUNNING)
-                    pulse_cond_wait();
-                pa_operation_unref(o);
-            }
-        }
+            wait_pa_operation_complete(pa_stream_flush(stream->stream, pulse_op_cb, &success));
+
         if (success || !stream->held_bytes)
         {
             stream->clock_lastpos = stream->clock_written = 0;
@@ -1819,7 +1813,7 @@ static BOOL alloc_tmp_buffer(struct pulse_stream *stream, SIZE_T bytes)
         stream->tmp_buffer_bytes = 0;
     }
     if (NtAllocateVirtualMemory(GetCurrentProcess(), (void **)&stream->tmp_buffer,
-                                zero_bits(), &bytes, MEM_COMMIT, PAGE_READWRITE))
+                                zero_bits, &bytes, MEM_COMMIT, PAGE_READWRITE))
         return FALSE;
 
     stream->tmp_buffer_bytes = bytes;
@@ -2060,6 +2054,237 @@ static NTSTATUS pulse_release_capture_buffer(void *args)
     return STATUS_SUCCESS;
 }
 
+static NTSTATUS pulse_is_format_supported(void *args)
+{
+    struct is_format_supported_params *params = args;
+    WAVEFORMATEXTENSIBLE in;
+    WAVEFORMATEXTENSIBLE *out;
+    const WAVEFORMATEX *fmt = &in.Format;
+    const BOOLEAN exclusive = params->share == AUDCLNT_SHAREMODE_EXCLUSIVE;
+
+    params->result = S_OK;
+
+    if (!params->fmt_in || (params->share == AUDCLNT_SHAREMODE_SHARED && !params->fmt_out))
+        params->result = E_POINTER;
+    else if (params->share != AUDCLNT_SHAREMODE_SHARED && params->share != AUDCLNT_SHAREMODE_EXCLUSIVE)
+        params->result = E_INVALIDARG;
+    else {
+        memcpy(&in, params->fmt_in, params->fmt_in->wFormatTag == WAVE_FORMAT_EXTENSIBLE ?
+                                    sizeof(in) : sizeof(in.Format));
+
+        if (fmt->wFormatTag == WAVE_FORMAT_EXTENSIBLE) {
+            if (fmt->cbSize < sizeof(WAVEFORMATEXTENSIBLE) - sizeof(WAVEFORMATEX))
+                params->result = E_INVALIDARG;
+            else if (fmt->nAvgBytesPerSec == 0 || fmt->nBlockAlign == 0 ||
+                    (in.Samples.wValidBitsPerSample > fmt->wBitsPerSample))
+                params->result = E_INVALIDARG;
+            else if (fmt->nChannels == 0)
+                params->result = AUDCLNT_E_UNSUPPORTED_FORMAT;
+        }
+    }
+
+    if (FAILED(params->result))
+        return STATUS_SUCCESS;
+
+    if (exclusive)
+        out = &in;
+    else {
+        out = params->fmt_out;
+        memcpy(out, fmt, fmt->wFormatTag == WAVE_FORMAT_EXTENSIBLE ?
+                         sizeof(*out) : sizeof((*out).Format));
+    }
+
+    switch (fmt->wFormatTag) {
+    case WAVE_FORMAT_EXTENSIBLE: {
+        if ((fmt->cbSize != sizeof(WAVEFORMATEXTENSIBLE) - sizeof(WAVEFORMATEX) &&
+             fmt->cbSize != sizeof(WAVEFORMATEXTENSIBLE)) ||
+             fmt->nBlockAlign != fmt->wBitsPerSample / 8 * fmt->nChannels ||
+             in.Samples.wValidBitsPerSample > fmt->wBitsPerSample ||
+             fmt->nAvgBytesPerSec != fmt->nBlockAlign * fmt->nSamplesPerSec) {
+            params->result = E_INVALIDARG;
+            break;
+        }
+
+        if (exclusive) {
+            UINT32 mask = 0, i, channels = 0;
+
+            if (!(in.dwChannelMask & (SPEAKER_ALL | SPEAKER_RESERVED))) {
+                for (i = 1; !(i & SPEAKER_RESERVED); i <<= 1) {
+                    if (i & in.dwChannelMask) {
+                        mask |= i;
+                        ++channels;
+                    }
+                }
+
+                if (channels != fmt->nChannels || (in.dwChannelMask & ~mask)) {
+                    params->result = AUDCLNT_E_UNSUPPORTED_FORMAT;
+                    break;
+                }
+            } else {
+                params->result = AUDCLNT_E_UNSUPPORTED_FORMAT;
+                break;
+            }
+        }
+
+        if (IsEqualGUID(&in.SubFormat, &KSDATAFORMAT_SUBTYPE_IEEE_FLOAT)) {
+            if (fmt->wBitsPerSample != 32) {
+                params->result = E_INVALIDARG;
+                break;
+            }
+
+            if (in.Samples.wValidBitsPerSample != fmt->wBitsPerSample) {
+                params->result = S_FALSE;
+                out->Samples.wValidBitsPerSample = fmt->wBitsPerSample;
+            }
+        } else if (IsEqualGUID(&in.SubFormat, &KSDATAFORMAT_SUBTYPE_PCM)) {
+            if (!fmt->wBitsPerSample || fmt->wBitsPerSample > 32 || fmt->wBitsPerSample % 8) {
+                params->result = E_INVALIDARG;
+                break;
+            }
+
+            if (in.Samples.wValidBitsPerSample != fmt->wBitsPerSample &&
+               !(fmt->wBitsPerSample == 32 &&
+                in.Samples.wValidBitsPerSample == 24)) {
+                params->result = S_FALSE;
+                out->Samples.wValidBitsPerSample = fmt->wBitsPerSample;
+                break;
+            }
+        } else {
+            params->result = AUDCLNT_E_UNSUPPORTED_FORMAT;
+            break;
+        }
+
+        break;
+    }
+    case WAVE_FORMAT_ALAW:
+    case WAVE_FORMAT_MULAW:
+        if (fmt->wBitsPerSample != 8) {
+            params->result = E_INVALIDARG;
+            break;
+        }
+    /* Fall-through */
+    case WAVE_FORMAT_IEEE_FLOAT:
+        if (fmt->wFormatTag == WAVE_FORMAT_IEEE_FLOAT && fmt->wBitsPerSample != 32) {
+            params->result = E_INVALIDARG;
+            break;
+        }
+    /* Fall-through */
+    case WAVE_FORMAT_PCM: {
+        if (fmt->wFormatTag == WAVE_FORMAT_PCM &&
+           (!fmt->wBitsPerSample || fmt->wBitsPerSample > 32 || fmt->wBitsPerSample % 8)) {
+            params->result = E_INVALIDARG;
+            break;
+        }
+
+        if (fmt->nChannels > 2) {
+            params->result = AUDCLNT_E_UNSUPPORTED_FORMAT;
+            break;
+        }
+
+        /* fmt->cbSize, fmt->nBlockAlign and fmt->nAvgBytesPerSec seem to be
+         * ignored, invalid values are happily accepted. */
+        break;
+    }
+    default:
+        params->result = AUDCLNT_E_UNSUPPORTED_FORMAT;
+        break;
+    }
+
+    if (exclusive) { /* This driver does not support exclusive mode. */
+        if (params->result == S_OK)
+            params->result = params->flow == eCapture ?
+                                             AUDCLNT_E_UNSUPPORTED_FORMAT :
+                                             AUDCLNT_E_EXCLUSIVE_MODE_NOT_ALLOWED;
+        else if (params->result == S_FALSE)
+            params->result = AUDCLNT_E_UNSUPPORTED_FORMAT;
+    }
+
+    return STATUS_SUCCESS;
+}
+
+static void sink_name_info_cb(pa_context *c, const pa_sink_info *i, int eol, void *userdata)
+{
+    uint32_t *current_device_index = userdata;
+    pulse_broadcast();
+
+    if (!i || !i->name || !i->name[0])
+        return;
+    *current_device_index = i->index;
+}
+
+struct find_monitor_of_sink_cb_param
+{
+    struct get_loopback_capture_device_params *params;
+    uint32_t current_device_index;
+};
+
+static void find_monitor_of_sink_cb(pa_context *c, const pa_source_info *i, int eol, void *userdata)
+{
+    struct find_monitor_of_sink_cb_param *p = userdata;
+    unsigned int len;
+
+    pulse_broadcast();
+
+    if (!i || !i->name || !i->name[0])
+        return;
+    if (i->monitor_of_sink != p->current_device_index)
+        return;
+
+    len = strlen(i->name) + 1;
+    if (len <= p->params->ret_device_len)
+    {
+        memcpy(p->params->ret_device, i->name, len);
+        p->params->result = STATUS_SUCCESS;
+        return;
+    }
+    p->params->ret_device_len = len;
+    p->params->result = STATUS_BUFFER_TOO_SMALL;
+}
+
+static NTSTATUS pulse_get_loopback_capture_device(void *args)
+{
+    struct get_loopback_capture_device_params *params = args;
+    uint32_t current_device_index = PA_INVALID_INDEX;
+    struct find_monitor_of_sink_cb_param p;
+    const char *device_name;
+    char *name;
+
+    pulse_lock();
+
+    if (!pulse_ml)
+    {
+        pulse_unlock();
+        ERR("Called without main loop running.\n");
+        params->result = E_INVALIDARG;
+        return STATUS_SUCCESS;
+    }
+
+    name = wstr_to_str(params->name);
+    params->result = pulse_connect(name);
+    free(name);
+
+    if (FAILED(params->result))
+    {
+        pulse_unlock();
+        return STATUS_SUCCESS;
+    }
+
+    device_name = params->device;
+    if (device_name && !device_name[0]) device_name = NULL;
+
+    params->result = E_FAIL;
+    wait_pa_operation_complete(pa_context_get_sink_info_by_name(pulse_ctx, device_name, &sink_name_info_cb, &current_device_index));
+    if (current_device_index != PA_INVALID_INDEX)
+    {
+        p.current_device_index = current_device_index;
+        p.params = params;
+        wait_pa_operation_complete(pa_context_get_source_info_list(pulse_ctx, &find_monitor_of_sink_cb, &p));
+    }
+
+    pulse_unlock();
+    return STATUS_SUCCESS;
+}
+
 static NTSTATUS pulse_get_mix_format(void *args)
 {
     struct get_mix_format_params *params = args;
@@ -2269,30 +2494,42 @@ static NTSTATUS pulse_set_sample_rate(void *args)
     struct set_sample_rate_params *params = args;
     struct pulse_stream *stream = handle_get_stream(params->stream);
     HRESULT hr = S_OK;
-    pa_operation *o;
     int success;
+    pa_sample_spec new_ss;
 
     pulse_lock();
-    if (!pulse_stream_valid(stream))
+    if (!pulse_stream_valid(stream)) {
         hr = AUDCLNT_E_DEVICE_INVALIDATED;
-    else
-    {
-        if (!(o = pa_stream_update_sample_rate(stream->stream, params->new_rate, pulse_op_cb, &success)))
-            success = 0;
-        else
-        {
-            while (pa_operation_get_state(o) == PA_OPERATION_RUNNING)
-                pthread_cond_wait(&pulse_cond, &pulse_mutex);
-            pa_operation_unref(o);
-        }
-
-        if (!success) hr = E_FAIL;
-        else
-        {
-            stream->ss.rate = params->new_rate;
-            stream->period_bytes = pa_frame_size(&stream->ss) * muldiv(stream->mmdev_period_usec, stream->ss.rate, 1000000);
-        }
+        goto exit;
     }
+    if (stream->dataflow != eRender) {
+        hr = E_NOTIMPL;
+        goto exit;
+    }
+
+    new_ss = stream->ss;
+    new_ss.rate = params->rate;
+
+    if (!wait_pa_operation_complete(pa_stream_update_sample_rate(stream->stream, params->rate, pulse_op_cb, &success)))
+        success = 0;
+
+    if (!success) {
+        hr = E_OUTOFMEMORY;
+        goto exit;
+    }
+
+    if (stream->held_bytes)
+        wait_pa_operation_complete(pa_stream_flush(stream->stream, pulse_op_cb, &success));
+
+    stream->clock_lastpos = stream->clock_written = 0;
+    stream->pa_offs_bytes = stream->lcl_offs_bytes = 0;
+    stream->held_bytes = stream->pa_held_bytes = 0;
+    stream->period_bytes = pa_frame_size(&new_ss) * muldiv(stream->mmdev_period_usec, new_ss.rate, 1000000);
+    stream->ss = new_ss;
+
+    silence_buffer(new_ss.format, stream->local_buffer, stream->real_bufsize_bytes);
+
+exit:
     pulse_unlock();
 
     params->result = hr;
@@ -2364,7 +2601,6 @@ static NTSTATUS pulse_get_prop_value(void *args)
     struct list *list = (params->flow == eRender) ? &g_phys_speakers : &g_phys_sources;
     PhysDevice *dev;
 
-    params->result = S_OK;
     LIST_FOR_EACH_ENTRY(dev, list, PhysDevice, entry) {
         if (strcmp(params->device, dev->pulse_name))
             continue;
@@ -2376,18 +2612,19 @@ static NTSTATUS pulse_get_prop_value(void *args)
             case 0:   /* FormFactor */
                 params->value->vt = VT_UI4;
                 params->value->ulVal = dev->form;
+                params->result = S_OK;
                 return STATUS_SUCCESS;
             case 3:   /* PhysicalSpeakers */
                 if (!dev->channel_mask)
                     goto fail;
                 params->value->vt = VT_UI4;
                 params->value->ulVal = dev->channel_mask;
+                params->result = S_OK;
                 return STATUS_SUCCESS;
-            default:
-                params->result = E_NOTIMPL;
             }
         }
 
+        params->result = E_NOTIMPL;
         return STATUS_SUCCESS;
     }
 
@@ -2412,7 +2649,8 @@ const unixlib_entry_t __wine_unix_call_funcs[] =
     pulse_release_render_buffer,
     pulse_get_capture_buffer,
     pulse_release_capture_buffer,
-    NULL,
+    pulse_is_format_supported,
+    pulse_get_loopback_capture_device,
     pulse_get_mix_format,
     pulse_get_device_period,
     pulse_get_buffer_size,
@@ -2427,14 +2665,16 @@ const unixlib_entry_t __wine_unix_call_funcs[] =
     pulse_test_connect,
     pulse_is_started,
     pulse_get_prop_value,
-    NULL,
-    NULL,
-    NULL,
-    NULL,
-    NULL,
-    NULL,
+    pulse_not_implemented,
+    pulse_not_implemented,
+    pulse_not_implemented,
+    pulse_not_implemented,
+    pulse_not_implemented,
+    pulse_not_implemented,
 };
 
+C_ASSERT(ARRAYSIZE(__wine_unix_call_funcs) == funcs_count);
+
 #ifdef _WIN64
 
 typedef UINT PTR32;
@@ -2579,6 +2819,55 @@ static NTSTATUS pulse_wow64_get_capture_buffer(void *args)
     return STATUS_SUCCESS;
 };
 
+static NTSTATUS pulse_wow64_is_format_supported(void *args)
+{
+    struct
+    {
+        PTR32 device;
+        EDataFlow flow;
+        AUDCLNT_SHAREMODE share;
+        PTR32 fmt_in;
+        PTR32 fmt_out;
+        HRESULT result;
+    } *params32 = args;
+    struct is_format_supported_params params =
+    {
+        .device = ULongToPtr(params32->device),
+        .flow = params32->flow,
+        .share = params32->share,
+        .fmt_in = ULongToPtr(params32->fmt_in),
+        .fmt_out = ULongToPtr(params32->fmt_out)
+    };
+    pulse_is_format_supported(&params);
+    params32->result = params.result;
+    return STATUS_SUCCESS;
+}
+
+static NTSTATUS pulse_wow64_get_loopback_capture_device(void *args)
+{
+    struct
+    {
+        PTR32 name;
+        PTR32 device;
+        PTR32 ret_device;
+        UINT32 ret_device_len;
+        HRESULT result;
+    } *params32 = args;
+
+    struct get_loopback_capture_device_params params =
+    {
+        .name = ULongToPtr(params32->name),
+        .device = ULongToPtr(params32->device),
+        .ret_device = ULongToPtr(params32->ret_device),
+        .ret_device_len = params32->ret_device_len,
+    };
+
+    pulse_get_loopback_capture_device(&params);
+    params32->result = params.result;
+    params32->ret_device_len = params.ret_device_len;
+    return STATUS_SUCCESS;
+}
+
 static NTSTATUS pulse_wow64_get_mix_format(void *args)
 {
     struct
@@ -2741,7 +3030,6 @@ static NTSTATUS pulse_wow64_set_volumes(void *args)
         float master_volume;
         PTR32 volumes;
         PTR32 session_volumes;
-        int channel;
     } *params32 = args;
     struct set_volumes_params params =
     {
@@ -2749,7 +3037,6 @@ static NTSTATUS pulse_wow64_set_volumes(void *args)
         .master_volume = params32->master_volume,
         .volumes = ULongToPtr(params32->volumes),
         .session_volumes = ULongToPtr(params32->session_volumes),
-        .channel = params32->channel
     };
     return pulse_set_volumes(&params);
 }
@@ -2860,7 +3147,8 @@ const unixlib_entry_t __wine_unix_call_wow64_funcs[] =
     pulse_release_render_buffer,
     pulse_wow64_get_capture_buffer,
     pulse_release_capture_buffer,
-    NULL,
+    pulse_wow64_is_format_supported,
+    pulse_wow64_get_loopback_capture_device,
     pulse_wow64_get_mix_format,
     pulse_wow64_get_device_period,
     pulse_wow64_get_buffer_size,
@@ -2871,16 +3159,18 @@ const unixlib_entry_t __wine_unix_call_wow64_funcs[] =
     pulse_wow64_get_position,
     pulse_wow64_set_volumes,
     pulse_wow64_set_event_handle,
-    NULL,
+    pulse_set_sample_rate,
     pulse_wow64_test_connect,
     pulse_is_started,
     pulse_wow64_get_prop_value,
-    NULL,
-    NULL,
-    NULL,
-    NULL,
-    NULL,
-    NULL,
+    pulse_not_implemented,
+    pulse_not_implemented,
+    pulse_not_implemented,
+    pulse_not_implemented,
+    pulse_not_implemented,
+    pulse_not_implemented,
 };
 
+C_ASSERT(ARRAYSIZE(__wine_unix_call_wow64_funcs) == funcs_count);
+
 #endif /* _WIN64 */
