@@ -21,8 +21,6 @@
 #ifndef __WOW64_STRUCT32_H
 #define __WOW64_STRUCT32_H
 
-#include "wine/server.h"
-
 typedef struct
 {
     ULONG Length;
@@ -75,6 +73,13 @@ typedef struct
     ULONG            DefaultNonPagedPoolCharge;
 } OBJECT_TYPE_INFORMATION32;
 
+typedef struct
+{
+    CONTEXT_CHUNK All;
+    CONTEXT_CHUNK Legacy;
+    CONTEXT_CHUNK XState;
+} CONTEXT_EX32;
+
 typedef struct
 {
     UNICODE_STRING32 ObjectName;
@@ -96,7 +101,11 @@ typedef struct
 
 typedef struct
 {
-    BOOLEAN ReplaceIfExists;
+    union
+    {
+        BOOLEAN ReplaceIfExists;
+        ULONG Flags;
+    };
     ULONG   RootDirectory;
     ULONG   FileNameLength;
     WCHAR   FileName[1];
@@ -192,6 +201,23 @@ typedef struct
     MEMORY_WORKING_SET_EX_BLOCK32 VirtualAttributes;
 } MEMORY_WORKING_SET_EX_INFORMATION32;
 
+typedef struct
+{
+    ULONG ImageBase;
+    ULONG SizeOfImage;
+    union
+    {
+        ULONG ImageFlags;
+        struct
+        {
+            ULONG ImagePartialMap : 1;
+            ULONG ImageNotExecutable : 1;
+            ULONG ImageSigningLevel : 4;
+            ULONG Reserved : 26;
+        };
+    };
+} MEMORY_IMAGE_INFORMATION32;
+
 typedef struct
 {
     NTSTATUS  ExitStatus;
@@ -202,13 +228,6 @@ typedef struct
     ULONG     InheritedFromUniqueProcessId;
 } PROCESS_BASIC_INFORMATION32;
 
-typedef struct
-{
-    ULONG Version;
-    ULONG Reserved;
-    ULONG Callback;
-} PROCESS_INSTRUMENTATION_CALLBACK_INFORMATION32;
-
 typedef struct
 {
     ULONG ReserveSize;
@@ -307,6 +326,30 @@ typedef struct
     ULONG                            DefaultBase;
 } RTL_PROCESS_MODULE_INFORMATION_EX32;
 
+typedef struct
+{
+    ULONG Flags;
+    union
+    {
+        ULONG TlsVector;
+        ULONG TlsModulePointer;
+    };
+    ULONG ThreadId;
+} THREAD_TLS_INFORMATION32;
+
+typedef struct
+{
+    ULONG Flags;
+    ULONG OperationType;
+    ULONG ThreadDataCount;
+    union
+    {
+        ULONG TlsIndex;
+        ULONG TlsVectorLength;
+    };
+    THREAD_TLS_INFORMATION32 ThreadData[1];
+} PROCESS_TLS_INFORMATION32;
+
 typedef struct
 {
     ULONG         BaseAddress;
@@ -341,6 +384,11 @@ typedef struct
     DWORD Attributes;
 } SID_AND_ATTRIBUTES32;
 
+typedef struct
+{
+    SID_AND_ATTRIBUTES32 Label;
+} TOKEN_MANDATORY_LABEL32;
+
 typedef struct
 {
     ULONG DefaultDacl;
@@ -357,6 +405,11 @@ typedef struct
     ULONG Owner;
 } TOKEN_OWNER32;
 
+typedef struct
+{
+    ULONG PrimaryGroup;
+} TOKEN_PRIMARY_GROUP32;
+
 typedef struct
 {
     SID_AND_ATTRIBUTES32 User;
@@ -493,6 +546,12 @@ typedef struct
     ULONG Flags;
 } SYSTEM_CACHE_INFORMATION32;
 
+typedef struct
+{
+    ULONG ProcessId;
+    UNICODE_STRING32 ImageName;
+} SYSTEM_PROCESS_ID_INFORMATION32;
+
 typedef struct
 {
     ULONG  OwnerPid;
@@ -667,29 +726,44 @@ typedef struct
     ULONG NumberOfBytes;
 } MEMORY_RANGE_ENTRY32;
 
-struct __server_iovec32
+typedef struct
 {
-    ULONG        ptr;
-    data_size_t  size;
-};
+    ULONG LowestStartingAddress;
+    ULONG HighestEndingAddress;
+    ULONG Alignment;
+} MEM_ADDRESS_REQUIREMENTS32;
 
-struct __server_request_info32
+typedef struct DECLSPEC_ALIGN(8)
 {
+    struct
+    {
+        DWORD64 Type : MEM_EXTENDED_PARAMETER_TYPE_BITS;
+        DWORD64 Reserved : 64 - MEM_EXTENDED_PARAMETER_TYPE_BITS;
+    };
     union
     {
-        union generic_request req;
-        union generic_reply   reply;
-    } u;
-    unsigned int            data_count;
-    ULONG                   reply_data;
-    struct __server_iovec32 data[__SERVER_MAX_DATA];
-};
+        DWORD64 ULong64;
+        ULONG   Pointer;
+        ULONG   Size;
+        ULONG   Handle;
+        ULONG   ULong;
+    };
+} MEM_EXTENDED_PARAMETER32;
 
 typedef struct
 {
-  ULONG LowestStartingAddress;
-  ULONG HighestEndingAddress;
-  ULONG Alignment;
-} MEM_ADDRESS_REQUIREMENTS32;
+    ULONG Token;
+    ULONG Thread;
+} PROCESS_ACCESS_TOKEN32;
+
+typedef struct
+{
+    ULONG PagedPoolLimit;
+    ULONG NonPagedPoolLimit;
+    ULONG MinimumWorkingSetSize;
+    ULONG MaximumWorkingSetSize;
+    ULONG PagefileLimit;
+    LARGE_INTEGER TimeLimit;
+} QUOTA_LIMITS32;
 
 #endif /* __WOW64_STRUCT32_H */
