@@ -437,11 +437,8 @@ struct security_descriptor *mode_to_sd( mode_t mode, const struct sid *user, con
     sd->sacl_len = 0;
     sd->dacl_len = dacl_size;
 
-    ptr = (char *)(sd + 1);
-    memcpy( ptr, user, sd->owner_len );
-    ptr += sd->owner_len;
-    memcpy( ptr, group, sd->group_len );
-    ptr += sd->group_len;
+    ptr = mem_append( sd + 1, user, sd->owner_len );
+    ptr = mem_append( ptr, group, sd->group_len );
 
     dacl = (struct acl *)ptr;
     dacl->revision = ACL_REVISION;
@@ -620,7 +617,7 @@ mode_t sd_to_mode( const struct security_descriptor *sd, const struct sid *owner
                     {
                         bits_to_set &= ~((mode << 6) | (mode << 3));  /* user + group */
                     }
-                    else if (equal_sid( sid, owner ))
+                    else if (equal_sid( sid, owner ) || equal_sid( sid, &owner_rights_sid ))
                     {
                         bits_to_set &= ~(mode << 6);  /* user only */
                     }
@@ -639,7 +636,7 @@ mode_t sd_to_mode( const struct security_descriptor *sd, const struct sid *owner
                         new_mode |= mode & bits_to_set;
                         bits_to_set &= ~mode;
                     }
-                    else if (equal_sid( sid, owner ))
+                    else if (equal_sid( sid, owner ) || equal_sid( sid, &owner_rights_sid ))
                     {
                         mode = (mode << 6);  /* user only */
                         new_mode |= mode & bits_to_set;
